#include <u.h>
#include <libc.h>
#include <mswin.h>
#include <winbase.h>

#include <userenv.h>

/*
 * windows quoting rules - I think
 * Words are seperated by space or tab
 * Words containing a space or tab can be quoted using "
 * 2N backslashes + " ==> N backslashes and end quote
 * 2N+1 backslashes + " ==> N backslashes + literal "
 * N backslashes not followed by " ==> N backslashes
 */
static char *
dblquote(char *cmd, char *s)
{
	int nb;
	char *p;

	for(p=s; *p; p++)
		if(*p == ' ' || *p == '\t' || *p == '\n' || *p == '\r' || *p == '"')
			break;

	if(p == s){					/* empty arg */
		strcpy(cmd, "\"\"");
		return cmd+2;
	}

	if(*p == 0){				/* easy case */
		strcpy(cmd, s);
		return cmd+(p-s);
	}

	*cmd++ = '"';
	for(;;) {
		for(nb=0; *s=='\\'; nb++)
			*cmd++ = *s++;

		if(*s == 0) {				/* trailing backslashes -> 2N */
			while(nb-- > 0)
				*cmd++ = '\\';
			break;
		}

		if(*s == '"') {			/* literal quote -> 2N+1 backslashes */
			while(nb-- > 0)
				*cmd++ = '\\';
			*cmd++ = '\\';			/* escape the quote */
		}
		*cmd++ = *s++;
	}

	*cmd++ = '"';
	*cmd = 0;

	return cmd;
}

void
usage(void)
{
	fprint(2, "usage: %s username[@domain]:passwd cmd...\n", argv0);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int i, n;
	void *env;
	HANDLE tok;
	char *buf, *p;
	DWORD status;
	STARTUPINFOW si;
	wchar_t cwd[8192];
	PROCESS_INFORMATION pi;
	wchar_t *user, *pass, *cli;

	ARGBEGIN{
	}ARGEND;

	if(argc < 2)
		usage();

	if((p = strchr(argv[0], ':')) == nil)
		usage();
	*p++ = 0;

	user = _winstr(argv[0]);
	pass = _winstr(p);

	n = 0;
	for(i = 1; i < argc; i++)
		n += strlen(argv[i]) +3;

	if((buf = malloc(n * sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");

	p = buf;
	for(i = 1; argv[i]; i++){
		p = dblquote(p, argv[i]);
		*p++ = ' ';
	}
	if(p != buf )
		p--;
	*p = 0;

	cli = _winstr(buf);

	*cwd = 0;
	GetCurrentDirectory(nelem(cwd), cwd);

	if(LogonUser(user, nil, pass, LOGON32_LOGON_INTERACTIVE, 
            LOGON32_PROVIDER_DEFAULT, &tok) == 0)
		sysfatal("LogonUser failed, %s", buf, _winerr());
    if(CreateEnvironmentBlock(&env, tok, TRUE) == 0)
		sysfatal("CreateEnvironmentBlock failed, %s", buf, _winerr());
	CloseHandle(tok);

 	memset(&pi, 0, sizeof(pi));
	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);

	if(CreateProcessWithLogonW(user, nil, pass, LOGON_WITH_PROFILE,
		nil, cli, CREATE_UNICODE_ENVIRONMENT, env, cwd, &si, &pi) == 0)
		sysfatal("CreateProcessAsUser('%s') failed, %s", buf, _winerr());

	if(WaitForSingleObject(pi.hProcess, INFINITE) != 0)
		sysfatal("WaitForSingleObject() failed, %s", _winerr());

	if(GetExitCodeProcess(pi.hProcess, &status) == 0)
		sysfatal("GetExitCodeProcess() failed, %s", _winerr());

	if(status != 0)
		sysfatal("%s exited - 0x%x", argv[1], status);

	exits(nil);
}
