/*
 * serial only terminal prog
 *
 * I could port con.c but windows doesn't allow files to be called con
 * as it looks like a device so, as we changed the name anyway, and
 * terminal config is somwhat broque in most operating systems, I created
 * this - mswin/term.
 * 
 * NB: <del> just interrupts us and sends nothing. we
 * could do better but this is good enough
 * for now.
 */

#include <u.h>
#include <libc.h>
#include <mswin.h>

static int strip;
static int returns;
static int notkbd;
static int cooked;
static int nltocr;
static int crtonl;
static int raw;
static int baud;
static int verbose;
static HANDLE cons;

enum {
	MAXMSG = 8192
};


int
wasintr(void)
{
	return 0;
}

void
rawon(void)
{
	DWORD state;

	if(raw)
		return;

	if(! cons)
		return;

	state = 0;
	if(SetConsoleMode(cons, state) == 0)
		sysfatal("cannot set console mode, %s", _winerr());

	raw = 1;
}

void
rawoff(void)
{
	DWORD state;

	if(! raw)
		return;

	if(! cons)
		return;

	state = ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT;
	if(SetConsoleMode(cons, state) == 0)
		sysfatal("cannot set console mode, %s", _winerr());

	raw = 0;
}

long
iread(HANDLE h, void *a, int n)
{
	DWORD m;

	do{
		m = 0;
		if(ReadFile(h, a, n, &m, nil) == 0)
			return -1;
	}while(m == 0);
	return m;
}

long
iwrite(HANDLE h, void *a, int n)
{
	DWORD m;

	m = 0;
	if(WriteFile(h, a, n, &m, nil) == 0)
		return -1;
	return m;
}

/*
 *  Read from the network and write to the screen.
 *  Filter out spurious carriage returns.
 */
void
fromnet(HANDLE net)
{
	long n;
	char buf[MAXMSG];
	char *cp, *ep;

	for(;;){
		n = iread(net, buf, sizeof(buf));
		if(n < 0)
			return;
		if(n == 0)
			continue;

		if(strip)
			for (cp=buf; cp<buf+n; cp++)
				*cp &= 0177;

		if(crtonl) {
			/* convert cr's to nl's */
			for (cp = buf; cp < buf + n; cp++)
				if (*cp == '\r')
					*cp = '\n';
		}
		else
		if(!returns){
			/* strip cariage returns */
			cp = buf;
			ep = buf + n;
			while(cp < ep && (cp = memchr(cp, '\r', ep-cp))){
				memmove(cp, cp+1, ep-cp-1);
				ep--;
				n--;
			}
		}
		if(n > 0 && write(1, buf, n) != n){
			fprint(2, "write error %r\n");
			return;
		}
	}
}

/*
 *  control menu
 */
#define STDHELP	"\t(b)reak, (q)uit, (i)nterrupt, toggle printing (r)eturns, (.)continue, (!cmd)\n"

int
menu(HANDLE net)
{
	char buf[MAXMSG];
	long n;
	int done;
	int wasraw = raw;

	if(wasraw)
		rawoff();

	fprint(2, ">>> ");
	for(done = 0; !done; ){
		n = read(0, buf, sizeof(buf)-1);
		if(n <= 0)
			return -1;
		buf[n] = 0;
		switch(buf[0]){
		case '.':
			done = 1;
			break;
		case 'q':
			return -1;
		case 'i':
			buf[0] = 0x1c;
			iwrite(net, buf, 1);
			done = 1;
			break;
		case 'r':
			returns = 1-returns;
			done = 1;
			break;
		default:
			fprint(2, STDHELP);
			break;
		}
		if(!done)
			fprint(2, ">>> ");
	}

	if(wasraw)
		rawon();
	else
		rawoff();
	return 0;
}
/*
 *  Read the keyboard and write it to the network.  '^\' gets us into
 *  the menu.
 */
void
fromkbd(HANDLE net)
{
	long m, n;
	char buf[MAXMSG];
	char *cp, *ep;
	int eofs;

	iwrite(net, "\n", 1);

	eofs = 0;
	for(;;){
		n = read(0, buf, sizeof(buf));
		if(n < 0){
			if(wasintr()){
				if(!raw){
					buf[0] = 0x7f;
					n = 1;
				} else
					continue;
			} else
				return;
		}
		if(n == 0){
			if(++eofs > 32)
				return;
		} else
			eofs = 0;
		if(n && memchr(buf, 0x1c, n)){
			if(menu(net) < 0)
				return;
		}else{
			if(!raw && n==0){
				buf[0] = 0x4;
				n = 1;
			}
			if(!returns){
				/* strip cariage returns */
				cp = buf;
				ep = buf + n;
				while(cp < ep && (cp = memchr(cp, '\r', ep-cp))){
					memmove(cp, cp+1, ep-cp-1);
					ep--;
					n--;
				}
			}
			if(nltocr){
				ep = buf+n;
				for(cp = buf; cp < ep; cp++)
					switch(*cp){
					case '\r':
						*cp = '\n';
						break;
					case '\n':
						*cp = '\r';
						break;
					}
			}
			if((m = iwrite(net, buf, n)) == 0 || n != m)
				return;
		}
	}
}

void
device(char *dev)
{
	DCB dcb;
	char *path;
	wchar_t *w;
	COMMTIMEOUTS cto;
	HANDLE net, thr;

	path = smprint("\\\\.\\%s", dev);
	w = _winstr(path);		/* not _winpath, as it translatest to an abs path */

	net = CreateFile(w, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, nil);
	if(net == INVALID_HANDLE_VALUE)
		sysfatal("%W cannot open, %s", w, _winerr());
	free(path);
	
	memset(&dcb, 0, sizeof(dcb));
	dcb.DCBlength = sizeof(dcb);
	if(GetCommState(net, &dcb) == 0)
		sysfatal("%s cannot get params, %r", dev, _winerr());

	dcb.BaudRate = baud;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;

	dcb.fBinary = FALSE;
	dcb.fParity = TRUE;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;

	if(SetCommState(net, &dcb) == 0)
		sysfatal("%s cannot set params", dev, _winerr());

	cto.ReadIntervalTimeout = 0;
	cto.ReadTotalTimeoutMultiplier = 0;
	cto.ReadTotalTimeoutConstant = 10;

	cto.WriteTotalTimeoutMultiplier = 0;
	cto.WriteTotalTimeoutConstant = 10;

	if(SetCommTimeouts(net, &cto) == 0)
		sysfatal("%s cannot set timeouts", dev, _winerr());

	if(verbose){
		if(GetCommState(net, &dcb) == 0)
			sysfatal("%s cannot get params", dev, _winerr());
   		print("baud=%d parity=%d data=%d stop=%d\n", 
              dcb.BaudRate, dcb.Parity, dcb.ByteSize, dcb.StopBits+1);
	}

	if(!cooked)
		rawon();

	EscapeCommFunction(net, SETDTR);

	thr = CreateThread(nil, 0, (LPTHREAD_START_ROUTINE)fromkbd, net, 0, nil);
	if(thr == INVALID_HANDLE_VALUE)
		sysfatal("cannot create thread, %s\n", _winerr());
	CloseHandle(thr);

	fromnet(net);

	EscapeCommFunction(net, CLRDTR);
	rawoff();

	fprint(2, "\nhangup\n");
}

static void
usage(void)
{
	fprint(2, "usage: %s [-rsnCRT] [-b rate] device\n", argv0);
	fprint(2, "   device        set the device-file for comms (default com4)\n");
	fprint(2, "   -b baudrate   set the baud rate (default 115200)\n");
	fprint(2, "   -C            cooked mode (local echo)\n");
	fprint(2, "   -s            strip data from remote to 7 bits\n");
	fprint(2, "   -r            keep carriage returns\n");
	fprint(2, "   -R            translate outgoing newlines to carriage returns, and vice verca\n");
	fprint(2, "   -T            translate incomming carriage returns to newlines\n");
	exits("usage");
}


void
main(int argc, char *argv[])
{
	char *dev;
	DWORD oldmode;

	_widefmtinstall();

	oldmode = 0;
	if(AttachConsole(getpid()) ||					/* successful attach */
	  GetLastError() == ERROR_ACCESS_DENIED){		/* was already attached */
		cons = GetStdHandle(STD_INPUT_HANDLE);
		if(GetConsoleMode(cons, &oldmode) == 0)
			cons = NULL;
	}

	dev = "com4";
	baud = CBR_115200;
	ARGBEGIN{
	case 'b':
		baud = atoi(EARGF(usage()));
		break;
	case 'r':
		returns = 1;
		break;
	case 's':
		strip = 1;
		break;
	case 'n':
		notkbd = 1;
		break;
	case 'C':
		cooked = 1;
		break;
	case 'R':
		nltocr = 1;
		break;
	case 'T':
		crtonl = 1;
		break;
	case 'v':
		verbose = 1;
		break;
	default:
		usage();
		break;
	}ARGEND;

	if(argc == 1)
		dev = argv[0];

	device(dev);

	if(cons)
		if(SetConsoleMode(cons, oldmode) == 0)
			sysfatal("cannot restore console mode, %s", _winerr());

	exits(nil);
}

