#include <u.h>
#include <libc.h>
#include <mswin.h>

int	failed;
int	gflag;
int	uflag;
int	xflag;
void	copy(char *from, char *to, int todir);

void
main(int argc, char *argv[])
{
	Dir *dirb;
	int todir, i;

	ARGBEGIN {
	case 'g':
		sysfatal("not supported on win32");
		break;
	case 'u':
		sysfatal("not supported on win32");
		break;
	case 'x':
		xflag++;
		break;
	default:
		goto usage;
	} ARGEND

	_widefmtinstall();
	quotefmtinstall();

	todir=0;
	if(argc < 2)
		goto usage;
	dirb = dirstat(argv[argc-1]);
	if(dirb!=nil && (dirb->mode&DMDIR))
		todir=1;
	if(argc>2 && !todir){
		fprint(2, "cp: %s not a directory\n", argv[argc-1]);
		exits("bad usage");
	}
	for(i=0; i<argc-1; i++)
		copy(argv[i], argv[argc-1], todir);
	if(failed)
		exits("errors");
	exits(0);

usage:
	fprint(2, "usage:\tcp [-gux] fromfile tofile\n");
	fprint(2, "\tcp [-x] fromfile ... todir\n");
	exits("usage");
}

/*
 * This is expensive and only works on NTFS
 * so this check is only done on the error path.
 */
void
copy(char *from, char *to, int todir)
{
	HANDLE h;
	FILETIME ft;
	SYSTEMTIME st;
	wchar_t *src, *dst;

	src = _winpath(from);
	if(todir){
		wchar_t *s, *elem;
		elem=s=src;
		while(*s++)
			if(s[-1] == L'/' || s[-1] == L'\\')
				elem=s;
		dst = _winpath2(to, elem);
	}
	else
		dst = _winpath(to);

	/* server side copy is worth having */
	if(CopyFile(src, dst, 0) == 0){
		werrstr("%s", _winerr());
		if(_samefile(src, dst))
			fprint(2, "cp: same file\n", src, dst);
		else
			fprint(2, "cp: %W %W copy failed - %r\n", src, dst);
		failed = 1;
	}
	free(src);

	 /* NB: CopyFile() clones attributes and times by default */
	if(! failed && ! xflag){
		if((h = CreateFile(dst, GENERIC_WRITE, 0, nil, OPEN_EXISTING, 0, nil)) == BadHand){
			fprint(2, "cp: %W cannot open to set file times - %r\n", dst);
			failed = 1;
		}
		else{
			GetSystemTime(&st);
			SystemTimeToFileTime(&st, &ft);
			if(SetFileTime(h, &ft, &ft, &ft) == 0){
				fprint(2, "cp: %W cannot set file times - %r\n", dst);
				failed = 1;
			}
			CloseHandle(h);
		}
	}
	free(dst);
}

