#define _SPAWN_EXTENSION
#include <u.h>
#include <libc.h>
#include <thread.h>
#include <9pclient.h>
#include <mswin.h>

#include <lm.h>
#include <userenv.h>

enum {
	Nname = 128,		/* max size of a user or domain name */
};

static char *host;
static char *plan9;
static char *cputype;

void
logmsg(int fatal, char *fmt, ...)
{
	int fd;
	va_list args;
	static char *logfile;

	if(logfile == nil)
		logfile = smprint("%s/sys/log/cpu", plan9);

	if(access(logfile, 2) == -1)
		return;

	va_start(args, fmt);
	if((fd = open(logfile, OWRITE)) == -1)
		return;
	seek(fd, 0, 2);				/* no append-only files on windows */
	fprint(fd, "%ld ", time(nil));
	vfprint(fd, fmt, args);
	fprint(fd, "\n");
	close(fd);

	if(fatal){
		fprint(2, "%s: ", argv0);
		vfprint(2, fmt, args);
		fprint(2, "\n");
		threadexitsall("failed");
	}
	va_end(args);
}

/*
 * strncpy() for wide chars, but guarantees to
 * terminate the destination string
 */
wchar_t *
wclipcpy(wchar_t *t, wchar_t *s, int sz)
{
	int n;

	n = _winlen(s)+1;
	if(n > sz)
		n = sz;
	memcpy(t, s, n*sizeof(wchar_t));
	t[n-1] = 0;

	return t;
}

void
wputenv(wchar_t *nam, wchar_t *val)
{
	char *n, *v;

	n = _ninestr(nam);
	v = _ninestr(val);
	putenv(n, v);
	free(n);
	free(v);
}

int
readstr(int fd, char *str, int len)
{
	int n, l;
	char *s;

	s = str;
	for(l = 0; l < len; l++) {
		n = read(fd, s, 1);
		if(n < 0)
			return -1;
		if(*s == '\0')
			return 0;
		s++;
	}
	return -1;
}

int
writestr(int fd, char *str)
{
	int l, n;

	l = strlen(str);
	n = write(fd, str, l+1);
	return n;
}

/* remove any vestigial drive mappings which could block our new ones */
void
unmapall(void)
{
	int e;
	HANDLE h;
	void *buf;
	NETRESOURCE *p;
	unsigned long n, l;
	enum { Nbuf = 16 * 1024 };

	if((buf = malloc(Nbuf)) == NULL)
		logmsg(1, "No memory, %r");

	if((e = WNetOpenEnum(RESOURCE_CONNECTED, RESOURCETYPE_ANY,
	    RESOURCEUSAGE_ALL, NULL, &h)) != NO_ERROR)
		logmsg(1, "cannot enumerate connected network resources, %s", _wneterr(e));

	n = -1;
	l = Nbuf;
	while(WNetEnumResource(h, &n, buf, &l) == NO_ERROR){
		for(p = buf; n--; p++){
			if(WNetCancelConnection2(p->lpLocalName, 0, 1) != NO_ERROR)
				fprint(2, "%s %s - unmap failed, %r", p->lpLocalName, p->lpRemoteName, _wneterr(e));
				logmsg(0, "%s %s - unmap failed, %r", p->lpLocalName, p->lpRemoteName, _wneterr(e));
		}
		n = -1;
		l = Nbuf;
	}
	WNetCloseEnum(h);
}

/* map drives as requested by remote */
void
mapdrives(char *dom)
{
#ifdef WAITING_FOR_9P_CLIENT
	int e;
	NETRESOURCE nr;
	char *path1, *path2, *ruser;
	wchar_t usr[Nname], pass[Nname];
	wchar_t loc[MAX_PATH], rem[MAX_PATH];

	ruser = getremuser();

	path1 = smprint("/mnt/term/usr/%s/lib/dosmap/%s", ruser, host);
	if((fd = np_open(path1, OREAD);
	if(fd == -1){
		path2 = smprint("/mnt/term/usr/%s/lib/dosmap/%s", ruser, dom);
		fd = np_open(path2, OREAD);
	}
	if(fd == -1){
		fprint(2, "%s: neither '%s' or '%s' found, no drives mapped\n", path1, path2)
		free(path1);
		free(path2);
		free(ruser);
	}

read file
factotum exchange on each one to get auth info

	while(getwstr(loc, sizeof(loc))){
		getwstr(fd, rem, sizeof(rem));
		getwstr(fd, usr, sizeof(usr));
		getwstr(fd, pass, sizeof(pass));

		memset(&nr, 0, sizeof(NETRESOURCE));
		nr.dwType = RESOURCETYPE_DISK;
		nr.lpLocalName = loc;  
		nr.lpRemoteName	= rem;

		if((e = WNetAddConnection2(&nr, pass, usr, CONNECT_TEMPORARY)) != NO_ERROR){
			werrstr(_wneterr(e));
			fprint(2, "map: %W %W %W failed, %r\n", loc, rem, usr);
		}
		memset(pass, 0, sizeof(pass));
	}
#endif
}

HANDLE
authenticate(char *dom, char *user, char *pass, void **env, char *home)
{
	int e;
	HANDLE tok;
	wchar_t *updc;
	USER_INFO_3 *ui;
	PROFILEINFO pri;
	wchar_t *wuser, *wdom, *wpass;

	wdom = _winstr(dom);
	wuser = _winstr(user);
	wpass = _winstr(pass);

	updc = NULL;
	if((e = NetGetDCName(NULL, wdom, (void *)&updc)) != NERR_Success){
		werrstr("get DC name %s", _wneterr(e));
		return nil;
	}

	if((e = NetUserGetInfo(updc, wuser, 3, (uchar **)&ui)) != NERR_Success){
		werrstr("get user info %s", _wneterr(e));
		return nil;
	}
	if(ui->usri3_home_dir)
		snprint(home, MAX_PATH, "%w/%w", ui->usri3_home_dir_drive, ui->usri3_home_dir);
	else
		snprint(home, MAX_PATH, "%w", ui->usri3_home_dir_drive);

	if(ui->usri3_password_expired){
		werrstr("password expired");
		return nil;
	}

	if(LogonUser(wuser, wdom, wpass, LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT, &tok) == 0){
		werrstr("%s", _winerr());
		return nil;
	}

	memset(&pri, 0, sizeof(pri));
	pri.dwSize = sizeof(pri);
	pri.lpUserName = wuser;
	pri.lpProfilePath = ui->usri3_profile;

	if(LoadUserProfile(tok, &pri) == 0){
		werrstr("load profile %s", _winerr());
		return nil;
	}

	free(wuser);
	free(wpass);
	free(wdom);
	NetApiBufferFree(updc);
	NetApiBufferFree(ui);
	return tok;
}

/*
 * We assume the environment variables $cputype,
 * $plan9 and $sysname have already been set by listen(1)
 */
void
threadmain(int argc, char *argv[])
{
	void *env;
	CFsys *fs;
	Waitmsg *w;
	HANDLE tok;
	int fd, pid;
	char buf[ERRMAX], user[Nname], pass[Nname], *dom;
	char *full, home[MAX_PATH], err[ERRMAX];
	char cwd[MAX_PATH], cmd[nelem(cwd)], authalg[Nname];

	argv0 = "cpu";	/* make error messages look more traditional */

	if((plan9 = getenv("plan9")) == nil)
		logmsg(1, "$plan9 environment variable not set");
	if((cputype = getenv("cputype")) == nil)
		logmsg(1, "$cputype environment variable not set");
	if((host = getenv("sysname")) == nil)
		logmsg(1, "$sysname environment variable not set");

	fd = 0;

	_widefmtinstall();
	quotefmtinstall();

	readstr(fd, authalg, nelem(authalg));
	if(strncmp(authalg, "tlspass", 6) != 0){
		writestr(fd, "username contains no '@'");
		logmsg(1, "%q only tlspass auth supported", authalg);
	}
	writestr(1, "");
	
//	fd = pushtls(0, "hashalg", "encalg", 0, "secret", "dir");

	if(readstr(fd, user, nelem(user)) == -1)
		logmsg(1, "get user@domain failed, %r");
	if(readstr(fd, pass, nelem(pass)) == -1)
		logmsg(1, "get password failed, %r");

	if((dom = strchr(user, '@')) == nil){
		writestr(fd, "username contains no '@'");
		logmsg(1, "%q contains no '@'", user);
	}
	*dom++ = 0;

	tok = authenticate(dom, user, pass, &env, home);
	if(tok == nil){
		errstr(err, sizeof err);
		writestr(fd, err);
		logmsg(1, "authentication failed, %r");
	}
	writestr(fd, "");		/* Auth ok (no error string) */

	unmapall();

	if(ImpersonateLoggedOnUser(tok) == 0)
		logmsg(1, "cannot impersonate user, %s", _winerr());

	cmd[0] = 0;
	readstr(fd, cwd, nelem(cwd));
	if(cwd[0] == '!'){
		strcpy(cmd, cwd+1);
		readstr(fd, cwd, nelem(cwd));
	}

#ifndef Plan9port_9pclient
	writestr(fd, "sorry, No 9p client yet\n");
#endif
	writestr(fd, "FS");					/* OK to start 9p server on client */
	writestr(fd, "/");					/* directory to mount */
	readstr(fd, buf, sizeof(buf));
	if(strcmp(buf, "OK") != 0)
		logmsg(1, "mount of server failed, %s", buf);

#ifdef Plan9port_9pclient
	fs = fsinit(fd);
#endif
	mapdrives(dom);

	if(strcmp(cwd, "NO") == 0)	/* client is lost so just go home */
		chdir(home);
	else
	if(chdir(cwd) == -1){
		fprint(2, "%s: failed to chdir to '%W'\n", argv0, cwd);
		chdir(home);			/* client in namespace we don't share so just go home */
	}

	if(rfork(RFCFDG|RFENVG) == -1)
		logmsg(1, "cpu fork failed, %r");

	/* build new environment */
	if(CreateEnvironmentBlock(&env, tok, TRUE) == 0)
		logmsg(1, "cannot create enviroment block, %s", _winerr());

	_importenv(env);
	putenv("service", "cpu");
	putenv("user", user);
	putenv("home", home);

/* FIXME: still need line dicipline to:
 * 		^D to EOF (zero length write down a pipe),
 * 		read /mnt/term/dev/cpunote and send singals
 *		read cwd pipe and write to /mnt/term/dev/wsys
 */

	/* open stdin, out and error */
// duplicating sockets might cause problems, maybe need a special dup.
//	dup(open("/mnt/term/dev/cons", OREAD), 0);
//	dup(open("/mnt/term/dev/cons", OREAD), 1);
//	dup(open("/mnt/term/dev/cons", OREAD), 2);

	/* create a thread reading /mnt/term/dev/cpunote */

	/* create a named pipe and a thread to read it and and send the results to /mnt/term/dev/wdir */

	full = smprint("%s/%s/bin/rc.exe", plan9, cputype);

	if(*cmd)
		pid = _spawnl(tok, full, "rc", "-lc", cmd, nil);
	else
		pid = _spawnl(tok, full, "rc", "-li", nil);

	logmsg(0, "cpu %s@%s %q %q %d", user, dom, cwd, cmd, pid);

loop:
	w = wait();
	if(w == nil){
		errstr(err, sizeof err);
		if(strcmp(err, "interrupted") == 0)
			goto loop;
		logmsg(1, "wait failed, %r");
	}
	if(w->pid == pid){
		if(w->msg[0])
			logmsg(0, "child exited %s", w->msg);
		free(w);
	}

	threadexitsall(nil);
}
