/*
 *
 * 9p client
 *	/dev/cpunote - interrupt
 *	/dev/wdir - for the plumber
 *	/dev/label - ?
 *	/dev/cons - stdin, stdout, stderr
 *	/env - export some/all of enviroment
 *	timezone?
 *
 * interrupt handling - <del> causes interrupt to be sent to shell
 * eof handling - ^d causes shell's input fifo to be closed.
 * exetype for rc(1)
 *
 * TLS before passwords exchanged
 *
 * Winstations should be allocated per user rather than per session.
 */
#include <windows.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <userenv.h>
#include <lm.h>
#include "ninep.h"

enum {
	MAX_MSG = 8000		// fits inside a 9p message
};

/* doesn't appear to be in mingw's headers or libraries */
#define SecureZeroMemory(p, n)	memset(p, 0, n)

static char user[128] = "guest";
static char sysname[128] = "";

extern char *argv0;

typedef struct {
	char *name;		// debug
	Fid *fid;
	HANDLE pipe;
	int pid;
} Relay;

static void
getstr(HANDLE h, char *buf, int n)
{
	char c, *p;
	unsigned long got;

	p = buf;
	while(n--){
		if(ReadFile(h, &c, 1, &got, NULL) == 0)
			break;
		if(got != 1 || c == '\n' || c == 0)
			break;
		*p++ = c;
	}
	*p = 0;
}

static void
putstr(HANDLE h, char *fmt, ...)
{
	va_list args;
	char s[1024];
	unsigned long x;

	va_start(args, fmt);
	x = vsnprintf(s, sizeof(s), fmt, args);
	va_end(args);

	WriteFile(h, s, x+1, &x, NULL);
}
char *
lookenv(wchar_t *wenv, char *tok)
{
	int l;
	static char val[Npath];
	wchar_t wtok[Npath], *p;

	l = mbstowcs(wtok, tok, sizeof(wtok)/sizeof(wtok[0]));
	for(p = wenv; *p; p++){
		if (wcsncmp(wtok, p, l) == 0)
			break;
		p += wcslen(p);
	}
	if (*p == 0)
		return NULL;
	wcstombs(val, p+l+1, sizeof(val));
	return val;
}


void
setenviro(char *home)
{
	char buf[Npath];
	SYSTEM_INFO si;
	unsigned long n;

	GetSystemInfo(&si);
	switch(si.wProcessorArchitecture){
	case PROCESSOR_ARCHITECTURE_AMD64:
		SetEnvironmentVariable("cputype", "win/amd64");
		SetEnvironmentVariable("objtype", "win/amd64");
		break;
	case PROCESSOR_ARCHITECTURE_IA64:
		SetEnvironmentVariable("cputype", "win/alpha");
		SetEnvironmentVariable("objtype", "win/alpha");
		break;
	case PROCESSOR_ARCHITECTURE_INTEL:
		SetEnvironmentVariable( "cputype", "win/386");
		SetEnvironmentVariable("objtype", "win/386");
		break;
	}

	/*
	 * Allow the sysname passed from the client to override
	 * the machines real system name - this is handy for me (steve)
	 * as our PCs are given unhelpful names based on their
	 * asset number.
	 */
	if(! *sysname){
		n = sizeof(sysname);
		GetComputerName(sysname, &n);
		strlwr(sysname);
		SetEnvironmentVariable("sysname", sysname);
	}

	SetEnvironmentVariable("service", "cpu");
	SetEnvironmentVariable("home", home);
	SetEnvironmentVariable("user", user);
	snprintf(buf, sizeof(buf), "%lu", si.dwNumberOfProcessors);
	SetEnvironmentVariable("NPROC", buf);

	/*
	 * Note: this strange prompt is understood by checkprompt()
	 * which uses it to set the current directory in /dev/wdir.
	 *
	 * The plumber consults this file to generate absolute paths
	 * for filenames it is given, so files in servers mounted on
	 * both local and remote systems can be plumed as though
	 * they are local; this idea was shamelessly stolen from
	 * conswdir(1) on plan9.
	 */
	snprintf(buf, sizeof(buf), "$E];$p\007%s%% ", sysname);
	SetEnvironmentVariable("prompt", buf);
}

void
getdefdom(char *dom, int n)
{
	HKEY key;
	DWORD sz;
	char *path, *attr;

	sz = n;
	attr = "DefaultDomainName";
	path = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon";

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, path, 0, KEY_READ, &key) != ERROR_SUCCESS)
		sysfatal("Reg open failed");
	if(RegQueryValueEx(key, attr, NULL, NULL, dom, &sz) != ERROR_SUCCESS)
		sysfatal("Reg query value failed");
	dom[sz] = 0;
}

int
auth(HANDLE hi, HANDLE ho, char *defdom, HANDLE *token)
{
	int rc;
	char *authdom, err[255], pass[Npath];

	putstr(ho, "%s", defdom);
	getstr(hi, user, sizeof(user));
	getstr(hi, pass, sizeof(pass));

	if ((authdom = strchr(user, '@')) != NULL)
		*authdom++ = 0;
	else
		authdom = defdom;

	SetLastError(0);
	rc = LogonUser(user, authdom, pass, LOGON32_LOGON_INTERACTIVE,
	    LOGON32_PROVIDER_DEFAULT, token);
	memset(pass, 0x00, sizeof(pass));
	if(rc == 0){
		errstr(GetLastError(), err, sizeof(err));
		putstr(ho, "%s", err);
		return -1;
	}
	putstr(ho, "OK");
	return 0;
}

/*
 * get users home and profile directories, we try to get these from
 * the network you are connected to, if not then we try to get the
 * local directory - c:/mnt/term/Documents and Settings/%USERNAME%
 * and as a last resort we use c:/
 */
void
getaccount(HANDLE token, char *defdom, char *prof, int nprof, char *home, int nhome)
{
	unsigned long n;
	USER_INFO_3 *ui;
	wchar_t wusr[Npath], wdom[Npath], *updc;

	*prof = 0;
	snprintf(home, nhome, "c:/");

	n = nhome;
	GetUserProfileDirectory(token, home, &n);

	updc = NULL;
	mbstowcs(wdom, defdom, sizeof(wdom)/sizeof(wdom[0]));
	if(NetGetDCName(NULL, wdom, (void *)&updc) != NERR_Success)
		return;

	mbstowcs(wusr, user, sizeof(wusr)/sizeof(wusr[0]));
	if(NetUserGetInfo(updc, wusr, 3, (unsigned char **)&ui) != NERR_Success){
		NetApiBufferFree(updc);
		return;
	}
	NetApiBufferFree(updc);

	wcstombs(prof, ui->usri3_profile, sizeof(prof));
	wcstombs(home, ui->usri3_home_dir_drive, sizeof(home));
	if(ui->usri3_password_expired)
		sysfatal("%s@%s - password expired", user, defdom);
	NetApiBufferFree(ui);
}
		

static unsigned long WINAPI
relayinp(void *v)
{
	int len;
	unsigned long n;
	char buf[MAX_MSG];
	Relay *r = (Relay *)v;


syslog("start: %s", r->name);
	while(1) {
		while((len = read9(r->fid, buf, sizeof(buf))) == 0)
			continue;
		if(len < 0)
			break;
syslog("%s: len=%d %.*s", r->name, len, len, buf);
		n = 0;
		if(WriteFile(r->pipe, buf, len, &n, 0) == 0 || n != len)
			break;
	}
syslog("exit: %s", r->name);
	return 0;
}

/*
 * We try to prevent packet fragmenetation by sleeping for a 1ms
 * once we read a char and then read what remains in the input fifo.
 */
static unsigned long WINAPI
relayout(void *v)
{
	char buf[MAX_MSG];
	unsigned long len;
	Relay *r = (Relay *)v;

syslog("start: %s", r->name);
	while(1) {
		len = 0;
		if(ReadFile(r->pipe, buf, 1, &len, 0) == 0){
			syslog("%s read1 fail %ld ", r->name, len);
			break;
		}
		Sleep(1);
		if(PeekNamedPipe(r->pipe, NULL, 0, NULL, &len, NULL) == 0){
			syslog("%s peek fail %ld ", r->name, len);
			break;
		}
		if(len > MAX_MSG-1)
			len = MAX_MSG-1;
		if(ReadFile(r->pipe, buf+1, len, &len, 0) == 0){
			syslog("%s read2 fail %ld ", r->name, len);
			break;
		}
		len++;

syslog("%s: len=%d %.*s", r->name, len, len, buf);
		if(write9(r->fid, buf, len) != len)
			break;
	}
syslog("exit: %s", r->name);
	return 0;
}


static unsigned long WINAPI
relaynotes(void *v)
{
	int len;
	char buf[MAX_MSG];
	Relay *r = (Relay *)v;

syslog("start: %s", r->name);
	while(1) {
		if((len = read9(r->fid, buf, sizeof(buf))) < 1)
			break;
		if (strcmp(buf, "hangup") == 0)
			break;
syslog("%s len=%d %.*s", r->name, len, len, buf);
		if (strcmp(buf, "interrupt") == 0)
			GenerateConsoleCtrlEvent(CTRL_C_EVENT, r->pid);
	}
syslog("exit: %s", r->name);
	return 0;
}

/*
 * Note the note Relay has no pipe associated with it, it sends
 * interrupt events to the child process rather than streams of text.
 */
int
runcmd(HANDLE token, NineP *np, char *cmdline, wchar_t *env)
{
	int i, n;
	STARTUPINFO si;
	unsigned long s;
	SECURITY_ATTRIBUTES sa;
	PROCESS_INFORMATION pi;
	Relay inp, out, err, note;
	HANDLE early[3], late[2], hnote;

	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	si.dwFlags = STARTF_USESTDHANDLES;
	si.lpDesktop = "";		// create a new winstation per session

	sa.nLength = sizeof(sa);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;

	CreatePipe(&si.hStdInput, &inp.pipe, &sa, MAX_MSG);
	CreatePipe(&out.pipe, &si.hStdOutput, &sa, MAX_MSG);
	CreatePipe(&err.pipe, &si.hStdError, &sa, MAX_MSG);
	SetHandleInformation(si.hStdInput, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);
	SetHandleInformation(si.hStdOutput, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);
	SetHandleInformation(si.hStdError, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);

	memset(&pi, 0, sizeof(pi));
	if(CreateProcessAsUser(token, NULL, cmdline, NULL, NULL, TRUE,
	    CREATE_UNICODE_ENVIRONMENT, env, NULL, &si, &pi) == 0)
		sysfatal("'%s' - cannot create process", cmdline);
	note.pid = pi.dwProcessId;

//	CloseHandle(si.hStdInput);
//	CloseHandle(si.hStdOutput);
//	CloseHandle(si.hStdError);

	RevertToSelf();

	if ((inp.fid = open9(np, "/mnt/term/dev/cons", OREAD)) == nil)
		sysfatal("/mnt/term/dev/cons cannot open for kbd");
	if ((note.fid = open9(np, "/mnt/term/dev/cpunote", OREAD)) == nil)
		sysfatal("/mnt/term/dev/cpunote cannot open for notes");
	if ((out.fid = open9(np, "/mnt/term/dev/cons", OWRITE)) == nil)
		sysfatal("/mnt/term/dev/cons cannot open for out");
	if ((err.fid = open9(np, "/mnt/term/dev/cons", OWRITE)) == nil)
		sysfatal("/mnt/term/dev/cons cannot open for err");

	early[0] = pi.hProcess;
	inp.name = "relay kbd";
	early[1] = CreateThread(NULL, 0, relayinp, (void *)&inp, 0, NULL);
	note.name = "relay notes";
	early[2] = CreateThread(NULL, 0, relaynotes,  (void *)&note, 0, NULL);

	out.name = "relay stdout";
	late[0] = CreateThread(NULL, 0, relayout, (void *)&out, 0, NULL);
	err.name = "relay stderr";
	late[1] = CreateThread(NULL, 0, relayout, (void *)&err, 0, NULL);

	/*
	 * first await keyboard hangup or process exit, then
	 * wait until output has drained before tidying up
	 */
	n = WaitForMultipleObjects(nelem(early), early, FALSE, INFINITE);
syslog("exited early: %d", n);
	n = WaitForMultipleObjects(nelem(late), late, TRUE, 100);	// allow output to drain
syslog("exited late: %d", n);

	if(GetExitCodeProcess(early[0], &s) && s == STILL_ACTIVE)
		TerminateProcess(early[0], 0);
	for(i = 1; i < nelem(early); i++)
		if(GetExitCodeThread(early[i], &s) && s == STILL_ACTIVE)
			TerminateThread(early[i], 0);
	for(i = 0; i < nelem(late); i++)
		if(GetExitCodeThread(late[i], &s) && s == STILL_ACTIVE)
			TerminateThread(late[i], 0);

	close9(inp.fid);
	close9(out.fid);
	close9(err.fid);
	close9(note.fid);

	CloseHandle(inp.pipe);
	CloseHandle(out.pipe);
	CloseHandle(err.pipe);
	CloseHandle(note.pipe);

	CloseHandle(pi.hThread);
	CloseHandle(early[0]);
	CloseHandle(early[1]);
	CloseHandle(late[0]);
	CloseHandle(late[1]);
	CloseHandle(hnote);
	return 0;
}

int
main(int argc, char *argv[])
{
	void *env;
	NineP *np;
	int gotcmd;
	PROFILEINFO pri;
	HANDLE hi, ho, token;
	char cmd[Npath*4], xdir[Npath*4], prof[Npath], home[Npath];
	char cmdline[Npath*4], alg[64], defdom[64], shell[Npath];

	argv0 = "cpud";
	/* argv0 = argv[0]; */
	hi = GetStdHandle(STD_INPUT_HANDLE); 
        ho = GetStdHandle(STD_OUTPUT_HANDLE); 
        SetConsoleMode(hi, ENABLE_PROCESSED_INPUT); 
        SetConsoleMode(ho, 0); 

	getstr(hi, alg, sizeof(alg));
	if(strncmp(alg, "tlspass ", 8) != 0){
		putstr(ho, "unsupported auth method");
		sysfatal("unsupported auth method");
	}
	putstr(ho, "");		// no news is good news

	getdefdom(defdom, sizeof(defdom));
	if (auth(hi, ho, defdom, &token) == -1)
		sysfatal("authentication failed");

	*cmd = 0;
	gotcmd = 0;
	getstr(hi, xdir, sizeof(xdir));
	if(xdir[0] == '!') {
		strcpy(cmd, &xdir[1]);
		gotcmd = 1;
		getstr(hi, xdir, sizeof(xdir));
	}

	getaccount(token, defdom, prof, sizeof(prof), home, sizeof(home));

	/*
	 * load users Windows Profile (registry hive)
	 */
	memset(&pri, 0, sizeof(pri));
	pri.dwSize = sizeof(pri);
	pri.lpUserName = user;
	pri.lpProfilePath = prof;
	if(LoadUserProfile(token, &pri) == 0)
		syslog("cannot load user profile");

	setenviro(home);

	if(CreateEnvironmentBlock(&env, token, TRUE) == 0)
		syslog("cannot create enviroment block");

	if (GetEnvironmentVariable("COMSPEC", shell, sizeof(shell)) == 0)
		strcpy(shell, "cmd.exe");
	if(gotcmd)
		snprintf(cmdline, sizeof(cmdline), "%s /q /a /c %s", shell, cmd);
	else
		snprintf(cmdline, sizeof(cmdline), "%s /q /a /k", shell);

	if ((np = attach9(hi, ho, "/mnt/term", "")) == NULL)
		sysfatal("cannot attach to 9 server");

	/*
	 * read mapping required and any
	 * additional bindings present in
	 * the users plan9 namespace.
	 */
	if(getmap(np, sysname) == -1)
		getmap(np, defdom);
	getns(np);

	/*
	 * map network drives
	 */
	unmapall();
	if(ImpersonateLoggedOnUser(token) == 0)
		syslog("impersonation failed");
	mapall();

	if(strcmp(xdir, "NO") == 0){
		SetCurrentDirectory(home);
	}
	else{
		if(SetCurrentDirectory(xdir) == 0)
			syslog("failed to chdir to '%s'", xdir);
		SetCurrentDirectory(home);
	}

/*
 *  tell other end to start their 9p server
 *
 *	putstr(ho, "FS");
 *	putstr(ho, "/");
 */
	runcmd(token, np, cmdline, env);

	DestroyEnvironmentBlock(env);
	UnloadUserProfile(token, pri.hProfile);
	detach9(np);
	exit(0);
}
