#include <windows.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <userenv.h>
#include <lm.h>
#include "ninep.h"

enum{			// states for prompt scanning
	None,
	Esc,
	Brack,
	Semi,
	Bell,
};

typedef struct Map Map;
struct Map {
	Map *next;
	char *path;	// plan9 path
	int plen;
	char *share;	// remote share
	char *mount;	// windows mountpoint
	int mlen;
};
Map *Maproot = nil;

static int
getpid9(NineP *np)
{
	Fid *fid;
	char buf[Npath];

	if ((fid = open9(np, "/mnt/term/dev/pid", OREAD)) == nil)
		return -1;
	if(line9(fid, buf, sizeof(buf)) < 0){
		close9(fid);
		return -1;
	}
	close9(fid);
	return atoi(buf);
}

static char *
getuser9(NineP *np)
{
	Fid *fid;
	static char buf[Npath];

	if ((fid = open9(np, "/mnt/term/dev/user", OREAD)) == nil)
		return nil;
	if(line9(fid, buf, sizeof(buf)) < 0){
		close9(fid);
		return nil;
	}
	close9(fid);
	return buf;
}

int
getmap(NineP *np, char *name)
{
	Fid *fid;
	Map *m, **p;
	char buf[Npath], file[Npath], *mount, *share, *path;

	snprintf(file, sizeof(file), "/mnt/term/usr/%s/lib/dosmap/%s", getuser9(np), name);
	if ((fid = open9(np, file, OREAD)) == nil)
		return -1;

	p = &Maproot;
	while(line9(fid, buf, sizeof(buf)) != -1){
		if((path = strtok(buf, " \t")) == NULL)
			continue;
		if(*path == '#')
			continue;
		if((mount = strtok(NULL, " \t")) == NULL)
			continue;
		if((share = strtok(NULL, " \t")) == NULL)
			continue;
		if ((m = malloc(sizeof(Map))) == nil)
			sysfatal("no memory - %r");
		if ((m->path = strdup(path)) == nil)
			sysfatal("no memory - %r");
		m->plen = strlen(m->path);
		if ((m->mount = strdup(mount)) == nil)
			sysfatal("no memory - %r");
		m->mlen = strlen(m->mount);
		if ((m->share = strdup(share)) == nil)
			sysfatal("no memory - %r");
		*p = m;
		p = &(m->next);
		*p = nil;
	}
	close9(fid);
	return 0;
}


int
getns(NineP *np)
{
	Fid *fid;
	Map *m, *n;
	char buf[Npath], file[Npath], *cmd, *src, *dst;

	snprintf(file, sizeof(file), "/mnt/term/proc/%d/ns", getpid9(np));
	if ((fid = open9(np, file, OREAD)) == nil)
		return -1;

	while(line9(fid, buf, sizeof(buf)) != -1){
		if((cmd = strtok(buf, " \t")) == NULL)
			continue;
		if(strcmp(cmd, "bind") != 0)
			continue;
		while((src = strtok(NULL, " \t")) != NULL)
			if(*src != '-')			// options
				break;
		if(src == nil)
			continue;
		if((dst = strtok(NULL, " \t")) == NULL)
			continue;

		for (n = Maproot; n; n = n->next)
			if (strncmp(n->path, src, n->plen) == 0)
				break;
		if (n == nil)
			continue;

		if ((m = malloc(sizeof(Map))) == nil)
			sysfatal("no memory - %r");
		if ((m->path = strdup(dst)) == nil)
			sysfatal("no memory - %r");
		m->plen = strlen(m->path);
		snprintf(buf, sizeof(buf), "%s%s", n->mount, src+n->plen);
		if ((m->mount = strdup(buf)) == nil)
			sysfatal("no memory - %r");
		m->mlen = strlen(m->mount);;
		m->share = nil;
		m->next = Maproot;
		Maproot = m;
	}
	close9(fid);
	return 0;
}

static void
setwdir(NineP *np, char *path)
{
	int n;
	Map *m;
	Fid *fid;
	char *p, buf[Npath];
	static char opath[Npath] = "";

	for (p = path; *p; p++)
		if (*p == '\\')
			*p = '/';

	if (strcmp(path, opath) == 0)
		return;
	snprintf(opath, sizeof(opath), "%s", path);

	for (m = Maproot; m; m = m->next)
		if (strnicmp(m->mount, path, m->mlen) == 0)
			break;
	if (m == nil)
		return;

	if((fid = open9(np, "/mnt/term/dev/wdir", OWRITE)) != nil){
		n = snprintf(buf, sizeof(buf), "%s%s", m->path, path+m->plen);
		write9(fid, buf, n);
		close9(fid);
	}
}

int
checkprompt(NineP *np, char c)
{
	static int state = None;
	static char *p, buf[Npath];

	switch(state){
	case None:
		p = buf;
		if(c == '\033')
			state++;
		break;
	case Esc:
		if(c == ']')
			state++;
		else
			state = None;
		break;
	case Brack:
		if(c == ';')
			state++;
		else
			state = None;
		break;
	case Semi:
		if(c == '\007')
			state++;
		else
		if (p-buf >= sizeof(buf)-1)
			state = None;
		else
			*p++ = c;
		break;
	case Bell:
		*p++ = 0;
		setwdir(np, buf);
		state = None;
		break;
	}
	return state;
}

void
unmapall(void)
{
	HANDLE h;
	void *buf;
	NETRESOURCE *p;
	unsigned long n, l;
	enum { Nbuf = 16 * Npath };

	if((buf = malloc(Nbuf)) == NULL){
		syswarn("no memory");
		return;
	}
	if(WNetOpenEnum(RESOURCE_CONNECTED, RESOURCETYPE_ANY,
	    RESOURCEUSAGE_ALL, NULL, &h) != NO_ERROR){
		syswarn("cannot enumerate connected network resources");
		return;
	}
	n = -1;
	l = Nbuf;
	while(WNetEnumResource(h, &n, buf, &l) == NO_ERROR){
		for(p = buf; n--; p++){
			if(WNetCancelConnection2(p->lpLocalName, 0, 1) != NO_ERROR)
				syswarn("%s %s - unmap failed", p->lpLocalName, p->lpRemoteName);
		}
		n = -1;
		l = Nbuf;
	}
	WNetCloseEnum(h);
	free(buf);
}

void
mapall(void)
{
	Map *m;
	NETRESOURCE nr;

	for (m = Maproot; m; m = m->next){
		if(! m->share)
			continue;
		memset(&nr, 0, sizeof(NETRESOURCE));
		nr.dwType = RESOURCETYPE_DISK;
		nr.lpLocalName = m->mount;  
		nr.lpRemoteName	= m->share;
		if(WNetAddConnection2(&nr, NULL, NULL, 0) != NO_ERROR)
			syswarn("%s %s - cannot map", m->mount, m->share);
	}
}

