#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "ninep.h"

enum {
	Icons,
	Icpunote,
	Idosmap,
	Iuser,
	Ipid,
	Iwdir,
	Ins
};

NineP *
attach9(HANDLE hi, HANDLE ho, char *mntpt, char *spec)
{
	NineP *np;

	if((np = calloc(sizeof(NineP), 1)) == nil)
		goto fail;
	if((np->mntpt = strdup(mntpt)) == NULL)
		goto fail;
	np->mntlen = strlen(mntpt);
	np->inp = hi;
	np->out = ho;
	if((np->note = CreateEvent(NULL, FALSE, FALSE, NULL)) == nil)
		goto fail;
	return np;
fail:
	free(np);
	return nil;
}

Fid *
open9(NineP *np, char *file, int mode)
{
	Fid *f;
	int id;
	char *buf;

	if(strncmp(np->mntpt, file, np->mntlen) != 0)
		return nil;
	file += np->mntlen;

	buf = nil;
	if (strcmp(file, "/dev/cons") == 0)
		id = Icons;
	else
	if (strcmp(file, "/dev/cpunote") == 0)
		id = Icpunote;
	else
	if (strcmp(file, "/dev/wdir") == 0)
		id = Iwdir;
	else
	if (strcmp(file, "/dev/user") == 0){
		id = Iuser;
		buf = "fred";
	}
	else
	if (strcmp(file, "/usr/fred/lib/dosmap/SNELLWILCOX") == 0){
		id = Idosmap;
		buf = "/n/dfs m: \\\\snellwilcox.local\\root\n";
	}
	else
	if (strcmp(file, "/proc/1234/ns") == 0){
		id = Ins;
		buf = "bind -a /n/dfs/eng/user/fred/work /usr/fred/work\n";
	}
	else
	if (strcmp(file, "/dev/pid") == 0){
		id = Ipid;
		buf = "1234";
	}
	else
		return nil;

	if ((f = malloc(sizeof(Fid))) == NULL)
		return nil;
	np->refs++;
	f->np = np;
	f->name = strdup(file);		// for debug
	f->id = id;
	f->mode = mode;
	f->buf = buf;
	f->off = 0;
	f->len = (buf)? strlen(buf): 0;

	return f;
}

int
interrupt(Fid *f, char *buf, int len)
{
	char *p;
	int i;

	for(p = buf, i = 0; i < len; i++, p++){
		if (*p == 3){				// ctrl-c for now
			SetEvent(f->np->note);
			return 1;
		}
	}
	return 0;
}

int
read9(Fid *f, void *buf, int len)
{
	unsigned long n;

	if(f->mode != OREAD && f->mode != ORDWR)
		return -1;

	if(f->buf){
		if(f->off >= f->len)
			return -1;
		if(len > f->len-f->off)
			len = f->len-f->off;
		memcpy(buf, f->buf+f->off, len);
		f->off += len;
		return len;
	}
	else
	if(f->id == Icons){
		n = -1;
		if(ReadFile(f->np->inp, buf, len, &n, NULL) == 0 || n == -1)
			return -1;
		if(*(signed char *)buf == -1 && n == 1)	// ^D typed -> EOF
			return -1;
if(*(signed char *)buf == 4)		// ^D typed -> EOF
	return -1;
		if(interrupt(f, buf, n))
			return 0;
		f->off += n;
		return n;
	}
	else
	if(f->id == Icpunote){
		WaitForSingleObject(f->np->note, INFINITE);
		return snprintf(buf, len, "interrupted") +1;	// +1 for trailing zero
	}
	return -1;
}

int
line9(Fid *f, char *buf, int len)
{
	char *p;
	int n, rc;

	p = buf;
	for(n = 0; n < len; n++){
		if((rc = read9(f, p, 1)) < 0)
			return (n)? 0: -1;	// EOF
		if(rc == 0 || *p == '\n')
			break;
		*p++;
	}
	*p = 0;
	return 0;
}

int
write9(Fid *f, void *buf, int len)
{
	unsigned long n;

	if(f->mode != OWRITE && f->mode != ORDWR)
		return -1;

	if(f->id == Icons){
		n = 0;
		if(WriteFile(f->np->out, buf, len, &n, NULL) == 0)
			return -1;
		return n;
	}
	if(f->id == Iwdir)
		return 0;
	return -1;
}

int
close9(Fid *f)
{
	f->np->refs--;
	free(f);
	return 0;
}


int 
detach9(NineP *np)
{
	if(np->refs)
		syswarn("ninep: detach with open files\n");
	free(np->mntpt);
	CloseHandle(np->note);
	free(np);
	return 0;
}

