/*
 * fc - floating point reverse polish notation calculator
 * version 2 - rewritten -- wrtp  1/91
 * ansified, swap, rep and dup added wrtp 9/95
 * bugs, comments, etc to roger peppe (rog@ohm.york.ac.uk)
 * converted to plan9 (exit->exits, that's all) 2/2/99
 * added some unicode equivalents for operators 6/99
 * further plan 9 conversion 4/04
 */
#include <u.h>
#include <libc.h>
#include <bio.h>

typedef struct Op Op;
struct Op {
	char *name;
	void *fn;
	short numargs;
	short argtypes;
};
/*#define PI 3.1415926535897932384626433833*/

Op *gettok(char *, double *);

typedef double	tdouble2(double, double),	/* two args */
		tdouble1(double),		/* one arg */
		tdoubleint(double, int),	/* one double, one int */
		tint2(int, int),		/* two ints */
		tint1(int),			/* one int */
		tmult(double *, int);	/* access to all values on stack */

tdouble2 fn_plus, fn_minus, fn_mult, fn_div;
tdouble1 fn_degree, fn_radian, fn_uminus, prnum, fn_log2;
tint2  fn_and, fn_or, fn_xor, fn_shiftr, fn_shiftl, fn_mod;
tint1 fn_not, fn_factorial;
double	fn_trunc(double, int);
tmult fn_rep, fn_swap, fn_sum;

#define T_DOUBLE	0
#define T_INT		1
#define tp2(x, y) ((x)|((y)<<4))

Biobuf bout;
Biobuf berr;

/*
 * list of functions:
 * comment out any lines containing functions not implemented in
 * the host math library
 */
Op ops[] = {
	{"swap",	fn_swap,	-2,	0},
	{"dup",	0,		1,	0},		/* dummy, for usage only */
	{"rep",	fn_rep,		0,	0},
	{"!",		fn_factorial,	1,	0x1},
	{"%",		fn_mod,		2,	0x11},
	{"p",		prnum,		1,	0x0},
	{"*",		fn_mult,	2,	0x00},
	{"×",		fn_mult, 	2,	0x00},
	{"x",		fn_mult,	2,	0x00},
	{"**",		pow,		2,	0x00},
	{"xx",	pow,		2,	0x00},
	{"^",		pow,		2,	0x00},
	{"pow",	pow,		2,	0x00},
	{"+",		fn_plus,	2,	0x00},
	{"-",		fn_minus,	2,	0x00},
	{"/",		fn_div,		2,	0x00},
	{"_",		fn_uminus,	1,	0x0},
	{"<<",		fn_shiftl,	2,	0x11},
	{"«",		fn_shiftl,	2,	0x11},
	{"shl",		fn_shiftl,	2,	0x11},
	{">>",		fn_shiftr,	2,	0x11},
	{"»",		fn_shiftr,	2,	0x11},
	{"shr",		fn_shiftr,	2,	0x11},
	{"and",		fn_and,		2,	0x11},
	{"⋀",			fn_and,		2,	0x11},
	{"⋁",			fn_or,		2, 	0x11},
	{"or",		fn_or,		2,	0x11},
	{"xor",		fn_xor,		2,	0x11},
	{"not",		fn_not,		1,	0x1},
	{"sum",		fn_sum,		0,	0},
	{"acos",	acos,		1,	0x0},
	{"asin",	asin,		1,	0x0},
	{"atan",	atan,		1,	0x0},
	{"atan2",	atan2,		2,	0x00},
	{"ceil",	ceil,		1,	0x0},
	{"cos",		cos,		1,	0x0},
	{"cosh",	cosh,		1,	0x0},
	{"deg",		fn_degree,	1,	0x0},
	{"exp",		exp,		1,	0x0},
	{"fabs",	fabs,		1,	0x0},
	{"floor",	floor,		1,	0x0},
	{"fmod",	fmod,		2,	0x00},
	{"ldexp",	ldexp,		2,	0x10},
	{"log",		log,		1,	0x0},
	{"ln",		log,		1,	0x0},
	{"log10",	log10,		1,	0x0},
	{"log2",	fn_log2,	1,	0x0},
	{"rad",		fn_radian,	1,	0x0},
	{"sin",		sin,		1,	0x0},
	{"sinh",	sinh,		1,	0x0},
	{"√",		sqrt,		1,	0x0},
	{"sqrt",	sqrt,		1,	0x0},
	{"tan",		tan,		1,	0x0},
	{"tanh",	tanh,		1,	0x0},
	{"trunc",	fn_trunc,	2,	0x01}
};

struct {
	char *op;
	double val;
} constants[] = {
	"π", PI,
	"pi",	PI,
	"e",	2.71828182845904509		/* not definitive... */
};

double *stak;
int sp = 0;
Op *last_op;	/* used by rep operation */

char bases[] = "dcxotbB";
char outbase;

void operate(Op *oper);
void printnum(Biobuf *f, double num);
int pbinary(Biobuf *f, ulong num);
void annotatebinary(Biobuf *f, ulong num);
ulong binconv(char *s, char **w);

#define x_abs(n) ((n) < 0 ? -(n) : (n))
void
usage(void)
{
	int i, j, cols, nops;
	Bprint(&berr, "Usage: fc -[%s] <postfix expression>\n", bases);
	Bprint(&berr, "Option specifies output format:\n");
	Bprint(&berr, "		-d decimal\n");
	Bprint(&berr, "		-c rune\n");
	Bprint(&berr, "		-x hex\n");
	Bprint(&berr, "		-o octal\n");
	Bprint(&berr, "		-t time\n");
	Bprint(&berr, "		-b binary\n");
	Bprint(&berr, "		-B annotated binary\n");

	Bprint(&berr, "Operands are decimal(default), hex(0x), octal(0), binary(0b),\n             rune(@), time(hh:mm.ss)\n");
	Bprint(&berr, "Operators are: (number of arguments in brackets)\n");
	nops = sizeof(ops) / sizeof(ops[0]);
	cols = 0;
	for (i = 0; i < nops;) {
		void *fn = ops[i].fn;
		for (j = 0; i < nops && ops[i].fn == fn; i++, j++)
			cols += Bprint(&berr, "%s%s", j?",":"", ops[i].name);
		cols += Bprint(&berr, "[%c] ", "n12"[x_abs(ops[i-1].numargs)]);
		if (cols > 60) {
			Bprint(&berr, "\n");
			cols = 0;
		}
	}
	if (cols > 0)
		Bprint(&berr, "\n");
	Bprint(&berr, "Constants are:\n");
	for (i = 0; i < sizeof(constants) / sizeof(constants[0]); i++) {
		Bprint(&berr, "\t%s=%g\n", constants[i].op, constants[i].val);
	}
	exits("error");
}

void
err(char *s)
{
	Bprint(&berr, "fc: %s\n", s);
	exits("error");
}

void
main(int argc, char **argv)
{
	Op *oper;
	int i;

	Binit(&bout, 1, OWRITE);
	Binit(&berr, 2, OWRITE);

	outbase = bases[0];

	if (argc < 1)
		usage();

	/* allocate more than adequate stack space */
	stak = (double *)mallocz(sizeof(double) * argc, 1);

	if (!stak)
		err("Couldn't allocate stack space");

	if (argc > 1 && *argv[1] && argv[1][0]=='-' && !strchr("0123456789.", argv[1][1])) {
		if (strlen(argv[1]) == 2 && strchr(bases, argv[1][1]))
			outbase = (*++argv)[1];
		else
			usage();
	}
	/* despite the following, it still dies on 0/0 */
	setfcr(getfcr() & ~(FPINEX | FPOVFL | FPUNFL | FPZDIV));

	while (*argv && *++argv) {
		if (oper = gettok(*argv, &stak[sp]))
			operate(oper);
		else
			sp++;
		last_op = oper;
	}

	/* print stack bottom first */
	for (i = 0; i < sp; i++)
		printnum(&bout, stak[i]);
	exits(nil);
}

double
prnum(double sofar)
{
	printnum(&berr, sofar);
	return sofar;
}

void
printnum(Biobuf *f, double sofar)
{
	char x[UTFmax];
	Rune r;
	int n;
	switch (outbase) {
	case 'c':
		r = sofar;
		n = runetochar(x, &r);
		Bprint(f, "@%.*s\n", n, x);			/* should use print() really */
		break;
	case 'd':
		Bprint(f, "%.10g\n", sofar);
		break;
	case 'b':
		pbinary(f, (ulong)sofar);
		Bprint(f, "\n");
		break;
	case 'B':
		annotatebinary(f, (ulong)sofar);
		break;
	case 'o':
		Bprint(f, "0%lo\n", (ulong)sofar);
		break;
	case 'x':
		Bprint(f, "0x%lx\n", (ulong)sofar);
		break;
	case 't':
		Bprint(f, "%.2d:", (int)sofar);
		sofar = (sofar - (int)sofar) * 60;	/* get minutes */
		Bprint(f, "%.2d", (int)sofar);
		sofar = (sofar - (int)sofar) * 60;	/* get seconds */
		if ((int)(sofar * 10000) != 0)		/* worthwhile seconds */
			Bprint(f, ".%g", sofar);
		Bprint(f, "\n");
		break;
	default :
		Bprint(&berr, "fc: Unsupported format -%c\n", outbase);
		exits("error");
	}
}

/*
 * print n as a binary number, always printing
 * a multiple of 8 binrary digits.
 * it returns the number of digits printed
 */
int
pbinary(Biobuf *f, ulong n)
{
	int i, j;
	Bprint(f, "0b");
	for (j = 3; j > 0; j--)
		if (n & (0xff << (j * 8)))
			break;
	for (i = 31; i >= 0; i--)
		if (i / 8 <= j)
			Bputc(f, ((n >> i) & 1) + '0');
	return (j + 1) * 8;
}

void
annotatebinary(Biobuf *f, ulong n)
{
	int i, ndig;
	ndig = pbinary(f, n);
	Bprint(f, "\n  ");
	for (i = ndig - 1; i >= 0; i--)
		Bprint(f, "%d", i % 10);
	if (ndig < 10) {
		Bprint(f, "\n");
		return;
	}
	Bprint(f, "\n  ");
	for (i = ndig - 1; i >= 10; i--) {
		if (i % 10 == 0)
			Bprint(f, "%d", i / 10);
		else
			Bprint(f, " ");
	}
	Bprint(f, "\n");
}
  
void
operate(Op *oper)
{
	int numargs = oper->numargs;
	double ret;
	void *fn = oper->fn;

	if (numargs < 0)
		numargs = -numargs;

	if (sp < numargs) {
		Bprint(&berr, "fc: Stack too small for op '%s'\n", oper->name);
		exits("error");
	}

	switch (oper->numargs) {
	default:
		if (oper->numargs < 0)
			(*(tmult*)fn)(stak, sp);
		else
			err("n>2 type not supported.");
		break;
	case 0:	/* cumulative... */
		ret = (*(tmult*)fn)(stak, sp);
		break;
	case 1:
		switch (oper->argtypes) {
		case T_DOUBLE :	ret = (*(tdouble1*)fn)(stak[sp-1]);	break;
		case T_INT :	ret = (*(tint1*)fn)(stak[sp-1]);		break;
		default :	Bprint(&berr, "1 type %x not supported.\n", oper->argtypes);
				exits("error");
		}
		break;
	case 2:
		switch (oper->argtypes) {
		case tp2(T_DOUBLE, T_DOUBLE):
			ret = (*(tdouble2*)fn)(stak[sp-2], stak[sp-1]);		break;
		case tp2(T_DOUBLE, T_INT):
			ret = (*(tdoubleint*)fn)(stak[sp-2], (int)stak[sp-1]);	break;
		case tp2(T_INT, T_INT):
			ret = (*(tint2*)fn)((int)stak[sp-2], (int)stak[sp-1]);	break;
		default :
			Bprint(&berr, "2 type %x not supported.\n", oper->argtypes);
			exits("error");
		}
		break;
	}
	if (oper->numargs >= 0) {
		if (oper->numargs)
			sp -= oper->numargs - 1;
		else if (oper->numargs == 0)
			sp = 1;
		stak[sp-1] = ret;
	}
}

Op *
gettok(char *s, double *num)
{
	char *whereto, *t;
	int i;
	if (!strcmp(s, "dup")) {
		*num = (num > stak) ? num[-1] : 0;
		return 0;
	}
	for (i = 0; i < nelem(constants); i++){
		if (!strcmp(s, constants[i].op)) {
			*num = constants[i].val;
			return 0;
		}
	}
	if (*s == '@') {
		Rune r;
		chartorune(&r, ++s);
		*num = r;
		return 0;
	}
	if ((*s=='-' && s[1]) || strchr("0123456789.", *s)) {
		/* if it's a time value */
		if ((t=strchr(s, ':')) != 0) {
			char *t2;
			*num = 0;
			*t++ = '\0';
			/* if it's got seconds as well... */
			if (t2 = strchr(t, '.')) {
				*t2++ = '\0';
				*num += (double)atoi(t2) / (60.0 * 60.0);
			}
			*num += (double)atoi(t) / 60.0 + atof(s);
		} else {
			/* if it's hex or octal */
			if (*s=='0' && !strpbrk(s, ".eE") || *s=='0' && s[1] == 'x') {
				*num = s[1]=='b'
					? (double)binconv(s + 2, &whereto)
					: (double)(long)strtoul(s, &whereto, 0);
			} else
				*num = strtod(s, &whereto);
			if (!whereto || *whereto) {
				Bprint(&berr, "Invalid number '%s'\n", s);
				exits("error");
			}
		}
		return 0;
	}
	for (i=0; i<nelem(ops); i++)
		if (!strcmp(ops[i].name, s))
			break;
	if (i==nelem(ops)) {
		Bprint(&berr, "Unknown operator '%s'\n", s);
		exits("error");
	}
	return &ops[i];
}

/* convert s from binary, setting w to point to first non-binary digit */
ulong
binconv(char *s, char **w)
{
	ulong sum = 0;
	if ('0' <= *s && *s <= '1') {
		for (; *s && '0' <= *s && *s <= '1'; s++)
			sum = (sum << 1) + (*s - '0');
		*w = s;
	}
	else *w = 0;
	return sum;
}

double
fn_swap(double st[], int n)
{
	if (n >= 2) {
		double tmp = st[n - 1];
		st[n - 1] = st[n - 2];
		st[n - 2] = tmp;
	}
	return 0;
}

/* repeat last operator until not enough elements left on stack */
double
fn_rep(double st[], int n)
{
	Op *rep_op = last_op;
	USED(n);
	if (rep_op == 0)
		err("No operator to rep");

	if (rep_op->numargs != 2)
		err("Invalid operator for rep");
	while (sp > 1)
		operate(rep_op);
	return st[0];
}

/* sum all elements on stack - same as + rep */
double
fn_sum(double st[], int n)
{
	double ret = 0;
	while (n--)
		ret += st[n];
	return ret;
}

double
fn_mod(int x, int y){return x % y;}
double
fn_plus(double x, double y){return x + y;}
double
fn_minus(double x, double y){return x - y;}
double
fn_mult(double x, double y){return x * y;}
double
fn_div(double x, double y){return x / y;}
double
fn_uminus(double x){return -x;}
double
fn_factorial(int n)
{
	double ret=1;
	while (n > 0) ret*=n--;
	return ret;
}
double
fn_degree(double x){return x / (2*PI) * 360.0;}
double
fn_radian(double x){return (x / 360.0) * 2 * PI;}
double
fn_and(int x,int y)	{return x & y;}
double
fn_or(int x,int y)	{return x | y;}
double
fn_xor(int x,int y)	{return x ^ y;}
double
fn_not(int x)		{return ~x;}
double
fn_shiftr(int x, int y){return x >> (unsigned)y;}
double
fn_shiftl(int x, int y){return x << (unsigned)y;}
double
fn_trunc(double x, int y)
{
	double dig = pow(10.0, (double)y);
	return floor(x * dig) / dig;
}
double
fn_log2(double x){return log(x) / log(2.0);}
