#include <winsock2.h>
#include <process.h>
#include <time.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "listen.h"

typedef struct {
	Ann *an;
	long raddr;				 // remote address
	int rport;				 // remote port
	SOCKET net;				 // network
	int pid;				 // cmd's process ID
} Call;

enum {
	MAX_MSG = 8200				 // fits in a 9p buffer
};

static char *
ipaddr(unsigned long ip)
{
	static char buf[20];

	snprintf(buf, sizeof(buf), "%lu.%lu.%lu.%lu",
		(ip >> 24) & 0xff, (ip >> 16) & 0xff,
		(ip >> 8) & 0xff, (ip >> 0) & 0xff);
	return buf;
}

void
reap(int proc, HANDLE hand)
{
	unsigned long s;

	if(GetExitCodeProcess(hand, &s) && s != STILL_ACTIVE)
		return;
	GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, proc);
	Sleep(100);
	if(GetExitCodeProcess(hand, &s) && s != STILL_ACTIVE)
		return;
	GenerateConsoleCtrlEvent(CTRL_C_EVENT, proc);
	Sleep(100);
	if(GetExitCodeProcess(hand, &s) && s != STILL_ACTIVE)
		return;
	TerminateProcess(hand, 0);
}

static unsigned long WINAPI
newcall(void *v)
{
	HANDLE ha[3];
	STARTUPINFO si;
	char cmd[1024];
	Call *c = (Call *)v;
	SECURITY_ATTRIBUTES sa;
	PROCESS_INFORMATION pi;

	memset(&sa, 0, sizeof(sa));
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;

	if(SetHandleInformation((HANDLE)c->net, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT) == 0)
		syslog("cannot set socket handle to inherit mode");
	if((ha[0] = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL)
		syslog("cannot create event");
	if(WSAEventSelect(c->net, ha[0], FD_CLOSE) != 0)
		syslog("cannot select on event");

	memset(&si, 0, sizeof(si));
	si.cb = sizeof(STARTUPINFO);
	si.wShowWindow = SW_HIDE;
	si.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
	si.hStdInput = (HANDLE)c->net;
	si.hStdOutput = (HANDLE)c->net;
	si.hStdError = (HANDLE)c->net;

	memset(&pi, 0, sizeof(pi));
	snprintf(cmd, sizeof(cmd), "%s %s %d %s %d",
		c->an->path, c->an->proto, c->an->port,
		ipaddr(c->raddr), c->rport);
	syslog("connect: %s", cmd);

	CreateProcess(NULL, cmd, NULL, NULL, TRUE, 0,
		NULL, Ctldir, &si, &pi);
	ha[1] = pi.hProcess;

	ha[2] = c->an->stop;
	WaitForMultipleObjects(3, ha, FALSE, INFINITE);
	
	debug("newcall: port=%d child exited / socket closed / service shutdown\n", c->an->port);

	CloseHandle(ha[0]);
	reap(pi.dwProcessId, pi.hProcess);

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
	shutdown(c->net, SD_BOTH);
	closesocket(c->net);

	free(c);
	return 0;
}

unsigned long WINAPI
doaccept(void *v)
{
	int n;
	Call *c;
	HANDLE rly;
	struct linger li; 
	Ann *an = (Ann *)v;
	struct sockaddr_in peer;

	if ((c = malloc(sizeof(Call))) == NULL){
		syslog("no memory");
		return -1;
	}
	c->an = an;

	debug("doaccept: port=%d waiting for call\n", an->port);
	n = sizeof(peer);
	if((c->net = accept(an->sock, (struct sockaddr *)&peer, &n)) == -1){
		syslog("doaccept: port=%d accept failed", an->port);
		return 1;
	}
	debug("\ndoaccept: port=%d connected\n", an->port);

	c->raddr = ntohl(peer.sin_addr.s_addr);
	c->rport = ntohs(peer.sin_port);

	n = 1;
	setsockopt(c->net, SOL_SOCKET, SO_KEEPALIVE, (char*)&n, sizeof(n));
	n = MAX_MSG;
	setsockopt(c->net, SOL_SOCKET, SO_SNDBUF, (char*)&n, sizeof(n));
	n = MAX_MSG;
	setsockopt(c->net, SOL_SOCKET, SO_RCVBUF, (char*)&n, sizeof(n));
	n = 1;
	setsockopt(c->net, SOL_SOCKET, SO_REUSEADDR, (char*)&n, sizeof(n));

   	if (strcmp(an->proto, "tcp") == 0){
   		li.l_onoff = 0;
    		li.l_linger = 0; 
		setsockopt(c->net, IPPROTO_TCP, SO_LINGER, (char *)&li, sizeof(li));
		n = 1;
		setsockopt(c->net, IPPROTO_TCP, TCP_NODELAY, (char*)&n, sizeof(n));
	}

	rly = CreateThread(NULL, 0, newcall, c, 0, NULL);
	CloseHandle(rly);
	CloseHandle(an->srv);
	an->srv = NULL;
	return 0;
}

int
dolisten(Ann *an)
{
	SOCKET sock;
	int proto, type;
	struct sockaddr_in addr;

	if (strcmp(an->proto, "tcp") == 0){
		type = SOCK_STREAM;
		proto = IPPROTO_TCP;
	}
	else
	if (strcmp(an->proto, "udp") == 0){
		type = SOCK_DGRAM;
		proto = IPPROTO_UDP;
	}
	else{
		syslog("%s - protocol not supported", an->proto);
		Sleep(5000);
		return -1;
	}

	/*
	 * we must use WSAsocket() rather than socket as the handles returned
	 * by WSAsocket can be used as the stdhandles for the child process
	 * whilst the handles returned by socket() cannot
	 */
	syslog("dolisten: port=%d proto=%s\n", an->port, an->proto);
	if((sock = WSASocket(PF_INET, type, proto, NULL, 0, 0)) == INVALID_SOCKET){
		syslog("socket failed");
		Sleep(5000);
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(an->port);
	addr.sin_addr.s_addr = INADDR_ANY;

	if(bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1){
		syslog("dolisten: addr=%s port=%d bind failed",
			ipaddr(ntohl(addr.sin_addr.s_addr)), ntohs(addr.sin_port));
		shutdown(sock, SD_BOTH);
		closesocket(sock);
		Sleep(5000);
		return -1;
	}

	if(listen(sock, SOMAXCONN) == -1){
		syslog("dolisten: listen failed");
		shutdown(sock, SD_BOTH);
		closesocket(sock);
		Sleep(5000);
		return -1;
	}
	return sock;
}


void
ignore(Ann *an)
{
	if(an->srv != NULL){
		TerminateThread(an->srv, 0);
		CloseHandle(an->srv);
		an->srv = NULL;
	}

	Sleep(100);
	if(an->stop != NULL){
		SetEvent(an->stop);
		Sleep(100);		// allow it to tidy itself up
		CloseHandle(an->stop);
		an->stop = NULL;
	}

/*
 * for some reason when I try to shutdown the
 * socket closesocket() and shutdown() hang,
 * with so_linger. 
 *
 *	if(an->sock != -1){
 *		shutdown(an->sock, SD_BOTH);
 *		closesocket(an->sock);
 *		an->sock = -1;
 *	}
 */
}
