#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <time.h>
#include "listen.h"

int Debug = 0;			// debug level, bigger value -> chattier
char *argv0 = "?";		// my name
Ann *Announce = NULL;		// ports we should announce on
HANDLE Finished = 0;		// flag this event on service shutdown
int Trusted = 0;
char *Logfile;
char *Ctldir;
char *Proto = "unknown";

int
addannounce(char *path, int port, char *proto)
{
	Ann *an;
	HANDLE stop;

	for (an = Announce; an; an = an->next){
		if(an->port == port && strcmp(an->proto, proto) == 0){
			if(an->state == Sunwanted)
				an->state = Sactive;
			else
			if(strcmp(path, an->path) != 0)
				syslog("addannounce: %s %s duplicate services", an->path, path);
			return 0;
		}
	}

	if((stop = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL){
		syslog("addannounce: cannot create event");
		return 1;
	}
	if((an = calloc(sizeof(Ann), 1)) == NULL){
		syslog("addannounce: no memory for Ann");
		return -1;
	}

	debug("addannounce: port=%d proto=%s path=%s\n", port, proto, path);
	an->stop = stop;
	an->proto = proto;
	an->path = strdup(path);
	an->port = port;
	an->sock = -1;
	an->state = Sactive;
	an->next = Announce;
	Announce = an;

	return 0;
}

int
scandir(char *ctldir, char *proto)
{
	DIR *d;
	Ann *an;
	int len;
	struct dirent *de;
	char *p, path[1024];

	for (p = ctldir; *p; p++)
		if (*p == '\\')
			*p = '/';

	for(an = Announce; an; an = an->next)
		if(an->state == Sactive)
			an->state = Sunwanted;

	if((d = opendir(ctldir)) == NULL){
		syslog("scandir: %s cannot open dir", ctldir);
		return -1;
	}

	len = strlen(proto);
	while((de = readdir(d)) != NULL){
		if(strncmp(de->d_name, proto, len) != 0)
			continue;
		strcpy(path, ctldir);
		if(path[strlen(path)-1] != '/')
			strcat(path, "/");
		strcat(path, de->d_name);
		addannounce(path, atoi(de->d_name+len), proto);
	}
	closedir(d);

	/*
	 * This one is no longer needed, ignore 
	 * this port.
	 */
	for(an = Announce; an; an = an->next)
		if (an->state == Sunwanted){
			debug("ignore: port=%d\n", an->port);
			ignore(an);
		}

	return 0;
}

/*
 * Await work to do or a 60 sec timeout
 */
int
awaitwork(HANDLE dir)
{
	Ann *an;
	int n, which;
	static HANDLE *h;
	static int alloc = 0;
	enum { Fixed = 2 };

	n = Fixed;
	for(an = Announce; an; an = an->next)
		n++;
	if (n >= alloc){
		alloc += 8;
		if ((h = malloc(sizeof(HANDLE)*n)) == NULL){
			syslog("no memory");
			exit(1);
		}
	}

	n = Fixed;
	for(an = Announce; an; an = an->next)
		if(an->sock != -1 && an->srv != NULL)
			h[n++] = an->srv;

	h[0] = dir;
	h[1] = Finished;
	debug("awaitwork: sleeping on %d sockets\n", n-1);
	which = WaitForMultipleObjects(n, h, FALSE, 60*1000);
	if(which == WAIT_OBJECT_0){
		debug("awaitwork: directory change\n");
		FindNextChangeNotification(dir);
	}
	if(which == WAIT_OBJECT_0+1){
		debug("awaitwork: shutdown listner\n");
		return -1;
	}
	debug("awaitwork: listener exited - respawning\n");
	return 0;
}

int
listendir(char *ctldir, char *proto)
{
	Ann *an;
	HANDLE dir;
	unsigned long flags;

	flags = FILE_NOTIFY_CHANGE_FILE_NAME;
	flags |= FILE_NOTIFY_CHANGE_LAST_WRITE;
	if((dir = FindFirstChangeNotification(ctldir, FALSE, flags)) == NULL){
		syslog("listendir: %s - cannot watch changes", ctldir);
		return -1;
	}

	if((Finished = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL){
		syslog("listendir: %s - create event", ctldir);
		return -1;
	}

	syslog("listendir: watching %s", ctldir);

	do{
		if (scandir(ctldir, proto) == -1)
			return -1;

		for(an = Announce; an; an = an->next){
			if(an->state != Sactive)
				continue;
			if(an->sock == -1 && an->srv == NULL)
				an->sock = dolisten(an);
			if(an->sock != -1 && an->srv == NULL)
				an->srv = CreateThread(NULL, 0, doaccept, an, 0, NULL);
		}
	}while(awaitwork(dir) != -1);

	for(an = Announce; an; an = an->next)
		if(an->state == Sactive)
			ignore(an);

	FindCloseChangeNotification(dir);
	Sleep(1000);
	CloseHandle(Finished);
	return 0;
}

void
usage(char *s)
{
	FILE *fp;

	if((fp = fopen(Logfile, "a")) != NULL){
		fprintf(fp, "usage error: %s\n", (s)? s: "NULL");
		fclose(fp);
	}

	fprintf(stderr, "%s\n", (s)? s: "NULL");
	fprintf(stderr, "usage: %s [-d dir] [-iuhrsD] proto\n", argv0);
	fprintf(stderr, "-l file   set logfile to <file>\n");
	fprintf(stderr, "-d dir    look in <dir> for programs to envoke\n");
	fprintf(stderr, "-i        install service\n");
	fprintf(stderr, "-u        uninstall service\n");
	fprintf(stderr, "-r        run service\n");
	fprintf(stderr, "-h        halt service\n");
	fprintf(stderr, "-s        start serving requests\n");
	fprintf(stderr, "-D        debug - don't run as a service\n");
	WSACleanup();
	exit(1);
}

int
main(int argc, char *argv[])
{
	WSADATA wsadata;
	int o, rc, op;
	char name[64];
		
	if(WSAStartup(MAKEWORD(2, 2), &wsadata) != 0){
		syslog("WSAStartup failed");
		return 0;
	}

	stamp("main", argc, argv);

	rc = -1;
	op = 's';
	optind = 0;
	optarg = NULL;
	argv0 = argv[0];
	while((o = getopt(argc, argv, "l:srhiDud:t")) != -1)
		switch (o){
		case 'l':
			Logfile = optarg;
			break;
		case 't':
			Trusted = 1;
			break;
		case 'r':
		case 'i':
		case 'u':
		case 'h':
		case 's':
			op = o;
			break;
		case 'd':
			Ctldir = optarg;
			break;
		case 'D':
			Debug++;
			break;
		case 'x':
			Debug++;
			op = o;
			break;
		default:
			usage("main: unknown option");
			break;
		}
	if (argc-optind != 1)
		usage("main: more than one arg after options");
	Proto = argv[optind];

	initenv();

	snprintf(name, sizeof(name), "listen-%s%s", Proto, (Trusted)? "-t": "");

	switch(op){
	case 'i':
		rc = svs_inst(name, "9win network listener");
		break;
	case 'u':
		rc = svs_uninst(name);
		break;
	case 'r':
		rc = svs_start(name);
		break;
	case 'h':
		rc = svs_stop(name);
		break;
	case 's':
		rc = svs_serv(name);		// service should block here.
		break;
	case 'x':				// Xperimental
		rc = listendir(Ctldir, Proto);
		break;
	default:
		break;
	}
	WSACleanup();
	exit((rc == 0)? 0: 42);
}
