#include <windows.h>
#include <winsvc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "listen.h"

/*
 * You would think that the argc, argv passed to StartService would be
 * then passed back to your service callback routine. This is almost true,
 * the SCM prepends the service name to the start of the argv array and
 * increments argc accordingly; However if the service is started from
 * the SCM gui or from auto boot these doesn't seem to happen! The program
 * is envoked correctly using the path and args passed when it was installed
 * my code then calls StartService() but the callback's args consist of only
 * the service's name and argc == 1. 
 *
 * This is why I pass Proto, Trusted, and Ctldir in global variables (ugh).
 */
static SERVICE_STATUS State;
static SERVICE_STATUS_HANDLE Hnd;

static void
svs_wstat(int new, int accept, int msec)
{
	debug("svs_wstat: state=%d accept=%d hint=%d\n", new, accept, msec);

	memset(&State, 0, sizeof(State));
	State.dwWaitHint = msec;
	State.dwCurrentState = new;
	State.dwWin32ExitCode = NO_ERROR;
	State.dwControlsAccepted = accept;
	State.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	SetServiceStatus(Hnd, &State);
}

static void WINAPI
cb_svs_state(DWORD req)
{
	debug("cb_svs_state: %ud\n", req);
	switch(req){
	case SERVICE_CONTROL_STOP:
		svs_wstat(SERVICE_STOP_PENDING, 0, 2000);
		SetEvent(Finished);
		return;
	case SERVICE_CONTROL_INTERROGATE:
		SetServiceStatus(Hnd, &State);
		break;
	default:
		syslog("cb_svs_state: unexpected state req=%d\n", req);
		break;
	}
}

static VOID WINAPI
cb_svs_main(int argc, char *argv[])
{
	char path[1024];

	GetModuleFileName(NULL, path, sizeof(path));
	Hnd = RegisterServiceCtrlHandler(path, cb_svs_state);
	svs_wstat(SERVICE_START_PENDING, 0, 2000);

	if(! Trusted)
		becomenone();

	svs_wstat(SERVICE_RUNNING, SERVICE_ACCEPT_STOP, 0);
	listendir(Ctldir, Proto);
	svs_wstat(SERVICE_STOPPED, 0, 0);
}

int
svs_inst(char *name, char *desc)
{
	SC_HANDLE scm, svs;
	SERVICE_DESCRIPTION sd;
	char depend[64], path[1024], cmd[1024];

	if((scm = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS )) == 0){
		syslog("svs_inst: cannot open SCM manager");
		return -1;
	}

	GetModuleFileName(NULL, path, sizeof(path));
	snprintf(cmd, sizeof(cmd), "%s %s %s -s -d %s -l %s %s",
		enquote(path, sizeof(path)),
		(Trusted)? "-t": "",
		(Debug)? "-D": "",
		enquote(Ctldir, sizeof(Ctldir)),
		enquote(Logfile, sizeof(Logfile)),
		Proto);

	memset(depend, 0, sizeof(depend));
	snprintf(depend, sizeof(depend), "%c%s", SC_GROUP_IDENTIFIER, "TDI");
	if ((svs = CreateService(scm,
				name,		// service name
				name,		// display name 
				SERVICE_ALL_ACCESS,
				SERVICE_WIN32_OWN_PROCESS|SERVICE_INTERACTIVE_PROCESS,
				SERVICE_AUTO_START,
				SERVICE_ERROR_NORMAL,
				cmd,
				NULL,		// load order group
				NULL,		// tag in load order 
				depend,		// load order dependency
				NULL,		// username
				NULL		// password
				)) == NULL){
		syslog("%s - cannot create service", name);
		CloseServiceHandle(scm);
		return -1;
	}

	if (desc && *desc){
		sd.lpDescription = desc;
		ChangeServiceConfig2(svs, SERVICE_CONFIG_DESCRIPTION, &sd);
	}

	CloseServiceHandle(svs);
	CloseServiceHandle(scm);
	return 0;
}

int
svs_uninst(char *name)
{
	SC_HANDLE scm, svs;

	if ((scm = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS)) == 0){
		syslog("cannot open SCM manager");
		return -1;
	}
	if ((svs = OpenService(scm, name, SERVICE_ALL_ACCESS)) == 0){
		syslog("%s - cannot open service", name);
		CloseServiceHandle(scm);
		return -1;
	}
	
	if (DeleteService(svs) == 0){
		syslog("%s - cannot delete service", name);
		CloseServiceHandle(svs);
		CloseServiceHandle(scm);
		return -1;
	}
	CloseServiceHandle(svs);
	CloseServiceHandle(scm);
	return 0;
}

int
svs_serv(char *name)
{
	static char sname[64];
	static SERVICE_TABLE_ENTRY Svstab[] = {
		{ sname, (void *)cb_svs_main },
		{ NULL, NULL }
	};

	snprintf(sname, sizeof(sname), "%s", name);

	/* This function blocks until the service is stopped */
	if (StartServiceCtrlDispatcher(Svstab) == 0){
		if(GetLastError() == 1063){
			syslog("%s: %s -s only valid from SCM", argv0, name);
		}else
			syslog("%s StartServiceCtrlDispatcher", name);
		return -1;
	}

	return 0;
}

int
svs_start(char *name)
{
	SC_HANDLE scm, svs;

	if ((scm = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS)) == 0){
		syslog("cannot open SCM manager");
		return -1;
	}
	
	if ((svs = OpenService(scm, name, SERVICE_ALL_ACCESS)) == 0){
		syslog("%s - cannot open service", name);
		CloseServiceHandle(scm);
		return -1;
	}
	if (StartService(svs, 0, NULL) == 0){
		syslog("%s - cannot start service", name);
		CloseServiceHandle(svs);
		CloseServiceHandle(scm);
		return -1;
	}
	CloseServiceHandle(svs);
	CloseServiceHandle(scm);
	return 0;
}

int
svs_stop(char *name)
{
	SERVICE_STATUS ss;
	SC_HANDLE scm, svs;

	if ((scm = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS)) == 0){
		syslog("cannot open SCM manager");
		return -1;
	}
	
	if ((svs = OpenService(scm, name, SERVICE_ALL_ACCESS)) == 0){
		syslog("%s - cannot open service", name);
		CloseServiceHandle(scm);
		return -1;
	}

	if (ControlService(svs, SERVICE_CONTROL_STOP, &ss) == 0){
		syslog("%s - cannot stop service", name);
		CloseServiceHandle(svs);
		CloseServiceHandle(scm);
		return -1;
	}

	CloseServiceHandle(svs);
	CloseServiceHandle(scm);
	return 0;
}
