#include <windows.h>
#include <stdio.h>
#include <time.h>
#include "listen.h"

char *
enquote(char *s, int n)
{
	int l;

	if (strpbrk(s, " \t") == 0)
		return s;
	l = strlen(s);
	if(l < n-3){
		memmove(s, s+1, l);
		s[0] = '"';
		s[l+2] = '"';
		s[l+3] = '\0';
	}
	return s;
}

static void
errmsg(FILE *fp, char *pre, char *fmt, va_list args, char *post)
{
	fprintf(fp, "%s: ", pre);
	vfprintf(fp, fmt, args);
	if(*post)
		fprintf(fp, " - %s", post);
	fprintf(fp, "\n");
	fflush(fp);
}

void
syslog(char *fmt, ...)
{
	int e;
	FILE *fp;
	time_t now;
	va_list args;
	char *p, *tstr, msg[1024];

	*msg = 0;
	if((e = GetLastError()) != 0){
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL,
			e, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			msg, sizeof(msg), 0);
		if ((p = strpbrk(msg, ".\n\r")) != NULL)
			*p = 0;
		SetLastError(0);
	}

	va_start(args, fmt);
	errmsg(stdout, argv0, fmt, args, msg);
	va_end(args);

	if((fp = fopen(Logfile, "a")) == NULL)
		return;
	time(&now);
	tstr = ctime(&now);
	if ((p = strpbrk(tstr, "\n\r")) != NULL)
		*p = 0;

	va_start(args, fmt);
	errmsg(fp, tstr, fmt, args, msg);
	va_end(args);
	fclose(fp);

}

void
debug(char *fmt, ...)
{
	FILE *fp;
	time_t now;
	va_list args;
	char *p, *tstr;

	if (! Debug)
		return;
	if((fp = fopen(Logfile, "a")) == NULL)
		return;

	time(&now);
	tstr = ctime(&now);
	if ((p = strpbrk(tstr, "\n\r")) != NULL)
		*p = 0;
	fprintf(fp, "%s: ", tstr);

	va_start(args, fmt);
	vfprintf(fp, fmt, args);
	va_end(args);
	fclose(fp);
}

void
stamp(char *str, int argc, char *argv[])
{
	int i;
	FILE *fp;
	time_t now;
	char *p, *preamble;

	if((fp = fopen(Logfile, "a")) == NULL)
		return;

	time(&now);
	preamble = ctime(&now);
	if ((p = strpbrk(preamble, "\r.\n")) != NULL)
		*p = 0;
	fprintf(fp, "%s: [%s] ", preamble, str);
	for(i = 0; i < argc; i++)
		fprintf(fp, "%d='%s' ", i, argv[i]);
	fprintf(fp, "\n");
	fclose(fp);
}

