#include <u.h>
#include <libc.h>
#include <mswin.h>

#include <userenv.h>
#include <lm.h>

void
local(wchar_t *old, wchar_t *new, int perm)
{
	wchar_t guidpath[MAX_PATH];

	if(perm){
		if(GetVolumeNameForVolumeMountPoint(new, guidpath, nelem(guidpath)) == 0)
			sysfatal("GetVolumeNameForVolumeMountPoint() failed, %s", _winerr());
		if(SetVolumeMountPoint(old, guidpath) == 0)
			sysfatal("SetVolumeMountPoint() failed, %s", _winerr());
	}
	else{
		if(DefineDosDevice(DDD_RAW_TARGET_PATH, new, old) == 0)
			sysfatal("DefineDosDevice() failed, %s", _winerr());
	}
}

void
network(wchar_t *old, wchar_t *new, int perm, wchar_t *user, wchar_t *pass)
{
	int e, flags;
	NETRESOURCE nr;

	flags = 0;
	if(perm)
		flags = CONNECT_UPDATE_PROFILE;

	memset(&nr, 0, sizeof(NETRESOURCE));
	nr.dwType = RESOURCETYPE_DISK;				// or RESOURCETYPE_PRINT
	nr.lpLocalName = old;  
	nr.lpRemoteName	= new;

	if((e = WNetAddConnection2(&nr, pass, user, flags)) != NO_ERROR)
		sysfatal("WNetAddConnection2() failed, %s", _wneterr(e));
}

void
usage(void)
{
	fprint(2, "usage: %s [-r] [-u user] [-p passwd] old new\n", argv0);
	fprint(2, "	-r	    make change permanant in the registry\n");
	fprint(2, "	-u user	remote username (only network mounts)\n");
	fprint(2, "	-p pass	remote password (only network mounts)\n");
	fprint(2, "	         missing username infers use current username,\n");
	fprint(2, "	         an empty username infers the user 'none'\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int perm;
	wchar_t *old, *new, *user, *pass;

	perm = 0;
	user = nil;
	pass = nil;
	ARGBEGIN{
	case 'r':
		perm = 1;
		break;
	case 'u':
		user = _winstr(EARGF(usage()));
		break;
	case 'p':
		pass = _winstr(EARGF(usage()));
		break;
	default:
		usage();
		break;
	}ARGEND;

	if(argc != 2)
		usage();

	old = _winpath(argv[0]);
	new = _winpath(argv[1]);

	if(strncmp(argv[0], "\\\\", 2) == 0 || strncmp(argv[0], "//", 2) == 0)
		network(old, new, perm, user, pass);
	else
		local(old, new, perm);

	free(old);
	free(new);
	free(user);
	free(pass);
	exits(nil);
}
