#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

#include <Sddl.h>
#include <Aclapi.h>

/* Why is this needed? */
#define FILE_GENERIC_ALL 0x10000000 

void
dumpace(Biobuf *bp, void *ace, int attr)
{
	SID *sid;
	int mask, type, flags;
	char *oper, *name;
	ACCESS_ALLOWED_ACE *allow;
	ACCESS_DENIED_ACE *deny;

	deny = (ACCESS_DENIED_ACE *)ace;
	allow = (ACCESS_ALLOWED_ACE *)ace;

	type = allow->Header.AceType;
	flags = allow->Header.AceFlags;

	switch(type){
	case ACCESS_ALLOWED_ACE_TYPE:
		oper = "allow";
		mask = allow->Mask;
		sid = (SID *)&allow->SidStart;
		break;
	case ACCESS_DENIED_ACE_TYPE:
		oper = "deny";
		mask = deny->Mask;
		sid = (SID *)&deny->SidStart;
		break;
	default:
		fprint(2, "%s: type=%d ACE neither is Allow not Deny\n", type, argv0);
		return;
	}


	if((name = _sid2name(sid, 0)) != nil)
		print("%s\n", name);
	free(name);

	if(attr & FILE_ATTRIBUTE_READONLY)
		print("	readonly attr\n");

	if(flags & OBJECT_INHERIT_ACE)
		print("	flag: object inherit\n");
	if(flags & CONTAINER_INHERIT_ACE)
		print("	flag: container inherit\n");
	if(flags & INHERIT_ONLY_ACE)
		print("	flag: inherit only ACE\n");
	if(flags & NO_PROPAGATE_INHERIT_ACE)
		print("	flag: no propogate inherit ACE\n");

	if((mask & FILE_ALL_ACCESS) == FILE_ALL_ACCESS){
		print("	%s all\n", oper);
		return;
	}
	if((mask & FILE_GENERIC_ALL) == FILE_GENERIC_ALL){
		print("	%s generic all\n", oper);
		return;
	}
	if((mask & FILE_GENERIC_READ) == FILE_GENERIC_READ){
		print("	%s generic read\n", oper);
		return;
	}
	if((mask & FILE_GENERIC_WRITE) == FILE_GENERIC_WRITE){
		print("	%s generic write\n", oper);
		return;
	}
	if((mask & FILE_GENERIC_EXECUTE) == FILE_GENERIC_EXECUTE){
		print("	%s generic execute\n", oper);
		return;
	}

	if(attr & FILE_ATTRIBUTE_DIRECTORY){
		if(mask & FILE_LIST_DIRECTORY)		print("	%s dir scan\n", oper);
		if(mask & FILE_ADD_FILE)			print("	%s add file to dir\n", oper);
		if(mask & FILE_ADD_SUBDIRECTORY)	print("	%s add subdir / pipe to dir\n", oper);
		if(mask & FILE_TRAVERSE)			print("	%s access dir\n", oper);
		if(mask & FILE_DELETE_CHILD)		print("	%s delete file from dir\n", oper);
	}
	else{
		if(mask & FILE_READ_DATA)			print("	%s read\n", oper);
		if(mask & FILE_WRITE_DATA)			print("	%s write\n", oper);
		if(mask & FILE_APPEND_DATA)			print("	%s append\n", oper);
		if(mask & FILE_EXECUTE)				print("	%s execute\n", oper);
	}

	if(mask & FILE_READ_EA)				print("	%s read EA\n", oper);
	if(mask & FILE_WRITE_EA)			print("	%s write EA\n", oper);
	if(mask & FILE_READ_ATTRIBUTES)		print("	%s read attr\n", oper);
	if(mask & FILE_WRITE_ATTRIBUTES)	print("	%s write attr\n", oper);
	if(mask & DELETE)					print("	%s delete\n", oper);
	if(mask & READ_CONTROL)				print("	%s read control\n", oper);
	if(mask & WRITE_DAC)				print("	%s write DAC\n", oper);
	if(mask & WRITE_OWNER)				print("	%s take ownership\n", oper);
	if(mask & SYNCHRONIZE)				print("	%s synchronise\n", oper);
}

void
main(int argc, char *argv[])
{
	int i, err, attr;
	void *ace;
	Biobuf bout;
	wchar_t *tmp;
	PACL dacl, sacl;
	PSID owner, group;
	SECURITY_DESCRIPTOR *sd;
	SECURITY_INFORMATION si;
	ACL_SIZE_INFORMATION sz;

	si = OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|DACL_SECURITY_INFORMATION;
	ARGBEGIN{
	}ARGEND;

	if(argc != 1){
		fprint(2, "usage: %s pathspec\n", argv0);
		exits("usage");
	}

	owner = nil;
	group = nil;
	dacl = nil;
	sacl = nil;
	sd = nil;

	tmp = _winpath(argv[0]);
	if((attr = GetFileAttributes(tmp)) == -1)
		sysfatal("%s GetFileAttributes() failed, %s", argv[0], _winerr());

	if((err = GetNamedSecurityInfo(tmp, SE_FILE_OBJECT, si, &owner, &group, &dacl, &sacl, &sd)) != ERROR_SUCCESS){
		SetLastError(err);
		sysfatal("%s GetFileSecurity() failed, %s", argv[0], _winerr());
	}

	if(dacl == nil)
		exits(0);

	if(GetAclInformation(dacl, (void *)&sz, sizeof(sz), AclSizeInformation) == 0)
		sysfatal("%s cannot get ACE size, %s", argv[0], _winerr());

	Binit(&bout, 1, OWRITE);
	for(i = 0; i < sz.AceCount; i++){
		if(GetAce(dacl, i, &ace) == 0 || ace == nil){
			fprint(2, "%s: cannot get ACE %d, %s\n", argv[0], i, _winerr());
			continue;
		}
		dumpace(&bout, ace, attr);
	}

	exits(nil);
}
