/*
 *
 * 9p client
 *	/dev/cpunote - interrupt
 *	/dev/wdir - for the plumber
 *	/dev/label - ?
 *	/dev/cons - stdin, stdout, stderr
 *	/env - export some/all of enviroment
 *
 * interrupt handling - <del> causes interrupt to be sent to shell
 * eof handling - ^d causes shell's input fifo to be closed.
 * exetype for rc(1)
 *
 * TLS before passwords exchanged
 *
 * Winstations should be allocated per user rather than per session.
 */
#include <stdio.h>
#include <stdarg.h>
#include <windows.h>
#include <userenv.h>
#include <lm.h>
#include "dosd.h"

char *argv0;

/* get the machines current domain */
static void
getdefdom(char *dom, int n)
{
	WKSTA_INFO_100  *buf;

	if(NetWkstaGetInfo(NULL, 100, (void *)&buf) != NERR_Success)
		sysfatal("NetWkstaGetInfo() failed");
	wcstombs(dom, buf->wki100_langroup, n);
	NetApiBufferFree(buf);

}

/* authenticate user */
static void
authenticate(HANDLE inp, HANDLE *token, char *user, char *authdom)
{
	char passwd[64];

	getstr(inp, passwd, sizeof(passwd));
	if(LogonUser(user, authdom, passwd, LOGON32_LOGON_INTERACTIVE,
	    LOGON32_PROVIDER_DEFAULT, token) == 0)
		sysfatal("authentication failed");
	memset(passwd, 0, strlen(passwd));

}

/* get users home dir and  profile script path */
static void
getpaths(HANDLE token, char *domain, char *user, char *prof, int nprof, char *home, int nhome)
{
	unsigned long n;
	USER_INFO_3 *ui;
	wchar_t uusr[128], udom[128], *updc;

	*prof = 0;
	snprintf(home, nhome, "c:/");

	n = nhome;
	GetUserProfileDirectory(token, home, &n);

	updc = NULL;
	mbstowcs(udom, domain, sizeof(udom)/sizeof(udom[0]));
	if(NetGetDCName(NULL, udom, (void *)&updc) != NERR_Success)
		return;

	mbstowcs(uusr, user, sizeof(uusr)/sizeof(uusr[0]));
	if(NetUserGetInfo(updc, uusr, 3, (unsigned char **)&ui) != NERR_Success){
		NetApiBufferFree(updc);
		return;
	}
	NetApiBufferFree(updc);

	wcstombs(prof, ui->usri3_profile, sizeof(prof));
	wcstombs(home, ui->usri3_home_dir_drive, sizeof(home));
	if(ui->usri3_password_expired)
		sysfatal("%s@%s password expired", user, domain);
	NetApiBufferFree(ui);
}

static int
chdir(char *path)
{
	int a;

	/*
	 * NB: if SetCurrentDirectory() fails then the subsequent
	 * CreateProcess() wedges as it creats a process in hyperspace
	 * thus we test the dir exists first. If the chdir(cwd) and
	 * chdir(home) both fail then we inherit the dir from listen(1)
	 * of c:\listen
	 */
	a = GetFileAttributes(path);
	if(a == -1 || (a & FILE_ATTRIBUTE_DIRECTORY) == 0)
		return -1;
	if(SetCurrentDirectory(path) == 0)
		return -1;
	return 0;
}

int
main(int argc, char *argv[])
{
	void *envblk;
	extern char *Logfile;
	HANDLE token, profhand, inp, out, err;
	char env[1024], *plan9, logfile[64], *p;
	char prof[512], cmd[4096], home[128], defdom[32];
	char cwd[1024], user[128], *authdom;

	argv0 = "dos";		/* so the error messages on plan9 look right */

	Logfile = "C:/temp/dosd.log";
	if((plan9 = getenv("plan9")) != NULL){
		snprintf(logfile, sizeof(logfile), "%s/sys/log/dosd", plan9);
		Logfile = logfile;
	}

	logmsg("started\n");

	inp = GetStdHandle(STD_INPUT_HANDLE); 
	out = GetStdHandle(STD_OUTPUT_HANDLE); 
	err = GetStdHandle(STD_ERROR_HANDLE); 
	SetConsoleMode(inp, ENABLE_PROCESSED_INPUT); 
	SetConsoleMode(out, 0);
	SetConsoleMode(err, 0);

	getdefdom(defdom, sizeof(defdom));
	putstr(out, "%s\n", defdom);
	getstr(inp, user, sizeof(user));
	if ((authdom = strchr(user, '@')) == NULL)
		sysfatal("%s protocol botch, no '@'", user);
	*authdom++ = 0;

	authenticate(inp, &token, user, authdom);

	getpaths(token, authdom, user, prof, sizeof(prof), home, sizeof(home));

	profhand = loadprofile(token, user, prof);

	initenv(user, home);
	while(getstr(inp, env, sizeof(env)) != 0){
		if((p = strchr(env, '=')) != NULL){
			*p = 0;
			SetEnvironmentVariable(env, p+1);
		}
	}

	if(CreateEnvironmentBlock(&envblk, token, TRUE) == 0)
		syswarn("cannot create enviroment block");

	unmapall();

	if(ImpersonateLoggedOnUser(token) == 0)
		syswarn("impersonation failed");

	mapdrives(inp);

	getstr(inp, cwd, sizeof(cwd));
	if(chdir(cwd) == -1){
		syswarn("failed to chdir to '%s'", cwd);
		chdir(home);
	}

	getstr(inp, cmd, sizeof(cmd));
	shell(inp, out, err, token, envblk, cmd);

	DestroyEnvironmentBlock(envblk);
	unloadprofile(token, profhand);
	exit(0);
}
