#include <stdlib.h>
#include <windows.h>
#include <userenv.h>
#include <lm.h>
#include "dosd.h"

static void
maperr(int err, char *local, char *remote, char *user)
{
	char *msg;

	switch(err){
	case ERROR_ACCESS_DENIED:				msg = "access denied"; break;
 	case ERROR_ALREADY_ASSIGNED:			msg = "drive already mapped"; break;
	case ERROR_BAD_DEV_TYPE:				msg = "name type mismatch"; break;
 	case ERROR_BAD_DEVICE:					msg = "bad local name"; break;
	case ERROR_BAD_NET_NAME:				msg = "bad remote name"; break;
 	case ERROR_BAD_PROFILE:					msg = "bad profile"; break;
 	case ERROR_BAD_PROVIDER:				msg = "bad network provider"; break;
	case ERROR_BAD_USERNAME:				msg = "invalid username"; break;
 	case ERROR_BUSY:						msg = "busy - try again"; break;
 	case ERROR_CANCELLED:					msg = "request cancelled"; break;
 	case ERROR_CANNOT_OPEN_PROFILE:			msg = "can't open profile"; break;
 	case ERROR_DEVICE_ALREADY_REMEMBERED: 	msg = "local already mapped"; break;
	case ERROR_INVALID_ADDRESS: 			msg = "invalid address"; break;
	case ERROR_INVALID_PARAMETER:			msg = "invalud parameter";  break;
 	case ERROR_INVALID_PASSWORD:			msg = "invalid password"; break;
	case ERROR_LOGON_FAILURE:				msg = "authentication failed"; break;
	case ERROR_NO_NET_OR_BAD_PATH:			msg = "cannot connect"; break;
	case ERROR_NO_NETWORK:					msg = "network unavailable"; break;
		syswarn("map: %s -> %s as %s - %s", local, remote, user, msg);
		break;
 	case ERROR_EXTENDED_ERROR:
		syswarn("map: %s -> %s as %s - network error", local, remote, user);
		break;
	default:
		syswarn("map: %s -> %s as %s - system error", local, remote, user);
		break;
	}
}

/* remove any vestigial (useless) drive mappings which could block our new ones */
void
unmapall(void)
{
	HANDLE h;
	void *buf;
	NETRESOURCE *p;
	unsigned long n, l;
	enum { Nbuf = 16 * 1024 };

	if((buf = malloc(Nbuf)) == NULL){
		syswarn("no memory");
		return;
	}

	if(WNetOpenEnum(RESOURCE_CONNECTED, RESOURCETYPE_ANY,
	    RESOURCEUSAGE_ALL, NULL, &h) != NO_ERROR){
		syswarn("cannot enumerate connected network resources");
		return;
	}

	n = -1;
	l = Nbuf;
	while(WNetEnumResource(h, &n, buf, &l) == NO_ERROR){
		for(p = buf; n--; p++){
			if(WNetCancelConnection2(p->lpLocalName, 0, 1) != NO_ERROR)
				syswarn("%s %s - unmap failed",
				    p->lpLocalName, p->lpRemoteName);
		}
		n = -1;
		l = Nbuf;
	}
	WNetCloseEnum(h);
}

/* map all the drives requested by remote */
void
mapdrives(HANDLE inp)
{
	int err;
	NETRESOURCE nr;
	char local[128], remote[128];
	char user[128], passwd[64];

	while(1){
		getstr(inp, local, sizeof(local));
		if(*local == 0)
			break;
		getstr(inp, remote, sizeof(remote));
		getstr(inp, user, sizeof(user));
		getstr(inp, passwd, sizeof(passwd));

		memset(&nr, 0, sizeof(NETRESOURCE));
		nr.dwType = RESOURCETYPE_DISK;
		nr.lpLocalName = local;  
		nr.lpRemoteName	= remote;

		if((err = WNetAddConnection2(&nr, passwd, user, CONNECT_TEMPORARY)) != NO_ERROR)
			maperr(err, local, remote, user);
		memset(passwd, 0, sizeof(passwd));
	}
}
