#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

enum {
	BUFCHARS = 8192
};

static int
nchars(wchar_t *p)
{
	int n;

	for(n = 0; *p++; n++)
		continue;
	return n;
}

void
list(Biobuf *bp, void *root, wchar_t *path)
{
	HKEY key;
	int i, l, len;
	ulong nt, nv;
	wchar_t *name, *value;

	if((name = malloc(BUFCHARS*sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");
	if((value = malloc(BUFCHARS*sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");

	if(RegOpenKeyEx(root, path, 0, KEY_READ, &key) != ERROR_SUCCESS)
		sysfatal("openkey failed, %s", _winerr());

	len = 0;
	for(i = 0; ; i++){
		nt = BUFCHARS;
		if(RegEnumValue(key, i, name, &nt, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
			break;
		l = nchars(name);
		if(l > len)
			len = l;
	}

	for(i = 0; ; i++){
		nt = BUFCHARS;
		if(RegEnumValue(key, i, name, &nt, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
			break;
		nv = BUFCHARS;
		if(RegQueryValueEx(key, name, NULL, NULL, (BYTE *)value, &nv) != ERROR_SUCCESS){
			fprint(2, "%Q queryvalue failed - %r\n", name);
			continue;
		}
		Bprint(bp, "%*W %W\n", -len, name, value);
	}

	RegCloseKey(key);
	free(name);
	free(value);
}


void
show(Biobuf *bp, void *root, wchar_t *path, wchar_t *name)
{
	HKEY key;
	ulong sz;
	wchar_t *value;

	if((value = malloc(BUFCHARS*sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");

	if(RegOpenKeyEx(root, path, 0, KEY_READ, &key) != ERROR_SUCCESS)
		sysfatal("openkey failed, %s", _winerr());

	sz = BUFCHARS;
	if(RegQueryValueEx(key, name, NULL, NULL, (BYTE *)value, &sz) != ERROR_SUCCESS)
		sysfatal("queryvalue failed, %s", _winerr());

	Bprint(bp, "%W %W\n", name, value);
	RegCloseKey(key);
	free(value);
}

void
change(void *root, wchar_t *path, wchar_t *name, wchar_t *value)
{
	HKEY key;
	ulong sz, type, junk;

	if(RegOpenKeyEx(root, path, 0, KEY_WRITE|KEY_READ, &key) != ERROR_SUCCESS)
		sysfatal("openkey failed, %s", _winerr());

	if(*value != 0){
		sz = 0;
		if(RegQueryValueEx(key, name, NULL, &type, NULL, &sz) != ERROR_SUCCESS)
			type = REG_SZ;
		sz = _winlen(value)*sizeof(wchar_t);
		if(RegSetValueEx(key, name, 0, type, (CONST BYTE *)value, sz) != ERROR_SUCCESS)
			sysfatal("queryvalue failed, %s", _winerr());
	}
	else{
		if(RegDeleteValue(key, name) != ERROR_SUCCESS)
			sysfatal("deletefalue failed, %s", _winerr());
	}

	RegFlushKey(key);
	RegCloseKey(key);

	SendMessageTimeout(HWND_BROADCAST, WM_SETTINGCHANGE, 0,
		(LPARAM)"Environment", SMTO_ABORTIFHUNG, 5000, &junk);
}

void
usage(void)
{
	fprint(2, "usage: %s [-a] [[name] value]\n", argv0);
	fprint(2, "       -a      environment for all users\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	void *root;
	Biobuf bout;
	wchar_t *path, *name, *value;

	root = HKEY_CURRENT_USER;
	path = L"Environment";
	ARGBEGIN{
	case 'a':
		root = HKEY_LOCAL_MACHINE;
		path = L"SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment";
		break;
	default:
		usage();
	}ARGEND;

	quotefmtinstall();
	_widefmtinstall();

	Binit(&bout, 1, OWRITE);
	switch(argc){
	case 0:
		list(&bout, root, path);
		break;
	case 1:
		name = _winstr(argv[0]);
		show(&bout, root, path, name);
		break;
	case 2:
		name = _winstr(argv[0]);
		value = _winstr(argv[1]);
		change(root, path, name, value);
		break;
	default:
		usage();
	}
	exits(nil);
}
