#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

#include <ntdef.h>
#include <ddk/ntapi.h>
#include <ntapi-missing.h>
#include <ddk/tdi.h>

/*
 * According to Gary Nebbett, in his book Windows NT/2000
 * Native API Reference, "[ObQueryNameString] ... will
 * hang when querying the names of pipes that have been
 * opened for synchronous access and that have a pending
 * read or write operation."
*/

#define NeedMore	STATUS_INFO_LENGTH_MISMATCH

static Biobuf bout;

static void
prthread(HANDLE h)
{
	HANDLE proc;
	int pid, tid;
	ULONG len, got;
	UNICODE_STRING *us;
	THREAD_BASIC_INFORMATION tbi;

	memset(&tbi, 0, sizeof(tbi));
	NtQueryInformationThread(h, ThreadBasicInformation, &tbi, 0x1c, NULL);
	pid = (int)tbi.ClientId.UniqueProcess;	/* yes, really, these are pointers in the kernel */
	tid = (int)tbi.ClientId.UniqueThread;
	Bprint(&bout, "tid=%ud pid=%ud ", tid, pid);
	if(tbi.ExitStatus != STATUS_PENDING)
		Bprint(&bout, "exited=%lud ", tbi.ExitStatus);

	if((proc = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, pid)) == BadHand){
		Bprint(&bout, "<cannot open process>");
		return;
	}

	us = NULL;
	len = 0xff;
	do{
		got = 0;
		len *= 2;
		us = realloc(us, len);
		memset(us, 0, len);
		us->MaximumLength = (len-sizeof(UNICODE_STRING))/sizeof(wchar_t);
	}while(NtQueryInformationProcess(proc, ProcessImageFileName, us, len, &got) == NeedMore);
	if(len >= got && us->Length){
		char *s = _ninepath(us->Buffer);
		Bprint(&bout, "%q ", s);
		free(s);
	}
	CloseHandle(proc);
	free(us);
}

static void
prproc(HANDLE h)
{
	ULONG len, got;
	UNICODE_STRING *us;

	Bprint(&bout, "pid=%lud ", GetProcessId(h));

	us = NULL;
	len = 0xff;
	do{
		got = 0;
		len *= 2;
		us = realloc(us, len);
		memset(us, 0, len);
		us->MaximumLength = (len-sizeof(UNICODE_STRING))/sizeof(wchar_t);
	}while(NtQueryInformationProcess(h, ProcessImageFileName, us, len, &got) == NeedMore);
	if(len >= got && us->Length){
		char *s = _ninepath(us->Buffer);
		Bprint(&bout, "%q ", s);
		free(s);
	}

	free(us);
}

static void
prfile(HANDLE h)
{
	ULONG len;
	IO_STATUS_BLOCK isb;
	FILE_NAME_INFORMATION *fni;

	fni = NULL;
	len = 0xff;
	do{
		len *= 2;
		fni = realloc(fni, len);
		memset(fni, 0, len);
	}while(NtQueryInformationFile(h, &isb, fni, len, FileNameInformation) == NeedMore);
	if(fni->FileNameLength){
		char *s = _ninepath(fni->FileName);
		Bprint(&bout, "%q ", s);
		free(s);
	}
	free(fni);
}

static char *
network(HANDLE h)
{
	char *proto;
	ULONG len, got;
	OBJECT_NAME_INFORMATION *ni;

	ni = NULL;
	len = 0xff;
	do{
		got = 0;
		len *= 2;
		ni = realloc(ni, len);
		memset(ni, 0, len);
	}while(NtQueryObject(h, ObjectNameInformation, ni, len, &got) == NeedMore);

	proto = NULL;
	if(len >= got && ni->Name.Length){
		if(_wincmp(ni->Name.Buffer, L"\\Device\\Udp") == 0)
			proto = "udp";
		if(_wincmp(ni->Name.Buffer, L"\\Device\\Tcp") == 0)
			proto = "tcp";
	}
	free(ni);
	return proto;
}

static void
objname(HANDLE h)
{
	ULONG len, got;
	OBJECT_NAME_INFORMATION *ni;

	ni = NULL;
	len = 0xff;
	do{
		got = 0;
		len *= 2;
		ni = realloc(ni, len);
		memset(ni, 0, len);
	}while(NtQueryObject(h, ObjectNameInformation, ni, len, &got) == NeedMore);

	if(len >= got && ni->Name.Length){
		char *s = _ninepath(ni->Name.Buffer);
		Bprint(&bout, "%q ", s);
		free(s);
	}
	free(ni);
}


static void
prnet(HANDLE obj, char *proto)
{
	HANDLE eve;
	NTSTATUS rc;
	char *buf[128];
	IO_STATUS_BLOCK IoStatusBlock;
	TDI_REQUEST_QUERY_INFORMATION trqi = { { 0 } , TDI_QUERY_ADDRESS_INFO };

	if((eve = CreateEvent(NULL, TRUE, FALSE, NULL)) == INVALID_HANDLE_VALUE)
		return;
	rc = NtDeviceIoControlFile(obj, eve, NULL, NULL, &IoStatusBlock, IOCTL_TDI_QUERY_INFORMATION,
		    &trqi, sizeof(trqi), &buf, sizeof(buf));
	CloseHandle(eve);

	if(rc != STATUS_SUCCESS)
		return;
/*
 * This appears to be broken, always returning a zero IP address.
 * needs more investigation
 *
 *[2009-08-03] Update1: The IoControlCode of recvfrom is 0x1201b, at the
offset 0x10 of the InputBuffer is the address of the sockaddr, in
which the remote ip (offset <del>0x10</del> 0x04) and port (offset
<del>0x08</del> 0x02) is stored.


[2009-08-01] Update2: The IoControlCode of accept is 0x1200c, at the
offset 0x0C of the OutputBuffer is the remote port and at the offset
of 0x0E is the remote ip.


 */
	Bprint(&bout, "proto=%s ip=%s port=%d ",
		proto,
		inet_ntoa(*(struct in_addr *)&buf[14]),
		ntohs(*(USHORT *)&buf[12]));
}


static DWORD
dohand(void *v)
{
	int e;
	HANDLE h = v;
	ULONG len, got;
	char *proto, *type;
	OBJECT_TYPE_INFORMATION *ti;
	OBJECT_BASIC_INFORMATION bi;

	if((e = NtQueryObject(h, ObjectBasicInformation, &bi, sizeof(bi), NULL)) != 0){
		Bprint(&bout, "\n");
		return 0;
	}

	Bprint(&bout, "%-4lud %-4lud ", bi.HandleCount, bi.PointerCount);

	ti = NULL;
	len = 0xff;
	do{
		got = 0;
		len *= 2;
		ti = realloc(ti, len);
		memset(ti, 0, len);
	}while(NtQueryObject(h, ObjectTypeInformation, ti, len, &got) == NeedMore);
	if(len < got)
		return 0;

	proto = "?";
	type = _ninestr(ti->Name.Buffer);
	if(strcmp(type, "File") == 0){
		if(GetFileType(h) == FILE_TYPE_PIPE)
			Bprint(&bout, "%-16s ", "Pipe");
		else
		if((proto = network(h)) != 0)
			Bprint(&bout, "%-16s ", "Network");
		else
			Bprint(&bout, "%-16s ", "File");
	}
	else
		Bprint(&bout, "%-16s ", type);

	if(strcmp(type, "Network") == 0)
		prnet(h, proto);
	else
	if(strcmp(type, "File") == 0)
		prfile(h);
	else
	if(strcmp(type, "Process") == 0)
		prproc(h);
	else
	if(strcmp(type, "Thread") == 0)
		prthread(h);
	else
		objname(h);
	Bprint(&bout, "\n");

	free(type);
	free(ti);
	return 0;
}

static char
stdhand(HANDLE proc, HANDLE h)
{
	PEB peb;
	DWORD got;
	PROCESS_BASIC_INFORMATION pbi;
	RTL_USER_PROCESS_PARAMETERS upp;

	NtQueryInformationProcess(proc, ProcessBasicInformation, &pbi, sizeof(pbi), &got);
	if(got != sizeof(pbi))
		return '?';

	ReadProcessMemory(proc, pbi.PebBaseAddress, &peb, sizeof(peb), &got);
	if(got != sizeof(peb))
		return '?';

	ReadProcessMemory(proc, peb.ProcessParameters, &upp, sizeof(upp), &got);
	if(got != sizeof(upp))
		return '?';
	if(h == upp.hStdInput)
		return 'I';
	if(h == upp.hStdOutput)
		return 'O';
	if(h == upp.hStdError)
		return 'E';
	return '-';
}

/*
 * some of the handle examination functions
 * can block so we just give up after a short timeout
 */
static void
hand(HANDLE proc, SYSTEM_HANDLE_TABLE_ENTRY_INFO *te, int tout)
{
	int e;
	ULONG inf;
	HANDLE raw, h, t;

	raw = (HANDLE *)(long)te->Handle;

	if((e = NtDuplicateObject(proc, raw, (HANDLE)0xffffffff, &h, 0, TRUE, 0xfffffffe)) != 0){
//		Bprint(&bout, "%s", _ntdllerr(e));
		return;
	}

	Bprint(&bout, "0x%-4x ", te->Handle);

	inf = 0;
	if(GetHandleInformation(raw, &inf) != 0){
		Bprint(&bout, "%c%c%c ",  ((inf & HANDLE_FLAG_INHERIT)? 'i': '-'),
			((inf & HANDLE_FLAG_PROTECT_FROM_CLOSE)? 'p': '-'),
			stdhand(proc, raw));
	}
	else{
		Bprint(&bout, "    ");
	}


	t = CreateThread(NULL, 20, (void *)dohand, h, 0, NULL);
	if(WaitForSingleObject(t, tout) == WAIT_TIMEOUT){
		TerminateThread(t, 0);
		Bprint(&bout, "[timeout]\n");
	}
	CloseHandle(t);
	CloseHandle(h);
}

int
handinfo(int pid, int tout)
{
	DWORD got;
	int i, len;
	HANDLE proc;
	SYSTEM_HANDLE_INFORMATION_correct *shi;
	int flags = PROCESS_DUP_HANDLE|PROCESS_QUERY_INFORMATION|PROCESS_VM_READ;

	if((proc = OpenProcess(flags, FALSE, pid)) == INVALID_HANDLE_VALUE){
		fprint(2, "%s: %d cannot open process - %s\n", argv0, pid, _winerr());
		return -1;
	}

	len = 1024;
	shi = NULL;
	do{
		len *= 2;
		shi = realloc(shi, len);
		NtQuerySystemInformation(SystemHandleInformation, shi, len, &got);
	}while(got > len);

	if(shi->NumberOfHandles == 0)
		sysfatal("%d - cannot get handle list\n", pid);

	for(i = 0; i < shi->NumberOfHandles; i++)
		if(shi->Handles[i].ProcessId == pid)
			hand(proc, &shi->Handles[i], tout);

	free(shi);
	CloseHandle(proc);
	return 0;
}

static void
usage(void)
{
	fprint(2, "usage: %s [-t n] pid...\n", argv0);
	fprint(2, "-t  n   set timeout for locked handles to n milliseconds\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int i, tout;

	tout = 30;		/* miliseconds */
	ARGBEGIN{
	case 't':
		tout = atoi(EARGF(usage()));
		break;
	default:
		usage();
	}ARGEND;

	if(tout < 1 || tout > 60000)
		sysfatal("silly timeout");
	if(argc < 1)
		usage();

	if(_takepriv(SE_DEBUG_NAME) == -1)
		fprint(2, "%s: cannot gain debug priviledge, owned processes only\n", argv0);

	quotefmtinstall();

	Binit(&bout, 1, OWRITE);
	for(i = 0; i < argc; i++)
		handinfo(atoi(argv[i]), tout);

	exits(nil);
}
