/*
 * Caveats.
 *
 * Hard links can can link only files and each link has its own copy of
 * the file's metadata. This metadata is updated only on the link
 * used to modify the file's contents. here can only be 1023 hard links
 * can be created per file.
 *
 * Take care when deleteing links, if you delete them with
 * explorer it will recursively delete the contents not just the link,
 * use the console rmdir command.
 *
 * When any path may contain at most 32 soft links.
 */

#include <u.h>
#include <libc.h>
#include <mswin.h>

#include <winnt.h>

int	failed;
int	sflag;
void	link(char *from, char *to, int todir);
void	hardlink(wchar_t *src, wchar_t *dst);
void	symlink(wchar_t *src, wchar_t *dst);

void
main(int argc, char *argv[])
{
	Dir *dirb;
	int todir, i;

	ARGBEGIN {
	case 's':
		sflag++;
		break;
	default:
		goto usage;
	} ARGEND

	_widefmtinstall();
	quotefmtinstall();

	todir=0;
	if(argc < 2)
		goto usage;
	dirb = dirstat(argv[argc-1]);
	if(dirb!=nil && (dirb->mode&DMDIR))
		todir=1;
	if(argc>2 && !todir){
		fprint(2, "ln: %s not a directory\n", argv[argc-1]);
		exits("bad usage");
	}
	for(i=0; i<argc-1; i++)
		link(argv[i], argv[argc-1], todir);
	if(failed)
		exits("errors");
	exits(0);

usage:
	fprint(2, "usage:\tcp [-s] fromfile tofile\n");
	fprint(2, "\tln [-s] fromfile ... todir\n");
	exits("usage");
}

void
link(char *from, char *to, int todir)
{
	wchar_t *src, *dst;
	src = _winpath(from);
	if(todir){
		wchar_t *s, *elem;
		elem=s=src;
		while(*s++)
			if(s[-1] == L'/' || s[-1] == L'\\')
				elem=s;
		dst = _winpath2(to, elem);
	}
	else
		dst = _winpath(to);

	if(sflag)
		symlink(src, dst);
	else
		hardlink(src, dst);
	free(src);
	free(dst);
}


void
ensureNTFS(wchar_t *path)
{
	int sz;
	wchar_t *drv, voltype[64];

	sz = GetFullPathName(path, 0, nil, nil);
	if((drv = malloc(sz * sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");
	sz = GetFullPathName(path, sz, drv, nil);

	drv[3] = 0;
	if(GetVolumeInformation(drv, NULL, 0, NULL, NULL, NULL, voltype, sizeof(voltype)) == 0) 
		sysfatal("%W cannot get volume info, %s", drv, _winerr());
	if(_wincmp(voltype, L"NTFS") != 0)
		sysfatal("%W not an NTFS filesystem", path);
	free(drv);
}

void
hardlink(wchar_t *src, wchar_t *dst)
{
	int a;

	ensureNTFS(src);
	ensureNTFS(dst);

	if((a = GetFileAttributes(src)) == -1)
		sysfatal("%W not found, %s", src, _winerr());
	if((a & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
		sysfatal("%W is a directory", src);

	if(GetFileAttributes(dst) != -1)
		sysfatal("%W already exists", dst);

	if(CreateHardLink(dst, src, NULL) == 0)
		sysfatal("%W %W cannot create link, %s", src, dst, _winerr());
}

BOOLEAN WINAPI
_CreateSymbolicLink(LPTSTR src, LPTSTR dst, DWORD flg)
{
	static BOOLEAN WINAPI (*f)(LPTSTR, LPTSTR, DWORD) = nil;
	if(f == nil)
		f = _latebind("kernel32", "CreateSymbolicLinkW");
	return f(src, dst, flg);
}

void
symlink(wchar_t *src, wchar_t *dst)
{
	int a, linktype;

	ensureNTFS(src);

	if((a = GetFileAttributes(src)) == -1)
		sysfatal("%W not found, %s", src, _winerr());
	if((a & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
		linktype = SYMBOLIC_LINK_FLAG_DIRECTORY;
	else
		linktype = 0;

	if(GetFileAttributes(dst) != -1)
		sysfatal("%W already exists", dst);

	if(_CreateSymbolicLink(dst, src, linktype) == 0)
		sysfatal("%W %W cannot symlink, %r", src, dst, _winerr());
}
