#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

#include <psapi.h>

#ifndef PAGE_WRITECOMBINE
#define PAGE_WRITECOMBINE 	0x400		/* missing in mingw */
#endif

enum {
	BUFCHARS = 8192
};

static int Allpath;

static void
vm(Biobuf *bp, HANDLE h)
{
	unsigned char *p;
	MEMORY_BASIC_INFORMATION mbi;
	char *base, *tmp, size[32], *state, *type, *prot;
	wchar_t *path;

	if((path = malloc(BUFCHARS*sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");

	p = nil;
	while(VirtualQueryEx(h, p, &mbi, sizeof(mbi)) == sizeof(mbi)){
		if(mbi.State == MEM_FREE){
			p += mbi.RegionSize;
			continue;
		}

		switch(mbi.State){
		case MEM_COMMIT:
			state = "commit";
			break;
		case MEM_RESERVE:
			state = "reserve";
			break;
		case MEM_FREE:
			state = "free";
			break;
		default:
			state = "?";
			break;
		}

		switch(mbi.Type){
		case MEM_IMAGE:
			type = "exe";
			break;
		case MEM_MAPPED:
			type = "file";
			break;
		case MEM_PRIVATE:
			type = "mem";
			break;
		default:
			if(mbi.State == MEM_FREE)
				type = " ";
			else
				type = "?";
		}

		switch (mbi.AllocationProtect & ~(PAGE_GUARD | PAGE_NOCACHE)) {
		case PAGE_NOACCESS:
			prot = "-";
			break;
		case PAGE_READONLY:
			prot = "r";
			break;
		case PAGE_READWRITE:
			prot = "rw";
			break;
		case PAGE_WRITECOPY:
			prot = "w,cow";
			break;
		case PAGE_EXECUTE:
			prot = "x";
			break;
		case PAGE_EXECUTE_READ:
			prot = "rx";
			break;
		case PAGE_EXECUTE_READWRITE:
			prot = "rwx";
			break;
		case PAGE_EXECUTE_WRITECOPY:
			prot = "x,cow";
			break;
		case PAGE_WRITECOMBINE:
			prot = "w,merge";
			break;
		default:
			if(mbi.State == MEM_FREE)
				prot = " ";
			else
				prot = "?";
		}

		*path = 0;
		tmp = nil;
		base = "";
		if(mbi.Type != MEM_PRIVATE && mbi.State != MEM_FREE){
			if(GetMappedFileName(h, mbi.BaseAddress, path, BUFCHARS) == 0){
				base = "?";
			}
			else{
				tmp = _ninepath(path);
				base = tmp;
				if(! Allpath){
					if((base = strrchr(tmp, '/')) == nil)
						base = tmp;
					else
						base++;
				}
			}
		}


		if(mbi.BaseAddress == mbi.AllocationBase)
			snprint(size, sizeof(size), "%ludK",
				mbi.RegionSize / 1024);
		else
			snprint(size, sizeof(size), "%lud+%dK",
				mbi.RegionSize / 1024,
				(mbi.BaseAddress - mbi.AllocationBase) / 1024);

		Bprint(bp, "0x%-8p %10s %-7s %-8s %-8s %s%s %s\n",
			mbi.BaseAddress,
			size, 
			state, type, prot,
			((mbi.AllocationProtect & PAGE_GUARD)? "g": " "),
			((mbi.AllocationProtect & PAGE_NOCACHE)? "n": " "),
			base);

		p += mbi.RegionSize;
		free(tmp);
	}
}

int
meminfo(Biobuf *bp, int pid)
{
	HANDLE h;

	if((h = OpenProcess(PROCESS_VM_READ | PROCESS_QUERY_INFORMATION, 0, pid)) == nil){
		fprint(2, "%s: %d cannot open process - %s\n", argv0, pid, _winerr());
		return -1;
	}
	vm(bp, h);
	CloseHandle(h);
	return 0;
}

static void
usage(void)
{
	fprint(2, "usage: %s [-a] pid...\n", argv0);
	fprint(2, "   -a       print full path names\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int i;
	Biobuf bout;

	ARGBEGIN{
	case 'a':
		Allpath = 1;
		break;
	default:
		usage();
	}ARGEND;

	if(argc < 1)
		usage();

	if(_takepriv(SE_DEBUG_NAME) == -1)
		fprint(2, "%s: cannot gain debug priviledge, owned processes only\n", argv0);

	quotefmtinstall();

	Binit(&bout, 1, OWRITE);
	for(i = 0; i < argc; i++)
		meminfo(&bout, atoi(argv[i]));
	exits(nil);
}
