#define _SPAWN_EXTENSION
#include <u.h>
#include <libc.h>
#include <ip.h>
#include <mswin.h>

#include <winsock2.h>
#include <ws2tcpip.h>
#include <iphlpapi.h>

static void
usage(void)
{
	fprint(2, "usage: %s [mac-addr]\n", argv0);
	exits("usage");
}

static char *
shellerr(int n)
{
	static char buf[32];

	switch(n){
	case 0:							return "os out of resources";
	case ERROR_FILE_NOT_FOUND:		return "file not found";
	case ERROR_PATH_NOT_FOUND:		return "path not found";
	case ERROR_BAD_FORMAT:			return "not executable";
	case SE_ERR_ACCESSDENIED:		return "access denied";
	case SE_ERR_ASSOCINCOMPLETE:	return "incomplete or invalid associateion";
	case SE_ERR_DDEBUSY:			return "DDE busy";
	case SE_ERR_DDEFAIL:			return "DDE fail";
	case SE_ERR_DDETIMEOUT:			return "DDE timeout";
	case SE_ERR_DLLNOTFOUND:		return "dde not found";
	case SE_ERR_NOASSOC:			return "no association for file";
	case SE_ERR_OOM:				return "no memory";
	case SE_ERR_SHARE:				return "sharing voliation";
	default:
		break;
	}
	snprint(buf, sizeof(buf), "unknown error %d\n", n);
	return buf;
}


void
main(int argc, char *argv[])
{
	ULONG len;
	char *type;
	wchar_t url[1024];
	uchar mac[6], buf[8192];
	MIB_IPNETTABLE *iptab;
	int i, found, err, match;

	ARGBEGIN{
	}ARGEND;

	fmtinstall('V', eipfmt);
	fmtinstall('E', eipfmt);
	_widefmtinstall();

	match = 0;
	found = 0;
	len = sizeof(buf);
	iptab = (MIB_IPNETTABLE *)buf;

	switch(argc){
	case 0:
		match = 0;
		break;
	case 1:
		match = 1;
		if(parseether(mac, argv[0]) < 0)
			sysfatal("%s - cannot parse ethernet address", argv[0]);
		break;
	default:
		usage();
	}

	if((err = GetIpNetTable(iptab, &len, 1)) != NO_ERROR)
		sysfatal("GetIpNetTable() failed, %d\n", err);

	for(i = 0; i < iptab->dwNumEntries; i++){
		switch(iptab->table[i].dwType){
		case MIB_IPNET_TYPE_OTHER:
			type = "Other";
			break;
		case MIB_IPNET_TYPE_DYNAMIC:
			type = "OK";
			break;
		case MIB_IPNET_TYPE_STATIC:
			type = "Static";
			break;
		case MIB_IPNET_TYPE_INVALID:
			type = "Invalid";
			break;
		}
		if(!match)
			print("ether	%-8s	%-16V %E\n",  type, &iptab->table[i].dwAddr, iptab->table[i].bPhysAddr);
		else{
			if(iptab->table[i].dwType != MIB_IPNET_TYPE_DYNAMIC && iptab->table[i].dwType != MIB_IPNET_TYPE_STATIC)
				continue;

			if(iptab->table[i].dwPhysAddrLen != 6)
				continue;

			if(memcmp(iptab->table[i].bPhysAddr, mac, 6) == 0){
				found = 1;
				runesnprint(url, sizeof(url), "http://%V/webclient\n", &iptab->table[i].dwAddr);
				err = (int)ShellExecute(nil, L"open", url, nil, nil, SW_SHOWNORMAL);

				if(err == 0 || err <= 32)
					fprint(2, "ShellExecute(open %w) failed, %s\n", url, shellerr(err));
				break;
			}
		}

	}

	if(! found)
		sysfatal("%E not found in ARP table");
	exits(0);
}
