#include <u.h>
#include <libc.h>
#include <mswin.h>

typedef struct {
	char *name;
	int prio;
} Namval;

static Namval Procprio[] = {
	{"high", HIGH_PRIORITY_CLASS},
	{"above", ABOVE_NORMAL_PRIORITY_CLASS},
	{"normal", NORMAL_PRIORITY_CLASS},
	{"below", BELOW_NORMAL_PRIORITY_CLASS},
	{"idle", IDLE_PRIORITY_CLASS},
};

static Namval Thrprio[] = {
	{"highest", THREAD_PRIORITY_HIGHEST},
	{"above", THREAD_PRIORITY_ABOVE_NORMAL},
	{"normal", THREAD_PRIORITY_NORMAL},
	{"below", THREAD_PRIORITY_BELOW_NORMAL},
	{"lowest", THREAD_PRIORITY_LOWEST},
	{"idle", THREAD_PRIORITY_IDLE},
};

void
usage(void)
{
	fprint(2, "usage: %s [opts] id...\n", argv0);
	fprint(2, "  -c       process ctrl-c\n");
	fprint(2, "  -b       process ctrl-break\n");
	fprint(2, "  -t       process terminate\n");
	fprint(2, "  -p name  process base priority\n");
	fprint(2, "  -S       thread suspend\n");
	fprint(2, "  -R       thread resume\n");
	fprint(2, "  -P name  thread base proirity\n");
	fprint(2, "  -T       thread terminate\n");
	fprint(2, "  prio names:  highest, above, normal, below, lowest, idle\n");
	exits("usage");
}

void
procaux(char op, int id, char *name)
{
	int i;
	HANDLE h;

	switch(op){
	case 't':
		if((h = OpenProcess(PROCESS_TERMINATE, FALSE, id)) == nil)
			sysfatal("%d cannot open process, %s", id, _winerr());
		if(TerminateProcess(h, 99) == 0)
			sysfatal("%s cannot terminate process, %s", id, _winerr());
		CloseHandle(h);
		break;

	case 'c':
		if(GenerateConsoleCtrlEvent(id, CTRL_C_EVENT) == 0)
			sysfatal("%d cannot send process ctrl-c, %s", id, _winerr());
		break;

	case 'b':
		if(GenerateConsoleCtrlEvent(id, CTRL_BREAK_EVENT) == 0)
			sysfatal("%d cannot send process ctrl-break, %s", id, _winerr());
		break;

	case 'p':
		if((h = OpenProcess(PROCESS_SET_INFORMATION, FALSE, id)) == nil)
			sysfatal("%d cannot open process, %s", id, _winerr());
		for(i = 0; i < nelem(Procprio); i++)
			if(cistrcmp(Procprio[i].name, name) == 0)
				break;
		if(i >= nelem(Procprio))
			sysfatal("%s unknown process priority, %s", name, _winerr());
		if(SetPriorityClass(h, Procprio[i].prio) == 0)
			sysfatal("%d set process priority failed, %s", id, _winerr());
		CloseHandle(h);
		break;

	case 'T':
		if((h = OpenThread(THREAD_TERMINATE, FALSE, id)) == nil)
			sysfatal("%d cannot open thread, %s", id, _winerr());
		if(TerminateThread(h, 99) == 0)
			sysfatal("%d terminate thread failed, %s", id, _winerr());
		CloseHandle(h);
		break;

	case 'S':
		if((h = OpenThread(THREAD_SUSPEND_RESUME, FALSE, id)) == nil)
			sysfatal("%d cannot open thread, %s", id, _winerr());
		if(SuspendThread(h) == 0)
			sysfatal("%d suspend thread failed, %s", id, _winerr());
		CloseHandle(h);
		break;

	case 'R':
		if((h = OpenThread(THREAD_SUSPEND_RESUME, FALSE, id)) == nil)
			sysfatal("%d cannot open thread, %s", id, _winerr());
		if(ResumeThread(h) == 0)
			sysfatal("%d resume thread failed, %s", id, _winerr());
		CloseHandle(h);
		break;

	case 'P':
		if((h = OpenThread(THREAD_SET_INFORMATION, FALSE, id)) == BadHand)
			sysfatal("%d cannot open thread, %s", id, _winerr());
		for(i = 0; i < nelem(Thrprio); i++)
			if(cistrcmp(Thrprio[i].name, name) == 0)
				break;
		if(i >= nelem(Thrprio))
			sysfatal("%s unknown thread priority, %s", name, _winerr());
		if(SetThreadPriority(h, Thrprio[i].prio) == 0)
			sysfatal("%d set thread priority failed, %s", id, _winerr());
		CloseHandle(h);
		break;
	default:
		sysfatal("%c bad operator, internal error", op);
	}
}

void
main(int argc, char *argv[])
{
	int i;
	char op, *name;

	op = 0;
	name = nil;
	ARGBEGIN{
	case 't':
	case 'c':
	case 'b':
	case 'T':
	case 'S':
	case 'R':
		op = ARGC();
		break;
	case 'P':
	case 'p':
		op = ARGC();
		name = EARGF(usage());
		break;
	default:
		usage();
	}ARGEND;

	if(_takepriv(SE_DEBUG_NAME) < 0)
		fprint(2, "%s: cannot take debug priviledge, owned processes only\n");

	if(op == 0)
		usage();
	for(i = 0; i < argc; i++)
		procaux(op, atoi(argv[i]), name);

	exits(nil);
}
