#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include <mswin.h>

#include <lm.h>
#include <lmdfs.h>

typedef struct _Usr Usr;
struct _Usr {
	char *serv;
	char *user;
	char *host;
	long login;
	long idle;
	Usr *left;
	Usr *right;
};

int multi = 0;
int lflag = 0;
int aflag = 0;
char *argv0;

int uwid = 0;
int swid = 0;
int hwid = 0;

char *
strlwr(char *s)
{
	char *p;

	for(p = s; *p; p++)
		if(isupper(*p))
			*p = tolower(*p);
	return s;
}

char *
elapsed(long sec)
{
	int days, hrs, min;
	static char buf[32];

	days = sec / (60L * 60L * 24L);
	sec -= days * (60L * 60L * 24L);
	hrs  = sec / (60L * 60L);
	sec -= hrs * (60L * 60L);
	min  = sec / 60L;
	if(days)
		snprint(buf, sizeof(buf), "%d,%d:%d ", days, hrs, min);
	else
		snprint(buf, sizeof(buf), "%d:%d ", hrs, min);
	return buf;
}

Usr *
add(Usr *u, char *serv, char *user, char *host, long login, long idle)
{
	int n;

	if(u == nil){
		if((u = mallocz(sizeof(Usr), 1)) == nil)
			sysfatal("No memory, %r");
		u->serv = serv;
		u->user = user;
		u->host = host;
		u->idle = idle;
		u->login = login;
		return u;
	}
	
	if((n = strcmp(u->serv, serv)) == 0){
		if((n = strcmp(u->user, user)) == 0){
			free(user);
			free(host);
			free(serv);
			return u;
		}
		u->left = add(u->left, serv, user, host, login, idle);
	}
	else
	if(n < 0)
		u->left = add(u->left, serv, user, host, login, idle);
	else
	if(n > 0)
		u->right = add(u->right, serv, user, host, login, idle);
	return u;
	
}

void
widths(Usr *u)
{
	if(u == nil)
		return;

	widths(u->right);
	uwid = max(uwid, strlen(u->user));
	swid = max(swid, strlen(u->serv));
	hwid = max(hwid, strlen(u->host));
	widths(u->left);
}

void
prnt(Biobuf *bp, Usr *u)
{
	int n;

	if(u == nil)
		return;

	prnt(bp, u->right);
	n = strlen(u->user);
	if(aflag || (n > 0 && u->user[n-1] != '$')){
		if(multi)
			Bprint(bp, "%*q ", swid, strlwr(u->serv));

		if(lflag){
			Bprint(bp, "%*q %10s %10s %*s\n",
				-uwid, strlwr(u->user),
				elapsed(u->login),
				elapsed(u->idle),
				-hwid, strlwr(u->host));
		}
		else
			Bprint(bp, "%q\n", strlwr(u->user));
	}
	prnt(bp, u->left);
}

Usr *
server(Usr *u, wchar_t *server)
{
	int i, err;
	char *user, *host, *serv;
	unsigned char *buf;
	SESSION_INFO_10	*p;
	DWORD got, n, tot, res;

	res = 0;
	got = 0;
	do{
		got = 0;
		buf = nil;
		
		err = NetSessionEnum(server, L"", L"", 10, &buf,
			MAX_PREFERRED_LENGTH, &n, &tot, &res);
		if(err != ERROR_SUCCESS && err != ERROR_MORE_DATA){
			fprint(2, "%s: %W - cannot get session info, %s\n", argv0, server, _ntdllerr(err));
			return u;
		}
		p = (SESSION_INFO_10 *)buf;
		for (i = 0; i < n; i++, p++){
			user = _ninestr(p->sesi10_username);
			if(p->sesi10_cname[0] == L'\\' && p->sesi10_cname[1] == L'\\')
				 host = _ninestr(p->sesi10_cname +2);
			else
				 host = _ninestr(p->sesi10_cname);
			serv = _ninestr(server);
			u = add(u, serv, user, host, p->sesi10_time, p->sesi10_idle_time);
		}
		if(buf)
			NetApiBufferFree(buf);
		got += n;
	}while(got > tot);

	if(got == 0)
		fprint(2, "%s: %W - no entries\n", argv0, server);
	return u;
}

int
seen(wchar_t *s)
{
	int i, n;
	static int num = 0;
	static int use = 0;
	static wchar_t **tab = nil;

	n = _winlen(s);
	for (i = 0; i < use; i++)
		if(memcmp(tab[i], s, n * sizeof(wchar_t)) == 0)
			return 1;

	if(use >= num){
		num += 16;
		if((tab = realloc(tab, num * sizeof(wchar_t *))) == nil)
			sysfatal("No memory, %r");
	}
	if((tab[use] = malloc((n+1) * sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");
	memcpy(tab[use], s, (n+1)*sizeof(wchar_t));
	use++;
	return 0;
}

#ifdef NO_LONGER_USED
	
Usr *
dfsroot(Usr *u, wchar_t *path)
{
	int i, err;
	DFS_INFO_3 *p;
	DFS_STORAGE_INFO *q;

	err = NetDfsGetInfo(path, nil, nil, 3, (void *)&p);
	if(err != ERROR_SUCCESS)
		return u;
	q = (DFS_STORAGE_INFO *)p->Storage;
	for (i = 0; i < p->NumberOfStorages; i++, q++){
		if(seen(q->ServerName))
			break;
		u = server(u, q->ServerName);
	}
	NetApiBufferFree(p);
	return u;
}


/*
 * This code iterates over all mounted SMB shares.
 * It can be very slow if you are using DFS.
 */
Usr *
mounts(Usr *u)
{
	HANDLE h;
	int i, err;
	NETRESOURCE *p;
	DWORD got, nbuf;
	unsigned char buf[4096];

	nbuf = sizeof(buf);
	if((err = WNetOpenEnum(RESOURCE_CONNECTED, RESOURCETYPE_ANY, 0, nil, &h)) != 0)
		sysfatal("WNetOpenEnum() failed, %s", _wneterr(err));

	while(1){
		err = WNetEnumResource(h, &got, buf, &nbuf);
		if(err == ERROR_NO_MORE_ITEMS)
			break;
		if(err != 0){
			fprint(2, "%s: WNetEnumResource(), %s\n", argv0, _wneterr(err));
			break;
		}
		p = (NETRESOURCE *)buf;
		for(i = 0; i < got; i++, p++)
			u = dfsroot(u, p->lpRemoteName);
	}
	WNetCloseEnum(h);
	return u;
}
#endif

void
usage(void)
{
	fprint(2, "usage: %s [-ld] [server]\n", argv0);
	fprint(2, "   -l        long list with server, terminal and login and idle times\n");
	fprint(2, "   -a        all, include hidden DFS sessions\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int i;
	Usr *u;
	Biobuf bout;
	unsigned long len;
	wchar_t *host, hostname[128];

	u = nil;
	ARGBEGIN{
	case 'a':		/* all users, even psudo users */
		aflag++;
		break;
	case 'l':		/* long list, source IP, connect time etc */	
		lflag++;
		break;
	default:
		usage();
	}ARGEND;

	_widefmtinstall();
	quotefmtinstall();

	Binit(&bout, 1, OWRITE);
	if(argc == 0){
		len = nelem(hostname);
		GetComputerName(hostname, &len);
		u = server(u, hostname);
	}
	else{
		if(argc > 1)
			multi = 1;
		for(i = 0; i < argc; i++){
			host = _winstr(argv[i]);
			u = server(u, host);
			free(host);
		}
	}

	widths(u);
	prnt(&bout, u);
	exits(nil);
}
