#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <windows.h>
#include <imagehlp.h>

#define PACKAGE "I HATE GNU"
#define PACKAGE_VERSION "I HATE GNU"
#include "bfd.h"

typedef struct Module Module;
struct Module {
	bfd *bdf;
	asymbol **symbols;
	int size;
	bfd_vma base;
	char *file;
	Module *next;
};

typedef struct Look Look;
struct Look {
	asymbol **symbols;
	bfd_vma addr;
	bfd_vma base;
	int found;
};

Module *Modlist;
char Cwd[1024];
char *argv0;
int Verbose;

#define nil NULL

void *
emalloc(long n)
{
	void *p;

	if((p = malloc(n)) == nil){
		fprintf(stderr, "%s: No memory\n", argv0);
		exit(-1);
	}
	return p;
}

const char *
basename(const char *s)
{
	const char *p;

	if((p = strrchr(s, '/')) != nil)
		s = p+1;
	if((p = strrchr(s, '\\')) != nil)
		s = p+1;
	return s;
}


char *
pathcpy(char *buf, const char *s, const char *e)
{
	char *t;

	for(t = buf; *s && t < e; s++, t++){
		if(*s == '\\')
			*t = '/';
		else
			*t = *s;
	}

	*t = 0;
	return buf;
}

const char *
relpath(const char *path)
{
	int n;
	const char *p;
	char *c, pp, cc;
	static char buf[1024], *e = &buf[sizeof(buf)-1];

	if(strpbrk(path, "/\\") == nil)
		return pathcpy(buf, path, e);

	c = Cwd;
	p = path;
	while(*c && *p){
		cc = *c;
		if(isupper(cc))
			cc = tolower(cc);
		pp = *p;
		if(isupper(pp))
			pp = tolower(pp);
		if(! (cc == pp) ||
			 (cc == '/' && pp == '\\') ||
			 (cc == '\\' && pp == '/'))
			break;
		p++;
		c++;
	}

	/* cwd is above or exactly at the target dir */
	if(*c == 0)
		return pathcpy(buf, p+1, e);

	/* managed to strip the drive only */
	if((c - Cwd) == 2)
		return pathcpy(buf, p+1, e);

	/* cwd is deeper than target */
	n = 0;
	while(c && *c){
		pathcpy(buf+n, "../", e);
		n += 3;
		c = strpbrk(c+1, "/\\");
	}

	if(*p == '/' || *p == '\\')
		p++;
	pathcpy(buf+n, p, e);
	return buf;
}

/* generally only used for things like kernel32.dll and ntdll.dll */
int
microsoft(char *file, bfd_vma base, bfd_vma addr)
{
	HANDLE h;
	HMODULE mh;
	DWORD disp;
	bfd_vma off;
	char buf[256];
	IMAGEHLP_LINE ln;
	int gotline, gotsym;
	union { IMAGEHLP_SYMBOL s; char n[1024]; } sym;

	/*
	 * allow for the library being loaded
	 * at a different address locally than in the
	 * program that faulted - rare but does happen
	 */
	if((mh = LoadLibrary(file)) == nil)
		return -1;

	off = (bfd_vma)mh - base;

	h = GetCurrentProcess();

	memset(&ln, 0, sizeof(ln));
	ln.SizeOfStruct = sizeof(ln);
	gotline = SymGetLineFromAddr(h, addr+off, &disp, &ln);


	disp = 0;
	memset(&sym, 0, sizeof(sym));
	sym.s.SizeOfStruct = sizeof(sym);
	sym.s.MaxNameLength = sizeof(sym)-sizeof(IMAGEHLP_SYMBOL);
	gotsym = SymGetSymFromAddr(h, addr+off, 0, &sym.s);

	if(! gotline && ! gotsym){
		FreeLibrary(mh);
		return -1;
	}

	if(gotline)
		snprintf(buf, sizeof(buf), "%s:%d ", relpath(ln.FileName), ln.LineNumber);
	else
		snprintf(buf, sizeof(buf), ": ");
	printf("%-32s ", buf);

	if(gotsym)
		printf("%s()", sym.s.Name);
	else
		printf("0x%p()", (void *)addr);

	FreeLibrary(mh);
	return 0;
}

void 
libbfd(bfd *abfd, asection *sect, void *ctx)
{
	Look *lk;
	bfd_vma vma;
	char buf[256];
	const char *file;
	const char *func;
	unsigned int line;

	lk = ctx;
	if((bfd_get_section_flags(abfd, sect) & SEC_ALLOC) == 0) 
		return;

	vma = bfd_get_section_vma(abfd, sect);
	if(lk->addr < vma || vma + bfd_get_section_size(sect) <= lk->addr) 
		return;

	file = nil;
	func = nil;
	line = 0;
	if(bfd_find_nearest_line(abfd, sect, lk->symbols, lk->addr - vma, &file, &func, &line) == 0)
		return;

	if(file && line)
		snprintf(buf, sizeof(buf), "%s:%d", relpath(file), line);
	else
	if(file && !line)
		snprintf(buf, sizeof(buf), "%s", relpath(file));
	else
	if(!file && line)
		snprintf(buf, sizeof(buf), ":%d", line);
	else
		snprintf(buf, sizeof(buf), ":");
	printf("%-32s ", buf);

	if(func)
		printf("%s()", func);
	else
		printf("0x%p()", (void *)lk->addr);
	lk->found = 1;
}

int
looksym(Module *m, bfd_vma addr)
{
	Look lk;

	memset(&lk, 0, sizeof(lk));
	lk.symbols = m->symbols;
	lk.base = m->base;
	lk.addr = addr;
	bfd_map_over_sections(m->bdf, &libbfd, &lk);

	return 0;
}

int
srcline(char *file, bfd_vma base, bfd_vma addr)
{
	int e, num;
	bfd *bdf;
	Module *m;
	void *symbols;
	unsigned size;

	for(m = Modlist; m; m = m->next)
		if(strcmp(m->file, file) == 0)
			return looksym(m, addr);

	if((bdf = bfd_openr(file, 0)) == nil){
		if(Verbose)
			fprintf(stderr, "%s open failed - %s\n", file, bfd_errmsg(bfd_get_error()));
		return -1;
	}
	if((e = bfd_check_format_matches(bdf, bfd_object, nil)) == 0){
		if(Verbose)
			fprintf(stderr, "%s incorrect format - %s\n", file, bfd_errmsg(bfd_get_error()));
		bfd_close(bdf);
		return -1;
	}

	if((bfd_get_file_flags(bdf) & HAS_SYMS) != HAS_SYMS){
		if(Verbose)
			fprintf(stderr, "%s symbols stripped\n", file);
		bfd_close(bdf);
		return microsoft(file, base, addr);		/* fall back to MS techniques */
	}

	size = 0;
	if((num = bfd_read_minisymbols(bdf, FALSE, &symbols, &size)) == 0){
		if((num = bfd_read_minisymbols(bdf, TRUE, &symbols, &size)) == 0){
			bfd_close(bdf);
			if(Verbose)
				fprintf(stderr, "%s failed to read symbols - %s\n", file, bfd_errmsg(bfd_get_error()));
			return -1;
		}
	}

	m = emalloc(sizeof(Module) + strlen(file)+1);
	m->file = (char *)(&m[1]);  strcpy(m->file, file);
	m->symbols = symbols;
	m->bdf = bdf;
	m->size = size;
	m->base = base;

	m->next = Modlist;
	Modlist = m;
	return looksym(m, addr);
}

void
usage(void)
{
	fprintf(stderr, "usage: %s [-V] [file]\n", argv0);
	exit(1);
}

int
parse(char *input, FILE *sp)
{
	int rc, line;
	char *file, *addr, *base, buf[1024];

	SymInitialize(GetCurrentProcess(),  0, 1);

	rc = 0;
	for(line = 1; fgets(buf, sizeof(buf), stdin); line++){
		if((file = strtok(buf, " \t\n")) == nil)
			continue;									/* blank line */
		if(strcmp(file, ".") == 0)
			break;										/* end of input */
		if((base = strtok(nil, " \t\n")) == nil){
			fprintf(stderr, "%s: %s:%d missing base address\n", argv0, input, line);
			continue;
		}
		if((addr = strtok(nil, " \t\n")) == nil){
			fprintf(stderr, "%s: %s:%d missing return address\n", argv0, input, line);
			continue;
		}

		printf("%-16s ", basename(file));
		rc |= srcline(file, bfd_scan_vma(base, nil, 0), bfd_scan_vma(addr, nil, 0));
		printf("\n");
	}
	return rc;
}


int
main(int argc, char *argv[])
{
	FILE *fp;
	int i, rc, o;

	while((o = getopt(argc, argv, "V")) != -1)
		switch(o){
		case 'V':
			Verbose++;
			break;
		default:
			usage();
		}

	argv0 = argv[0];
	if(getcwd(Cwd, sizeof(Cwd)) == nil)
		fprintf(stderr, "%s: cannot get cwd\n", argv0);

	rc = 0;
	if(optind == argc)
		rc = parse("stdin", stdin);
	else
	for(i = optind; i < argc; i++){
		if((fp = fopen(argv[i], "r")) == nil)
			fprintf(stderr, "%s: %s cannot open file - %s\n", argv0, argv[i], strerror(errno));
		else{
fprintf(stderr, "%s\n", argv[i]);
			rc |= parse(argv[i], fp);
		}
		if(fp)
			fclose(fp);
	}

	return rc;
}
