#include <u.h>
#include <libc.h>
#include <mswin.h>

void * 
mapaddr(uchar *buf, IMAGE_SECTION_HEADER *ish, int nsect, int size, int addr)
{
	int i, off;

	for(i = 0; i < nsect; i++){
      if(addr >= ish->VirtualAddress && addr < ish->VirtualAddress+ish->SizeOfRawData)
			break;
		ish++;
	}
	if(i >= nsect)
		return 0;

	off = ish->PointerToRawData + addr - ish->VirtualAddress;
	if(off >= size)
		return 0;
	return buf + off;
}

static void
usage(void)
{
	fprint(2, "usage: %s [-v] [ntdll.dll]\n", argv0);
	fprint(2, "   -v     verbose, print failed decodes also\n");
	exits("usage");
}

void
main(int argc, char* argv[])
{
	WORD  *ords;
	HANDLE h, map;
	uchar *buf, *text;
	DWORD *funcs, *names;
	wchar_t *file, *name;
	IMAGE_NT_HEADERS *NThdr;
	IMAGE_DOS_HEADER *DOShdr;
	IMAGE_SECTION_HEADER *sects;
	IMAGE_EXPORT_DIRECTORY *dir;
	int syscall, verbose, all, i, j, va, size, nsect;

	_widefmtinstall();

	all = 0;
	verbose = 0;
	ARGBEGIN{
	case 'v':
		verbose++;
		break;
	case 'a':
		all++;
		break;
	default:
		usage();
		break;
	}ARGEND;

	switch(argc){
	case 0:
		file = L"c:/windows/system32/ntdll.dll";
		break;
	case 1:
		file = _winstr(argv[0]);
		break;
	default:
		usage();
		break;
	}

	if((h = CreateFile(file, 
		GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_DELETE, nil, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, nil)) == INVALID_HANDLE_VALUE )
			sysfatal("%w - cannot open, %s\n", file, _winerr());

	size = GetFileSize(h, nil);

	if((map = CreateFileMapping( h, nil, PAGE_READONLY, 0, 0, nil )) == nil)
		sysfatal("%w - cannot map into memory, %s", file, _winerr());

	if((buf = (uchar *)MapViewOfFile( map, FILE_MAP_READ, 0, 0, 0 )) == nil)
		sysfatal("%w - cannot map view of file, %s\n", file, _winerr());

	if( buf[0] != 'M' || buf[1] != 'Z' )
		sysfatal("%s - mad magic number\n", file);

	DOShdr = (IMAGE_DOS_HEADER *)buf;
	NThdr  = (IMAGE_NT_HEADERS *)(buf + DOShdr->e_lfanew);
	nsect  = NThdr->FileHeader.NumberOfSections;
	sects  = (IMAGE_SECTION_HEADER *)(buf +DOShdr->e_lfanew +sizeof(IMAGE_NT_HEADERS));

	if((va = NThdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress) == 0)
		sysfatal("%w no export directory\n", file);

	dir   = mapaddr(buf, sects, nsect, size, va);
	funcs = mapaddr(buf, sects, nsect, size, dir->AddressOfFunctions);
	ords  = mapaddr(buf, sects, nsect, size, dir->AddressOfNameOrdinals);
	names = mapaddr(buf, sects, nsect, size, dir->AddressOfNames);

	for(i = 0; i < dir->NumberOfNames; i++){
		name = mapaddr(buf, sects, nsect, size, names[i]);
		text = mapaddr(buf, sects, nsect, size, funcs[ords[i]]);

		/*
		 * Check if the API entry begins with:
		 *
		 *	MOV  EAX, IMM32
		 *  XOR  ECX, ECX
		 *  LEA  EDX, [ESP+04h]
		 *  CALL FS:[C0h]
		 *
		 * Or
		 *
		 *	MOV  EAX, IMM32
		 *  MOV  ECX, IMM32
		 *  LEA  EDX, [ESP+04h]
		 *  CALL FS:[C0h]
		 */
		if(text[0] == 0xb8){									    	    // mov eax, imm32
			syscall = *(DWORD *)(text+1);

			if(text[5] == 0x33 && text[6] == 0xC9 &&		            	// xor ecx, ecx
				!memcmp( &text[7],  "\x8D\x54\x24\x04", 4) &&         	 	// lea edx, [esp+04h]
				!memcmp( &text[11], "\x64\xFF\x15\xC0\x00\x00\x00", 7))		// call fs:[C0h]
					print("%-4d %-4d 32 %s\n", ords[i], syscall, name);
			else
			if(text[5] == 0xB9 &&								        	// mov ecx, imm32
				!memcmp( &text[10], "\x8D\x54\x24\x04", 4) &&          		// lea edx, [esp+04h]
				!memcmp( &text[13], "\x64\xFF\x15\xC0\x00\x00\x00", 7)) 	// call fs:[C0h]
					print("%-4d %-4d 64 %s\n", ords[i], syscall, name);
			else
			if(text[5] == 0xB9 &&								        	// mov ecx, imm32
				!memcmp( &text[10], "\x8D\x54\x24\x04", 4) &&          		// lea edx, [esp+04h]
				!memcmp( &text[14], "\x64\xFF\x15\xC0\x00\x00\x00", 7)) 	// call fs:[C0h]
					print("%-4d %-4d 64 %s\n", ords[i], syscall, name);
			else
			if(verbose){
				print("%-4d %-4s ?  %-40s %02x ", ords[i], "?", name, text[0]);
				for(j = 1; j < 24 && text[j] != 0xb8; j++)
					print("%02x ", text[j]);
				print("\n");
			}

		}
		else
		if(all)
			print("%-4d %-4s    %s\n", ords[i], "", name);

	}

	CloseHandle(h);
	UnmapViewOfFile(buf);
	CloseHandle(map);
}
