#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

#include <ntdef.h>
#include <ddk/ntapi.h>
#include <ntapi-missing.h>

#define ll2f(x)	((double)((x).HighPart) * 4.294967296E9 + (double)((x).LowPart))

static double
cpumhz(void)
{
	HKEY hk;
	unsigned long mhz, sz;

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
		0, KEY_READ, &hk) != 0)
		return -1;
	sz = sizeof(mhz);
	if(RegQueryValueEx(hk, L"~MHz", NULL, NULL, (void *)&mhz, &sz) != 0)
		return -1;
	return mhz;
}

static int
cputimes(double *uptime, double *idle)
{
	int e;
	unsigned long stl, spl;
	SYSTEM_TIME_INFORMATION *sti;         
	SYSTEM_PERFORMANCE_INFORMATION	*spi;

	stl = 1024;
	if((sti = malloc(stl)) == nil)
		sysfatal("No memory, %r");
	while((e = NtQuerySystemInformation(SystemTimeOfDayInformation, sti, stl, &stl)) == 0xC0000004)
		if((sti = realloc(sti, stl)) == nil)
			sysfatal("No memory, %r");
	if(e != NO_ERROR){
		fprint(2, "%s: cannot query SystemTimeInformation failed sz=%d e=%x\n", argv0, stl, e);
		return -1;
	}
	*uptime = ll2f(sti->liKeSystemTime);
	
	spl = 1024;
	if((spi = malloc(spl)) == nil)
		sysfatal("No memory, %r");
	while((e = NtQuerySystemInformation(SystemPerformanceInformation, spi, spl, &spl)) == 0xC0000004)
		if((spi = realloc(spi, spl)) == nil)
			sysfatal("No memory, %r");
	if(e != NO_ERROR){
		fprint(2, "%s: cannot query SystemPerformanceInformation sz=%ld e=%lx\n", argv0, spl, e);
		return -1;
	}
	*idle = ll2f(spi->IdleTime);

	free(spi);
	free(sti);
	return 0;
}

double
cpuload(int nproc)
{
	double load, uptime[2], idle[2], id, up;

	if(cputimes(&uptime[0], &idle[0]) < 0)
		return -1;
	sleep(1000);
	cputimes(&uptime[1], &idle[1]);

	up = uptime[1]-uptime[0];
	id = (idle[1]-idle[0]) / nproc;
	load = 100.0-(id/up)*100.0;

	if(load < 0)
		load = 0;

	return load;
}

void
main(int argc, char *argv[])
{
	SYSTEM_INFO si;
	MEMORYSTATUSEX ms;
	BOOL WINAPI (*GlobalMemoryStatusEx)(MEMORYSTATUSEX *);

	GlobalMemoryStatusEx = _latebind("kernel32.dll", "GlobalMemoryStatusEx");

	ms.dwLength = sizeof(ms);
	if(GlobalMemoryStatusEx(&ms) == 0)
		fprint(2, "%s: GlobalMemoryStatusEx failed, %s", argv[0], _winerr());
	else
		print("mem	%6.2f%% %.1fGb\n",
			(double)ms.dwMemoryLoad,
			(double)ms.ullTotalPhys / (1024.0 * 1024.0 * 1024.0) + 0.05);

	GetNativeSystemInfo(&si);
	print("cpu %6.2f%% %.1fGHz x %d\n",
		cpuload(si.dwNumberOfProcessors),
		cpumhz() / 1000.0 + 0.05,
		si.dwNumberOfProcessors);

	exits(nil);
}

