#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

#include <wtsapi32.h>
#include <wtsapi32-missing.h>

Biobuf bout;
int aflag;
int lflag;
int nflag;
int multi;
int userwid;
int hostwid;
int termwid;

typedef struct Usr Usr;
struct Usr {
	Usr *left;
	Usr *right;

	char *user;
	char *host;
	int sess;
	vlong idle;
	vlong login;
	char *term;
	int state;
};

int
ciwcscmp(wchar_t *s1, wchar_t *s2)
{
	Rune c1, c2;

	for(;;) {
		c1 = *s1++;
		if(isupperrune(c1))
			c1 = tolowerrune(c1);
		c2 = *s2++;
		if(isupperrune(c2))
			c1 = tolowerrune(c2);
		if(c1 != c2) {
			if(c1 > c2)
				return 1;
			return -1;
		}
		if(c1 == 0)
			return 0;
	}
}

/* local session appears on different session IDs depending on windows version */
void
statestr(int n, char *buf, int len)
{
	switch(n){
	case WTSActive:
		snprint(buf, len, "");
		break;
	case WTSConnected:
		snprint(buf, len, "<connected>");
		break;
	case WTSConnectQuery:
		snprint(buf, len, "<connect query>");
		break;
	case WTSShadow:
		snprint(buf, len, "<dhadow>");
		break;
	case WTSDisconnected:
		snprint(buf, len, "<disconnected>");
		break;
	case WTSIdle:
		snprint(buf, len, "<idle>");
		break;
	case WTSListen:
		snprint(buf, len, "<listen>");
		break;
	case WTSReset:
		snprint(buf, len, "<reset>");
		break;
	case WTSDown:
		snprint(buf, len, "<down>");
		break;
	case WTSInit:
		snprint(buf, len, "<init>");
		break;
	default:
		snprint(buf, len, "<unknwon %d>", n);
		break;
	}
}


void
elap(int sec, char *buf, int len)
{
	int days, hrs, min;

	if(sec == -1){
		*buf = 0;
		return;
	}
	days = sec / (60L * 60L * 24L);
	sec -= days * (60L * 60L * 24L);
	hrs  = sec / (60L * 60L);
	sec -= hrs * (60L * 60L);
	min  = sec / 60L;
	if(days)
		snprint(buf, len, "%d %d:%d ", days, hrs, min);
	else
		snprint(buf, len, "%d:%d ", hrs, min);
}

void
show(Usr *u, int locsess)
{
	char *flag, login[16], idle[16], state[16];

	if(! lflag){
		if(u->state == WTSActive){
			if(multi)
				Bprint(&bout, "%*s ", -hostwid, u->host);
			Bprint(&bout, "%s\n", u->user);
		}
		return;
	}
		
	elap(u->idle, idle, sizeof(idle));
	elap(u->login, login, sizeof(login));
	statestr(u->state, state, sizeof(state));

	if(u->sess == locsess)
		flag = " <console>";
	else
		flag = "";
	if(multi)
		Bprint(&bout, "%*s ", -hostwid, u->host);

	Bprint(&bout, "%*q %10s %10s %*s %q %s\n", -userwid, u->user, idle, login, -termwid, u->term, state, flag);
}


void
prnt(Usr *u, int locsess)
{
	if(u){
		prnt(u->left, locsess);
		if(aflag || u->state == WTSActive)
			show(u, locsess);
		prnt(u->right, locsess);
	}
}

Usr *
add(Usr *u, char *host, char *user, int sess, char *term, long login, long idle, int state)
{
	int n;

	if(u == nil){
		if((u = mallocz(sizeof(Usr), 1)) == nil)
			sysfatal("No memory, %r");
		n = strlen(user);
		if(n > userwid)
			userwid = n;
		n = strlen(host);
		if(n > hostwid)
			hostwid = n;
		u->host = host;
		u->user = user;
		u->sess = sess;
		u->login = login;
		u->idle = idle;
		u->term = term;
		n = strlen(term);
		if(n > termwid)
			termwid = n;
		u->state = state;
		return u;
	}

	n = strcmp(host, u->host);
	if(sess == u->sess && n == 0){
		free(user);
		free(term);
		return u;
	}

	if(n < 0 || (n == 0 && sess < u->sess))
		u->left = add(u->left, host, user, sess, term, login, idle, state);
	if(n > 0 || (n == 0 && sess > u->sess))
		u->right = add(u->right, host, user, sess, term, login, idle, state);
	return u;
}

void *
inforaw(HANDLE h, int sess, int class)
{
	wchar_t *buf;
	DWORD got;

	got = 0;
	buf = nil;
	if(WTSQuerySessionInformation(h, sess, class, &buf, &got) == 0){
		werrstr(_winerr());
		return nil;
	}
	if(got == 0)
		return nil;
	return buf;
}


Usr *
connection(Usr *u, char *host, HANDLE h, int sess, int state)
{
	long idle, login;
	struct in_addr a;
	char *tmp, *term, *user;
	WTSINFO *wsi;
	WTSCLIENT *wci;
	WTS_CLIENT_ADDRESS *wca;

	idle = 0;
	login = 0;
	term = nil;
	user = "?";

	if((tmp = inforaw(h, sess, WTSClientInfo)) == nil){
		fprint(2, "%s: WTSClientInfo failed - %r\n", argv0);
		return u;
	}
	wci = (WTSCLIENT *)tmp;
	if(! nflag)
		term = _ninestr(wci->ClientName);
//	user = _ninestr(wci->UserName);

	WTSFreeMemory(tmp);

	if((tmp = inforaw(h, sess, WTSSessionInfo)) == nil){
		fprint(2, "%s: WTSSessionInfo failed - %r\n", argv0);
		free(term);
		free(user);
		return u;
	}
	wsi = (WTSINFO *)tmp;
	if(wsi->LastInputTime.QuadPart)
		idle = (wsi->CurrentTime.QuadPart - wsi->LastInputTime.QuadPart) / 1e9;
	if(wsi->LogonTime.QuadPart)
		login = (wsi->CurrentTime.QuadPart - wsi->LogonTime.QuadPart) / 1e9;
	user = _ninestr(wsi->UserName);
	WTSFreeMemory(tmp);

	if(term == nil || *term == 0){
		if((tmp = inforaw(h, sess, WTSClientAddress)) != nil){
			wca = (WTS_CLIENT_ADDRESS *)tmp;
			if(wca->AddressFamily == AF_INET){
				memcpy(&a, (char *)(&wca->Address)+2, sizeof(a.S_un.S_addr));		/* why +2? because microsoft says so (sigh) */
				term = smprint("%s", inet_ntoa(a));
				if(term == nil)
					sysfatal("No memory, %r");
			}
			WTSFreeMemory(tmp);
		}
	}

	return add(u, host, user, sess, term, login, idle, state);
}

Usr *
sessions(Usr *u, char *host, HANDLE h)
{
	int i;
	DWORD num;
	WTS_SESSION_INFO *sess;

	num = 0;
	sess = nil;
	if(WTSEnumerateSessions(h, 0, 1, &sess, &num) == 0){
		if(GetLastError() != ERROR_NO_MORE_ITEMS)
			fprint(2, "%s: %s - cannot get session list, %s\n", argv0, host, _winerr());
		return u;
	}

	for(i = 0; i < num; i++)
		u = connection(u, host, h, sess[i].SessionId, sess[i].State);
	return u;
}

Usr *
server(Usr *u, wchar_t *whost)
{
	HANDLE h;
	char *host;

	host = _ninestr(whost);
	if((h = WTSOpenServer(whost)) == nil){
		fprint(2, "%s: %s - cannot connect - %s\n", argv0, host, _winerr());
		return u;
	}

	u = sessions(u, host, h);
	WTSCloseServer(h);
	return u;
}
	

/* This plain doesn't work on modern networks as it relies on netbios */
Usr *
domain(Usr *u)
{
	int i;
	DWORD num;
	WTS_SERVER_INFO *srv;

	num = 0;
	srv = nil;
	if(WTSEnumerateServers(nil, 0, 1, &srv, &num) == 0)
		sysfatal("WTSEnumerateServers() failed, %s", _winerr());

	if(num > 1)
		multi = 1;
	for(i = 0; i < num; i++)
		u = server(u, srv[i].pServerName);
	WTSFreeMemory(srv);
	return u;
}

int
localsess(void)
{
	int i;
	DWORD num;
	WTS_PROCESS_INFO *wp;

	if(WTSEnumerateProcesses(0, 0, 1, &wp, &num) == 0)
		sysfatal("WTSEnumerateProcesses failed, %r");

	for(i = 0; i < num; i++)
		if(ciwcscmp(wp[i].pProcessName, L"winlogon.exe") == 0)
			return wp[i].SessionId;
	return -1;
}

void
usage(void)
{
	fprint(2, "usage: %s [-an]\n", argv0);
	fprint(2, "   -n        numeric client address\n");
	fprint(2, "   -a        all sessions, don't hide dull ones\n");
	fprint(2, "   -l        long listing\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Usr *u;
	int i;

	ARGBEGIN{
	case 'l':
		lflag++;
		break;
	case 'a':
		aflag++;
		break;
	case 'n':
		nflag++;
		break;
	default:
		usage();
		break;
	}ARGEND;

	quotefmtinstall();
	_widefmtinstall();

	Binit(&bout, 1, OWRITE);
	if(_takepriv(SE_DEBUG_NAME) == -1)
		fprint(2, "%s: cannot gain debug priviledge, limited info only\n", argv0);

	u = nil;
	if(argc == 0)
		u = sessions(u, "localhost", WTS_CURRENT_SERVER_HANDLE);
	else{
		if(argc > 1)
			multi = 1;
		for(i = 0; i < argc; i++)
			u = server(u, _winstr(argv[i]));
	}
	prnt(u, localsess());
	exits(nil);
}
