#include <u.h>
#include <libc.h>
#include <mswin.h>

#include <wtsapi32.h>
#include <wtsapi32-missing.h>

#include <ntdef.h>
#include <ddk/ntapi.h>
#include <ntapi-missing.h>

#define ll2f(x)	((double)((x).HighPart) * 4.294967296E9 + (double)((x).LowPart))

static int
cputimes(double *uptime, double *idle)
{
	int e;
	unsigned long stl, spl;
	SYSTEM_TIME_INFORMATION *sti;         
	SYSTEM_PERFORMANCE_INFORMATION	*spi;

	stl = 1024;
	if((sti = malloc(stl)) == nil)
		sysfatal("No memory, %r");
	while((e = NtQuerySystemInformation(SystemTimeOfDayInformation, sti, stl, &stl)) == 0xC0000004)
		if((sti = realloc(sti, stl)) == nil)
			sysfatal("No memory, %r");
	if(e != NO_ERROR){
		fprint(2, "%s: cannot query SystemTimeInformation failed sz=%d e=%x\n", argv0, stl, e);
		return -1;
	}
	*uptime = ll2f(sti->liKeSystemTime);
	
	spl = 1024;
	if((spi = malloc(spl)) == nil)
		sysfatal("No memory, %r");
	while((e = NtQuerySystemInformation(SystemPerformanceInformation, spi, spl, &spl)) == 0xC0000004)
		if((spi = realloc(spi, spl)) == nil)
			sysfatal("No memory, %r");
	if(e != NO_ERROR){
		fprint(2, "%s: cannot query SystemPerformanceInformation sz=%ld e=%lx\n", argv0, spl, e);
		return -1;
	}
	*idle = ll2f(spi->IdleTime);

	free(spi);
	free(sti);
	return 0;
}
double
cpuload(int nproc)
{
	double load, uptime[2], idle[2], id, up;

	if(cputimes(&uptime[0], &idle[0]) < 0)
		return -1;
	sleep(1000);
	cputimes(&uptime[1], &idle[1]);

	up = uptime[1]-uptime[0];
	id = (idle[1]-idle[0]) / nproc;
	load = 100.0-(id/up)*100.0;
	if(load < 0)
		load = 0;

	return load;
}

int
numusers(void)
{
	DWORD n;
	WTS_SESSION_INFO *p;

	if(WTSEnumerateSessions(WTS_CURRENT_SERVER_HANDLE, 0, 1, &p, &n) == 0)
		return -1;
	WTSFreeMemory(p);
	return n;
}

void
main(void)
{
	uvlong t;
	int users;
	double load;
	SYSTEM_INFO si;
	FILETIME ft[3];
	char *host;

	if((host = sysname()) == nil)
		sysfatal("cannot get sysname, %r");

	GetSystemInfo(&si);
	load = cpuload(si.dwNumberOfProcessors);
	users = numusers();

	if(GetSystemTimes(ft, ft+1, ft+2) == 0)
		sysfatal("GetSystemTimes failed, %r");

	t = (_ft2elap(ft)+_ft2elap(ft+1)+_ft2elap(ft+2))/(1000*si.dwNumberOfProcessors);

	print("%s up %llud days, %llud:%02llud:%02llud, %d users, load: %.1f%%\n",
		host, t/86400L, (t/3600L)%24L, (t/60L)%60L, t%60L, users, load);

	exits(nil);
}

