#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <win32.h>

#include <ntdef.h>
#include <ddk/ntapi.h>
#include <ntapi-missing.h>

static int
handtype(HANDLE h, char *buf, int len)
{
	int n;
	DWORD arg;
//	DISK_GEOMETRY dg;
	CONSOLE_SCREEN_BUFFER_INFO cinfo;

	if(GetConsoleScreenBufferInfo(h, &cinfo)){
		snprint(buf, len, "cons inp");
		return 0; 
	}
	if(GetConsoleMode(h, &arg)){
		snprint(buf, len, "cons out");
		return 0; 
	}
	if(GetCommMask(h, &arg)){
		snprint(buf, len, "EIA");
		return 0; 
	}

//	if(DeviceIoControl(h, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &dg, sizeof(dg), &arg, nil) == 0){
//		snprint(buf, len, "disk volume");
//		return 0; 
//	}
	
	n = GetFileType(h);
	switch(n){
	case FILE_TYPE_REMOTE:
		snprint(buf, len, "remote CIFS");
		return 0; 
	case FILE_TYPE_DISK:
		snprint(buf, len, "local disk");
		return 0; 
	case FILE_TYPE_CHAR:
		snprint(buf, len, "char dev");
		return 0; 
	case FILE_TYPE_PIPE:
		snprint(buf, len, "pipe");
		return 0; 
	default:
		snprint(buf, len, "%d", n);
		return 0; 
	}

	return -1;
}

static int
objecttype(HANDLE h, char *buf, int len)
{
	DWORD got;
	int sz, rc;
	OBJECT_TYPE_INFORMATION *ti;

	sz = 1024;
	ti = malloc(sz);
again:
	rc = NtQueryObject(h, ObjectTypeInformation, ti, sz, &got);
	if(rc == STATUS_INFO_LENGTH_MISMATCH){
		free(ti);
		sz *= 2;
		ti = malloc(sz);
		goto again;
	}

	if(rc != STATUS_SUCCESS){
		free(ti);
		return -1;
	}
	snprint(buf, len, "%.*w", ti->Name.Length, ti->Name.Buffer);
	free(ti);
	return 0;
}



int
prnt(char *src, wchar_t *dst)
{
	int err;
	HANDLE h;
	char buf[128];
	OBJECT_ATTRIBUTES oa;
	UNICODE_STRING us;
	IO_STATUS_BLOCK io;
	BY_HANDLE_FILE_INFORMATION hfi;

	memset(&io, 0, sizeof(io));
	RtlInitUnicodeString(&us, dst);
	InitializeObjectAttributes(&oa, &us, OBJ_CASE_INSENSITIVE, NULL, NULL);
	err = NtCreateFile(&h, FILE_READ_ATTRIBUTES, &oa, &io, nil, 0, 
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
		FILE_OPEN, FILE_NON_DIRECTORY_FILE, nil, 0);
	if(! NT_SUCCESS(err)){
		fprint(2, "%w %s\n", dst, _ntdllerr(err));
		return -1;
	}

	if(GetFileInformationByHandle(h, &hfi) == 0){
		print("N %s %s\n", src, _winerr());
		return -1;
	}

//	if(handtype(h, buf, sizeof(buf)) < 0)
//	if(objecttype(h, buf, sizeof(buf)) < 0)
//		strcpy(buf, "???");
	CloseHandle(h);

	print("%08x %-24q %w\n", hfi.dwFileAttributes, src, dst);
	return 0;
}

int
deref(char *src, wchar_t *dst)
{
	int err;
	ULONG dw;
	HANDLE h;
	UNICODE_STRING us;
	wchar_t buf[1024];
	OBJECT_ATTRIBUTES oa;

	RtlInitUnicodeString(&us, dst);
	InitializeObjectAttributes(&oa, &us, OBJ_CASE_INSENSITIVE, NULL, NULL);
	err = NtOpenSymbolicLinkObject(&h, SYMBOLIC_LINK_QUERY, &oa);
	if(NT_SUCCESS(err)){
		us.Length = 0;
		us.MaximumLength = sizeof(buf);
		us.Buffer = buf;
		err = NtQuerySymbolicLinkObject(h, &us, &dw);
		NtClose(h);
		deref(src, us.Buffer);

	}
	else{
		prnt(src, dst);
	}
	return 0;
}

int
ls(char *path)
{
	HANDLE h;
	ULONG dw, i;
	char *name, *src;
	wchar_t *wpath, *dst;
	NTSTATUS err;
	char buf[8192];
	UNICODE_STRING us;
	OBJECT_ATTRIBUTES oa;
	OBJDIR_INFORMATION *doi;

	if(path){
		wpath = _winpath(path);
		RtlInitUnicodeString(&us, wpath);
	}
	else{
		path = "";
		RtlInitUnicodeString(&us, L"\\");
	}

	InitializeObjectAttributes(&oa, &us, OBJ_CASE_INSENSITIVE, NULL, NULL);
	err = NtOpenDirectoryObject(&h, STANDARD_RIGHTS_READ|DIRECTORY_QUERY, &oa);
	free(wpath);

	if(! NT_SUCCESS(err)){
		werrstr("NtOpenDirectoryObject(%s) - %s", path,  _ntdllerr(err));
		NtClose(h);
		return -1;
	}


	i = 0;
	doi = (OBJDIR_INFORMATION *)&buf;
	while(1){
		err = NtQueryDirectoryObject(h, doi, sizeof(buf), TRUE, FALSE, &i, &dw);
		if(err == STATUS_NO_MORE_ENTRIES)
			break;
		if(! NT_SUCCESS(err)){
			werrstr("NtQueryDirectoryObject(%s,%d) - %s", path, i, _ntdllerr(err));
			NtClose(h);
			return -1;
		}

		name = _ninepath(doi->ObjectName.Buffer);
		src = smprint("%s/%s", path, name);
		dst = _winpath(src);

		deref(name, dst);

		free(name);
		free(src);
		free(dst);
	}

	NtClose(h);
	return 0;
}

void
main(int argc, char *argv[])
{
	int rc;

	_debugpriv();
	quotefmtinstall();
	fmtinstall('w', _wcfmt);

	if(argc == 1){
		rc = ls("/DosDevices");
		rc |= ls("/Global??");
	}
	else
		rc = ls(argv[1]);
	if(rc < 0)
		sysfatal("%r");

	exits(nil);
}

