#include <u.h>
#include <libc.h>
#include <mswin.h>

/*
 * Differences:
 * This uses a win2 server size copy which can be a big 
 * performance win.
 *
 * mv can change filename capitalisation on windows  
 * filesystems where case is normally ignored.
 *
 * We also allow moving directories, as Windows
 * supports this even though plan9 does not.
 */

int	failed;
void	move(char *from, char *to, int todir);

void
main(int argc, char *argv[])
{
	Dir *dirb;
	int todir, i;

	ARGBEGIN {
	default:
		goto usage;
	} ARGEND

	_widefmtinstall();
	quotefmtinstall();

	todir=0;
	if(argc < 2)
		goto usage;
	dirb = dirstat(argv[argc-1]);
	if(dirb!=nil && (dirb->mode&DMDIR))
		todir=1;
	if(argc>2 && !todir){
		fprint(2, "mv: %s not a directory\n", argv[argc-1]);
		exits("bad usage");
	}
	for(i=0; i<argc-1; i++)
		move(argv[i], argv[argc-1], todir);
	if(failed)
		exits("errors");
	exits(0);

usage:
	fprint(2, "usage:\tmv fromfile tofile\n");
	fprint(2, "\tmv fromfile ... todir\n");
	exits("usage");
}

void
move(char *from, char *to, int todir)
{
	wchar_t *src, *dst;

	src = _winpath(from);
	if(todir){
		wchar_t *s, *elem;
		elem=s=src;
		while(*s++)
			if(s[-1] == L'/' || s[-1] == L'\\')
				elem=s;
		dst = _winpath2(to, elem);
	}
	else
		dst = _winpath(to);

	if(MoveFileEx(src, dst, MOVEFILE_COPY_ALLOWED|MOVEFILE_REPLACE_EXISTING) == 0){
		werrstr("%s", _winerr());
		if(_samefile(src, dst))
			fprint(2, "mv: same file\n", src, dst);
		else
			fprint(2, "mv: %W %W move failed - %r\n", src, dst);
		failed = 1;
	}
	free(src);
	free(dst);
}
