// Need to link with Iphlpapi.lib and Ws2_32.lib

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ip.h>

#include <mswin.h>

#include <winsock2.h>
#include <ws2tcpip.h>
#include <iphlpapi.h>
#include <iprtrmib-missing.h>
#include <wtsapi32.h>
#include <wtsapi32-missing.h>
#include <iptypes.h>

int resolve;

typedef struct Owner Owner;
struct Owner {
	int pid;
	char *uid;
};

Owner *
procowners(void)
{
	int i;
	Owner *o;
	DWORD num;
	WTS_PROCESS_INFO *wp;

	if(WTSEnumerateProcesses(WTS_CURRENT_SERVER_HANDLE, 0, 1, &wp, &num) == 0)
		return nil;

	if((o = malloc(sizeof(Owner) * (num+1))) == nil)
		sysfatal("No memory, %r");

	for(i = 0; i < num; i++){
		o[i].uid = nil;
		o[i].pid = wp[i].ProcessId;
		if(wp[i].ProcessId == 0)
			o[i].uid = "idle";
		else if(wp[i].ProcessId == 4)
			o[i].uid = "System";
		else if(wp[i].pUserSid)
			o[i].uid = _sid2name(wp[i].pUserSid, 100);
	}
	o[i].pid = -1;
	WTSFreeMemory(wp);
	return o;
}

char *
pidowner(Owner *o, int pid)
{
	int i;

	if(pid == 0)
		return "none";		/* typically sockets in time-wait state */

	if(o == nil)
		return "NA";

	/* N is small enough that searching a linear list if fine */
	for(i = 0; o[i].pid != -1; i++)
		if(o[i].pid == pid && o[i].uid)
			return o[i].uid;
	return "-";
}


void
portname(char *buf, int len, int port, char *proto)
{
	struct servent *s;

	if(resolve && (s = getservbyport(port, proto)) != nil)
		snprint(buf, len, "%s", s->s_name);
	else
		snprint(buf, len, "%d", ntohs(port));
}

void
x_hostname(char *buf, int len, int addr)
{
	struct hostent *h;

	if(addr == 0){
		snprint(buf, len, "::");
		return;
	}

	if(resolve && (h = gethostbyaddr((char *)&addr, sizeof(addr), PF_INET)) != nil)
		snprint(buf, len, "%s", h->h_name);
	else
		snprint(buf, len, "%I", &addr);
}

void
hostname(char *buf, int len, int addr)
{
	struct sockaddr_in sa;

	if(addr == 0){
		snprint(buf, len, "::");
		return;
	}

	if(resolve){
	    sa.sin_family = AF_INET;
    	sa.sin_addr.s_addr = addr;
    	if(getnameinfo((struct sockaddr *)&sa, sizeof(sa), buf, len, nil, 0, 0) == 0)
			return;
	}
	snprint(buf, len, "%V", &addr);
}

void
tcpstate(char *buf, int len, int state)
{
	char *p;

	switch(state){
	case MIB_TCP_STATE_CLOSED:
		p = "Closed";
		break;
	case MIB_TCP_STATE_LISTEN:
		p = "Listen";
		break;
	case MIB_TCP_STATE_SYN_SENT:
		p = "SYN-sent";
		break;
	case MIB_TCP_STATE_SYN_RCVD:
		p = "SYN-received";
		break;
	case MIB_TCP_STATE_ESTAB:
		p = "Established";
		break;
	case MIB_TCP_STATE_FIN_WAIT1:
		p = "FIN-wait1";
		break;
	case MIB_TCP_STATE_FIN_WAIT2:
		p = "FIN-wait2 ";
		break;
	case MIB_TCP_STATE_CLOSE_WAIT:
		p = "Close-wait";
		break;
	case MIB_TCP_STATE_CLOSING:
		p = "Closing";
		break;
	case MIB_TCP_STATE_LAST_ACK:
		p = "Last-ACK";
		break;
	case MIB_TCP_STATE_TIME_WAIT:
		p = "Time-wait";
		break;
	case MIB_TCP_STATE_DELETE_TCB:
		p = "Delete-tcb";
		break;
	default:
		p = "Unknown";
		break;
	}
	snprint(buf, len, "%s", p);
}

void *
gettab(void *func, DWORD sz)
{
	int e;
	void *tab;
	DWORD (*get)(void *ptr, DWORD *sz, BOOL sort);

	if((tab = malloc(sz)) == nil)
		sysfatal("No memory, %r");

	get = func;
	while((e = get(tab, &sz, 0)) == ERROR_INSUFFICIENT_BUFFER)
		if((tab = realloc(tab, sz)) == nil)
			sysfatal("No memory, %r");

   if(e != NO_ERROR){
		SetLastError(e);
		free(tab);
		return nil;
	}

	return tab;
}

void *
getexttab(void *v, DWORD sz, int class)
{
	int e;
	void *tab;
	DWORD WINAPI (*get)(void *tab, DWORD *sz, BOOL order, ULONG af, int class, ULONG reserved);

	if((tab = malloc(sz)) == nil)
		sysfatal("No memory, %r");

	get = v;
	while((e = get(tab, &sz, 0, AF_INET, class, 0)) == ERROR_INSUFFICIENT_BUFFER)
		if((tab = realloc(tab, sz)) == nil)
			sysfatal("No memory, %r");

    if(e != NO_ERROR){
		SetLastError(e);
		free(tab);
		return nil;
	}

	return tab;
}

int
tcpsess(Biobuf *bp, Owner *o)
{
	int i;
	MIB_TCPTABLE_OWNER_PID *tab;
	char locport[32], remport[32], remhost[1024], state[32];

	if((tab = getexttab(GetExtendedTcpTable, sizeof(MIB_TCPTABLE_OWNER_PID), TCP_TABLE_OWNER_PID_ALL)) == nil)
		sysfatal("GetExtendedTcpTable failed, %s", _winerr());

	for(i = 0; i < (int)tab->dwNumEntries; i++) {
		tcpstate(state, sizeof(state), tab->table[i].dwState);
		portname(locport, sizeof(locport), tab->table[i].dwLocalPort, "tcp");
		portname(remport, sizeof(remport), tab->table[i].dwRemotePort, "tcp");
		hostname(remhost, sizeof(remhost), tab->table[i].dwRemoteAddr);

		Bprint(bp, "tcp %-4d %-16s ", i, pidowner(o, tab->table[i].dwOwningPid));
		Bprint(bp, "%-16s %-16s %-16s %s\n", state, locport, remport, remhost);
		Bflush(bp);
	}
	free(tab);
	return 0;
}

int
udpsess(Biobuf *bp, Owner *o)
{
	int i;
	MIB_UDPTABLE_OWNER_PID *tab;
	char locport[32];

	if((tab = getexttab(GetExtendedUdpTable, sizeof(MIB_UDPTABLE_OWNER_PID), UDP_TABLE_OWNER_PID)) == nil)
		sysfatal("GetExtendedUdpTable failed, %s", _winerr());

	for(i = 0; i < (int)tab->dwNumEntries; i++) {
		portname(locport, sizeof(locport), tab->table[i].dwLocalPort, "udp");

		Bprint(bp, "udp %-4d %-16s ", i, pidowner(o, tab->table[i].dwOwningPid));
		Bprint(bp, "%-16s %-16s %-16s %s\n", "Open", locport, "0", "::");
		Bflush(bp);
	}
	free(tab);
	return 0;
}

void
ifname(char *buf, int len, int type, int idx)
{
	char *p;

	switch(type){
	case IF_TYPE_OTHER:
		p = "/net/other";
		break;
	case IF_TYPE_ETHERNET_CSMACD:
		p = "/net/ether";
		break;
	case IF_TYPE_ISO88025_TOKENRING:
		p = "/net/tokenring";
		break;
	case IF_TYPE_FDDI:
		p = "/net/fddi";
		break;
	case IF_TYPE_PPP:
		p = "/net/ppp";
		break;
	case IF_TYPE_SOFTWARE_LOOPBACK:
		p = "/dev/null";
		idx = -1;
		break;
	case IF_TYPE_ATM:
		p = "/dev/atm";
		break;
	case IF_TYPE_IEEE80211:
		p = "/dev/wifi";
		break;
	case IF_TYPE_TUNNEL:
		p = "/dev/tunnel";
		break;
	case IF_TYPE_IEEE1394:
		p = "/dev/firewire";
		break;
#ifdef IF_TYPE_IEEE80216_WMAN
	case IF_TYPE_IEEE80216_WMAN:
		p = "/dev/wimax";
		break;
#endif
#ifdef IF_TYPE_WWANPP
	case IF_TYPE_WWANPP:
		p = "/dev/gsm";
		break;
#endif
#ifdef IF_TYPE_WWANPP2
	case IF_TYPE_WWANPP2:
		p = "/dev/gsm-cdma";
		break;
#endif
	default:
		p = "unknown";
		break;
	}
	
	if(idx == -1)
		snprint(buf, len, "%s", p);
	else
		snprint(buf, len, "%s%d", p, idx);
}

uchar *
ip2arr(int addr)
{
	static uchar buf[4];

	buf[3] = addr >> 24;
	buf[2] = addr >> 16;
	buf[1] = addr >> 8;
	buf[0] = addr;
	return buf;
}

static char *
operstatus(int stat)
{
	static char buf[32];

	switch(stat){
	case IF_OPER_STATUS_NON_OPERATIONAL:
			return "disabled";
	case IF_OPER_STATUS_UNREACHABLE:
		return "unreachable";
	case IF_OPER_STATUS_DISCONNECTED:
		return "disconnected";
	case IF_OPER_STATUS_CONNECTING:
		return "connect in progress";
	case IF_OPER_STATUS_CONNECTED:
		return "connected, idle";
	case IF_OPER_STATUS_OPERATIONAL:
		return "up";
	default:
		snprint(buf, sizeof(buf), "'? %d'", stat);
		return buf;
	}
}


void
interfaces(Biobuf *bp)
{
    int i, j;
	char name[32];
	MIB_IPADDRROW *ap;
	MIB_IFROW *ip; 
    MIB_IPADDRTABLE *iptab;
	MIB_IFTABLE *iftab;
	uchar addr[IPaddrlen], net[IPaddrlen], mask[IPaddrlen];

	if((iftab = gettab(GetIfTable, sizeof(MIB_IFTABLE))) == nil)
		sysfatal("GetIfTable failed, %s", _winerr());

	if((iptab = gettab(GetIpAddrTable, sizeof(MIB_IPADDRTABLE))) == nil)
		sysfatal("GetIpAddrTable failed, %s", _winerr());

	for(i = 0; i < iftab->dwNumEntries; i++){	
		ip = &iftab->table[i];

		if(ip->dwOperStatus == IF_OPER_STATUS_NON_OPERATIONAL)
			continue;

		ap = nil;
		for(j = 0; j < iptab->dwNumEntries; j++)
			if(iptab->table[j].dwIndex == ip->dwIndex){
				ap = &iptab->table[j];
				break;
			}

		ifname(name, sizeof(name), ip->dwType, ip->dwIndex);
	
		if(ap){
			v4tov6(addr, ip2arr(ap->dwAddr));
			v4tov6(mask, ip2arr(ap->dwMask));
		}
		else{
			memset(addr, 0, sizeof(addr));
			memset(mask, 0, sizeof(mask));
		}

		memset(mask, 0xff, 10);
		maskip(addr, mask, net);

		Bprint(bp, "%-14s %5d %-16I %5M %-15I %8lud %8lud %8lud %8lud\n",
			name, ip->dwMtu, addr, mask, net,
			(ulong)(ip->dwInUcastPkts + ip->dwInNUcastPkts),
			(ulong)(ip->dwOutUcastPkts + ip->dwOutNUcastPkts),
			(ulong)(ip->dwInErrors + ip->dwInDiscards + ip->dwInUnknownProtos),
			(ulong)(ip->dwOutErrors + ip->dwOutDiscards));
	}

	free(iptab);
	free(iftab);
}

void
ethernics(Biobuf *bp)
{
    int i;
	char name[32];
	MIB_IFROW *ip;
	MIB_IFTABLE *iftab;

	if((iftab = gettab(GetIfTable, sizeof(MIB_IFTABLE))) == nil)
		sysfatal("GetIfTable failed, %s", _winerr());

	for(i = 0; i < iftab->dwNumEntries; i++){	
		ip = &iftab->table[i];

		if(ip->dwOperStatus == IF_OPER_STATUS_NON_OPERATIONAL)
			continue;

		ifname(name, sizeof(name), ip->dwType, ip->dwIndex);
		Bprint(bp, "%-14s %-5ld %E   %-12s   %s\n",
			name,
			ip->dwSpeed/1000000,
			ip->bPhysAddr, 
			operstatus(ip->dwOperStatus),
			ip->bDescr);
	}
	free(iftab);
}

void
usage(void)
{
	fprint(2, "usage: %s [-i] [-n] [-e] [-p proto]\n", argv0);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Owner *o;
	char *proto;
	WSADATA wsaData;
	Biobuf bout;

	fmtinstall('I', eipfmt);
	fmtinstall('V', eipfmt);
	fmtinstall('E', eipfmt);
	fmtinstall('M', eipfmt);

	Binit(&bout, 1, OWRITE);
	if(WSAStartup(MAKEWORD(1, 1), &wsaData))
		sysfatal("WSAStartup() failed, %s", _winerr());

	proto = nil;
	resolve = 1;
	ARGBEGIN{
	case 'i':
		interfaces(&bout);
		exits(nil);
		break;
	case 'e':
		ethernics(&bout);
		exits(nil);
	case 'n':
		resolve = 0;
		break;
	case 'p':
		proto = EARGF(usage());
		break;
	default:
		usage();
		break;
	}ARGEND;


	if(	_takepriv(SE_DEBUG_NAME) == -1)
		fprint(2, "%s: cannot gain debug priviledge, process owners elided\n", argv0);
	o = procowners();
	if(proto == nil || strcmp(proto, "tcp") == 0)
		tcpsess(&bout, o);
	if(proto == nil || strcmp(proto, "udp") == 0)
		udpsess(&bout, o);

    exits(nil);  
}

