#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

#include <setupapi.h>
#include <devguid.h>
#include <regstr.h>
#include <ddk/cfgmgr32.h>

enum {
	Nres = 16
};

int debug;

typedef struct Hinv Hinv;
typedef struct Mmap Mmap;
typedef struct Ment Ment;

struct Ment {		/* reserved memory segment */
	uvlong base;
	uvlong limit;
};
struct Mmap {		/* set of reserved segmenets, RAM or I/O */
	Ment *tab;
	int used;
	int alloc;
};

struct Hinv {
	struct {
		short bus;
		short device;
		short function;
		int class;
		int subclass;
		int iface;
		int vendorid;
		int deviceid;
		char *desc;
	} pci;
	int irq;
	int dma;
	Mmap io;
	Mmap mem;
};

struct {
	int type;
	char *desc;
	int class;
	int subclass;
	int iface;
	char *name;
} Pci[] = {
	{ 0x0000, "",					0x01, 0x00, 0x00, "scsi" },		/* disks */
	{ 0x0007, "IDE ATA/ATAPI controllers",		0x01, 0x00, 0x00, "ide" },
	{ 0x0000, "",					0x01, 0x02, 0x00, "flop" },
	{ 0x0000, "",					0x01, 0x04, 0x00, "raid" },
	{ 0x0012, "Network adapters",			0x02, 0x00, 0x00, "ether" },		/* networks */
	{ 0x0000, "",					0x02, 0x01, 0x00, "ring" },		/* the cambridge ring is dead (RIP) */
	{ 0x0000, "",					0x02, 0x02, 0x00, "fddi" },
	{ 0x0000, "",					0x02, 0x03, 0x00, "atm" },
	{ 0x0023, "Display adapters",			0x03, 0x00, 0x00, "vid" },		/* display */
	{ 0x001d, "Sound, video and game controllers",	0x04, 0x00, 0x00, "aud" },		/* audio */
	{ 0x0000, "",					0x05, 0x00, 0x00, "ram" },		/* memory */
	{ 0x0000, "",					0x05, 0x01, 0x00, "flash" },
	{ 0x002a, "System devices",			0x06, 0x00, 0x00, "host-pci" },		/* bridge */
	{ 0x0000, "",					0x06, 0x01, 0x00, "pci-isa" },
	{ 0x0000, "",					0x06, 0x02, 0x00, "pci-eisa" },
	{ 0x0000, "",					0x06, 0x03, 0x00, "pci-mca" },
	{ 0x0000, "",					0x06, 0x04, 0x00, "pci-pci" },
	{ 0x0000, "",					0x06, 0x05, 0x00, "pci-pcmcia" },
	{ 0x0000, "",					0x06, 0x06, 0x00, "pci-nubus" },
	{ 0x0000, "",					0x06, 0x07, 0x00, "pci-cardbus" },
	{ 0x001b, "",					0x07, 0x00, 0x00, "8050" },		/* serial */
	{ 0x0000, "",					0x07, 0x00, 0x01, "16450" },
	{ 0x0000, "",					0x07, 0x00, 0x02, "16550" },
	{ 0x0016, "",					0x07, 0x01, 0x00, "par" },		/* parallel */
	{ 0x0000, "",					0x07, 0x00, 0x01, "bi-par" },
	{ 0x0000, "",					0x07, 0x00, 0x02, "epc" },
	{ 0x0000, "",					0x08, 0x00, 0x00, "8259" },
	{ 0x0000, "",					0x08, 0x00, 0x01, "isa-pic" },
	{ 0x0000, "",					0x08, 0x00, 0x02, "eisa-pic" },
	{ 0x0000, "",					0x08, 0x01, 0x00, "8237" },
	{ 0x0000, "",					0x08, 0x01, 0x01, "isa-dma" },
	{ 0x0000, "",					0x08, 0x01, 0x02, "eisa-dma" },
	{ 0x0000, "",					0x08, 0x02, 0x00, "8254" },
	{ 0x0000, "",					0x08, 0x02, 0x01, "isa-timer" },
	{ 0x0000, "",					0x08, 0x02, 0x02, "eisa-timer" },
	{ 0x0000, "",					0x08, 0x03, 0x00, "rtc" },
	{ 0x0000, "",					0x08, 0x03, 0x01, "isa-rtc" },
	{ 0x0000, "",					0x08, 0x03, 0x02, "eisa-rtc" },
	{ 0x0000, "",					0x09, 0x00, 0x00, "kbd" },
	{ 0x000b, "",					0x09, 0x01, 0x00, "tablet" },
	{ 0x0000, "",					0x09, 0x02, 0x00, "mouse" },
	{ 0x000f, "",					0x0a, 0x00, 0x00, "dock" },
	{ 0x0000, "",					0x0b, 0x00, 0x00, "386" },
	{ 0x0000, "",					0x0b, 0x01, 0x00, "486" },
	{ 0x0000, "",					0x0b, 0x02, 0x00, "pentium" },
	{ 0x0000, "",					0x0b, 0x10, 0x00, "alpha" },
	{ 0x0000, "",					0x0b, 0x20, 0x00, "ppc" },
	{ 0x0000, "",					0x0b, 0x40, 0x00, "coprocessor" },
	{ 0x0000, "IEEE 1394 Bus host controllers",	0x0c, 0x00, 0x00, "1394" },
	{ 0x0000, "",					0x0c, 0x01, 0x00, "access" },
	{ 0x0000, "",					0x0c, 0x02, 0x00, "SSA" },
	{ 0x0000, "Universal Serial Bus controllers",	0x0c, 0x03, 0x00, "usb" },
	{ 0x0000, "",					0x0c, 0x05, 0x00, "smb" },
	{ 0x0000, "",					0x0d, 0x00, 0x00, "rad" },
	{ 0x0000, "",					0x10, 0x00, 0x00, "crypt" },
};

void
addmap(Mmap *mm, uvlong base, uvlong limit)
{
	if(mm->used >= mm->alloc){
		mm->alloc += 8;
		mm->tab = realloc(mm->tab, mm->alloc * sizeof(Ment));
		if(mm->tab == nil)
			sysfatal("No memory");
	}
	mm->tab[mm->used].base = base;
	mm->tab[mm->used].limit = limit;
	mm->used++;
}
	
int
numprop(HDEVINFO hdi, SP_DEVINFO_DATA *didp, int req)
{
	unsigned long val, got;

	val = 0;
	got = 0;
	if(SetupDiGetDeviceRegistryProperty(hdi, didp, req, nil, (void *)&val, sizeof(val), &got) == 0){
		if(debug)
			fprint(2, "%s: SetupDiGetDeviceRegistryProperty(%d) failed - %s\n", argv0, req, _winerr());
		return -1;
	}
	return val;
}

char *
strprop(HDEVINFO hdi, SP_DEVINFO_DATA *didp, int req)
{
	wchar_t buf[1024];

	if(SetupDiGetDeviceRegistryProperty(hdi, didp, req, nil, (BYTE *)buf, nelem(buf), nil) == 0){
		if(debug)
			fprint(2, "%s: SetupDiGetDeviceRegistryProperty(%d) failed - %s\n", argv0, req, _winerr());
		*buf = 0;
	}
	return _ninestr(buf);
}

void
getclass(Hinv *hi, HDEVINFO hdi, SP_DEVINFO_DATA *didp)
{
	char *tmp;
	int i, type;
	wchar_t buf[128];

	type = numprop(hdi, didp, SPDRP_DEVTYPE);
	for(i = 0; i < nelem(Pci); i++){
		if(Pci[i].type == type){
			hi->pci.class = Pci[i].class;
			hi->pci.subclass = Pci[i].subclass;
			hi->pci.iface = Pci[i].iface;
			return;
		}
	}

	/*
	 * This is utterly vile but there doesn't seem to be a better ay to do it.
	 */
	if(SetupDiGetClassDescription(&didp->ClassGuid, buf, nelem(buf), nil) == 0)
		*buf = 0;

	tmp = _ninestr(buf);
	for(i = 0; i < nelem(Pci); i++){
		if(strcmp(tmp, Pci[i].desc) == 0){
			hi->pci.class = Pci[i].class;
			hi->pci.subclass = Pci[i].subclass;
			hi->pci.iface = Pci[i].iface;
			free(tmp);
			return;
		}
	}

	hi->pci.class = -1;
	hi->pci.subclass = -1;
	hi->pci.iface = -1;
	if(debug)
		fprint(2, "'%s' - unknown device class\n", buf);
	free(tmp);
}


void
getvendor(Hinv *hi, HDEVINFO hdi, SP_DEVINFO_DATA *didp)
{
	int n;
	char *s, *a[8];

	s = strprop(hdi, didp, SPDRP_HARDWAREID);

	/*
	 * we are parsing strings of the form:
	 * PCI\VEN_8086&DEV_3A02&SUBSYS_04201028&REV_02
	 */
	if(strncmp(s, "PCI\\VEN_", 8) != 0){
		free(s);
		return;
	}
	if((n = gettokens(s, a, nelem(a), "_&")) != 8){
		if(debug)
			fprint(2, "getvendor: bad field count (%d != 8)\n", n);
		free(s);
	}

	hi->pci.vendorid = strtoul(a[1], nil, 16);
	hi->pci.deviceid = strtoul(a[3], nil, 16);
	free(s);
}

void
getposition(Hinv *hi, HDEVINFO hdi, SP_DEVINFO_DATA *didp)
{
	int n, bus;
	short addr;
	char *s, *a[8];

	addr = numprop(hdi, didp, SPDRP_ADDRESS);
	bus = numprop(hdi, didp, SPDRP_BUSNUMBER);
	if(bus != -1 && addr != -1){
		hi->pci.bus = bus;
		hi->pci.device = addr >> 16;
		hi->pci.function = addr;
		return;

	}

	/*
	 * somtimes SPDRP_BUSNUMBER and SPDRP_ADDRESS
	 * fail for no good reason so we fall back to 
	 * string parsing (ugh)
	 */
	s = strprop(hdi, didp, SPDRP_LOCATION_INFORMATION);

	/*
	 * we are parsing strings of the form:
	 * PCI bus 1, device 0, function 0
	 */
	if(strncmp(s, "PCI bus", 7) != 0){
		free(s);
		return;
	}
	if((n = gettokens(s, a, nelem(a), " \t,")) != 7){
		if(debug)
			fprint(2, "getposition: bad field count (%d != 7\n", n);
		free(s);
		return;
	}

	hi->pci.bus = atoi(a[2]);
	hi->pci.device = atoi(a[4]);
	hi->pci.function = atoi(a[6]);
	free(s);
}


void
getres(Hinv *hi, HDEVINFO hdi, SP_DEVINFO_DATA *didp)
{
	int i;
	void *buf;
	RES_DES ord;
	RES_DES rd;
	RESOURCEID got;
	unsigned long n, alloc;
	LOG_CONF cfg;
	IO_RESOURCE *io;
	MEM_RESOURCE *mem;
	DMA_RESOURCE *dma;
	IRQ_RESOURCE *irq;
	int logtypes[] = { ALLOC_LOG_CONF, FORCED_LOG_CONF, BOOT_LOG_CONF };
	RESOURCEID restypes[] = { ResType_Mem, ResType_IO, ResType_DMA, ResType_IRQ};

	USED(hdi);

	cfg = 0;
	for(i = 0; i < nelem(logtypes); i++)
		if(CM_Get_First_Log_Conf(&cfg, didp->DevInst, logtypes[i]) == CR_SUCCESS)
			break;
	if(i >= nelem(logtypes))
		return;
	buf = nil;
	alloc = 0;
	for(i = 0; i < nelem(restypes); i++){

		rd = 0;
		got = restypes[i];
		ord = (RES_DES)cfg;
		while(CM_Get_Next_Res_Des(&rd, ord, got, &got, 0) == CR_SUCCESS){
			if(ord != cfg) 
				CM_Free_Res_Des_Handle(ord);
	
			ord = rd;
			if(CM_Get_Res_Des_Data_Size(&n, rd, 0) != CR_SUCCESS)
				continue;
	
			if(n >= alloc){
				alloc = n + 64;
				if((buf = realloc(buf, alloc)) == nil)
					sysfatal("No memory, %r");
			}
			memset(buf, 0, alloc);

			if(CM_Get_Res_Des_Data(rd, buf, alloc, 0) != CR_SUCCESS)
				continue;
	
			switch(got){
			case ResType_Mem:
				mem = buf;
				addmap(&hi->mem, mem->MEM_Header.MD_Alloc_Base, mem->MEM_Header.MD_Alloc_End);
				break;
			case ResType_IO:
				io = buf;
				addmap(&hi->io, io->IO_Header.IOD_Alloc_Base, io->IO_Header.IOD_Alloc_End);
				break;
			case ResType_DMA:
				dma = buf;
				hi->dma = dma->DMA_Header.DD_Alloc_Chan;
				break;
			case ResType_IRQ:
				irq = buf;
				hi->irq = irq->IRQ_Header.IRQD_Alloc_Num;
				break;
			default:
				break;
			}
		}
		if(ord != cfg)
			CM_Free_Res_Des_Handle(ord);
	}
	free(buf);
	CM_Free_Log_Conf_Handle(cfg);
}

Hinv *
hinv(void)
{
	HDEVINFO hdi;
	Hinv *base, *hi;
	SP_DEVINFO_DATA did;
	int i, num, alloc;

	if((hdi = SetupDiGetClassDevs(nil, nil, 0, DIGCF_ALLCLASSES)) == INVALID_HANDLE_VALUE)
		sysfatal("SetupDiGetClassDevs() failed, %s", _winerr());

	i = 0;
	num = 0;
	alloc = 0;
	base = nil;
	did.cbSize = sizeof(did);
	while(SetupDiEnumDeviceInfo(hdi, i++, &did)){
		if(num >= (alloc-1)){			/* -1 as we must allow a placeholder for the end of the list */
			alloc += 8;
			if((base = realloc(base, alloc * sizeof(Hinv))) == nil)
				sysfatal("No memory, %r");
		}
		hi = &base[num++];
		memset(hi, 0, sizeof(Hinv));

		getposition(hi, hdi, &did);
		getvendor(hi, hdi, &did);
		getres(hi, hdi, &did);
		getclass(hi, hdi, &did);
		hi->pci.desc = strprop(hdi, &did, SPDRP_DEVICEDESC);
	}
	if(GetLastError() != NO_ERROR && GetLastError() != ERROR_NO_MORE_ITEMS)
		sysfatal("SetupDiEnumDeviceInfo() failed, %s", _winerr());

	SetupDiDestroyDeviceInfoList(hdi);

	base[num].irq = -1;		/* end of list marker */
	return base;
}

static void
usage(void)
{
	fprint(2, "usage: %s [-d] [-a] [-v [-v]]\n", argv0);
	fprint(2, " -a    all - even fake windows pci devices\n");
	fprint(2, " -v    verbose\n");
	fprint(2, " -d    debug\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Hinv *hi;
	char buf[32];
	int i, all, verbose;
	Biobuf bout;

	all = 0;
	verbose = 0;
	ARGBEGIN{
	case 'a':
		all++;
		break;
	case 'd':
		debug++;
		break;
	case 'v':
		verbose++;
		break;
	default:
		usage();
	}ARGEND;

	if(argc != 0)
		usage();

	Binit(&bout, 1, OWRITE);

	hi = hinv();
	for(; hi && hi->irq != -1; hi++){

		if(! all && (hi->pci.vendorid == 0 || hi->pci.deviceid == 0))
			continue;

		snprint(buf, sizeof(buf), "%d.%d.%d", hi->pci.bus, hi->pci.device, hi->pci.function);

		Bprint(&bout, "%-8s %3d %3d %04x/%04x %2d\n",
			buf,
			hi->pci.class, hi->pci.subclass,
			hi->pci.vendorid, hi->pci.deviceid,
			hi->irq);

		if(verbose > 1){
			if(hi->dma)
				Bprint(&bout, "\tdma=%2d\n", hi->dma);

			for(i = 0; i < hi->io.used; i++)
				Bprint(&bout, "\tio 0x%llx - 0x%llx\n",
					hi->io.tab[i].base, hi->io.tab[i].limit);

			for(i = 0; i < hi->mem.used; i++)
				Bprint(&bout, "\tmem 0x%llx - 0x%llx\n",
					hi->mem.tab[i].base, hi->mem.tab[i].limit);
		}

		if(verbose > 0)
			Bprint(&bout, "\t%s\n", hi->pci.desc);
	}
	exits(nil);
}

