/*
 * Plan 9 versions of system-specific functions
 *	By convention, exported routines herein have names beginning with an
 *	upper case letter.
 */
#define _SPAWN_EXTENSION
#define _GETENVP_EXTENSION
#include "rc.h"
#include "exec.h"
#include "io.h"
#include "fns.h"
#include "getflags.h"

#include <ctype.h>

char *Signame[] = {
	"sigexit",	"sighup",	"sigint",	"sigquit",
	"sigalrm",	"sigkill",	"sigfpe",	"sigterm",
	nil
};

char *Rcmain = "undefined";

void execfinit(void);
void execnewpgrp(void);

builtin Builtin[] = {
	"cd",		execcd,
	"whatis",	execwhatis,
	"eval",		execeval,
	"exec",		execexec,	/* but with popword first */
	"exit",		execexit,
	"shift",	execshift,
	"wait",		execwait,
	".",		execdot,
	"finit",	execfinit,
	"flag",		execflag,
	"rfork",	execnewpgrp,
	nil
};

char *syssigname[] = {
	"exit",		/* can't happen */
	"hangup",
	"interrupt",
	"quit",		/* can't happen */
	"alarm",
	"kill",
	"sys: fp: ",
	"term",
	0
};

#define	SEP	';'
char **environp;

/*
 * Windows preserves case but ignores it, even in environment variables.
 * We need to be able to find these vars so we force their
 * case to that below as they are parsed from the environment.
 */
char *Force[] = {
	"path",
	"pathext",
	nil
};


struct word*
enval(char *s)
{
	char *t, c;
	struct word *v;

	for(t = s;*t && *t != SEP;t++)
		continue;

	c=*t;
	*t='\0';
	v = newword(s, c=='\0'?(struct word *)0:enval(t+1));
	*t = c;
	return v;
}

int
cmpvar(void *aa, void *ab)
{
	struct var * const *a = aa, * const *b = ab;

	return strcmp((*a)->name, (*b)->name);
}

char *
exportenv(void)
{
	char *env, *p, *q;
	struct var **h, *v, **idx;
	struct word *a;
	int nvar = 0, nchr = 0, sep, i;

	/*
	 * Slightly kludgy loops look at locals then globals.
	 * locals no longer exist - geoff
	 */
	for(h = gvar-1; h != &gvar[NVAR]; h++)
	for(v = h >= gvar? *h: runq->local; v ;v = v->next){
		if(v==vlook(v->name) && v->val){
			nvar++;
			nchr+=strlen(v->name)+1;
			for(a = v->val;a;a = a->next)
				nchr+=strlen(a->word)+1;
		}
		if(v->fn){
			nvar++;
			nchr+=strlen(v->name)+strlen(v->fn[v->pc-1].s)+8;
		}
	}

	idx = (struct var **)emalloc(nvar * sizeof(struct var *));

	i = 0;
	for(h = gvar-1; h != &gvar[NVAR]; h++)
	for(v = h >= gvar? *h: runq->local; v ;v = v->next){
		if((v==vlook(v->name) && v->val) || v->fn)
			idx[i++] = v;
	}

	qsort((void *)idx, nvar, sizeof idx[0], cmpvar);

	env = (char *)emalloc(nvar+nchr+1);

	p = env;
	for(i = 0; i < nvar; i++){
		v = idx[i];
		if((v==vlook(v->name)) && v->val){
			q = v->name;
			while(*q) *p++=*q++;
			sep='=';
			for(a = v->val;a;a = a->next){
				*p++=sep;
				sep = SEP;
				q = a->word;
				while(*q) *p++=*q++;
			}
			*p++='\0';
		}
		if(v->fn){
			*p++='f'; *p++='n'; *p++='#';
			q = v->name;
			while(*q) *p++=*q++;
			*p++='=';
			q = v->fn[v->pc-1].s;
			while(*q) *p++=*q++;
			*p++='\0';
		}
	}
	*p = 0;

	return env;	
}

void
execnewpgrp(void)
{
	int arg;
	char *s;
	switch(count(runq->argv->words)){
	case 1:
		arg = RFENVG|RFNAMEG|RFNOTEG;
		break;
	case 2:
		arg = 0;
		for(s = runq->argv->words->next->word;*s;s++) switch(*s){
		default:
			goto Usage;
		case 'n':
			arg|=RFNAMEG;  break;
		case 'N':
			arg|=RFCNAMEG;
			break;
		case 'm':
			arg|=RFNOMNT;  break;
		case 'e':
			arg|=RFENVG;   break;
		case 'E':
			arg|=RFCENVG;  break;
		case 's':
			arg|=RFNOTEG;  break;
		case 'f':
			arg|=RFFDG;    break;
		case 'F':
			arg|=RFCFDG;   break;
		}
		break;
	default:
	Usage:
		pfmt(err, "Usage: %s [fnesFNEm]\n", runq->argv->words->word);
		setstatus("rfork usage");
		poplist();
		return;
	}
	if(rfork(arg)==-1){
		pfmt(err, "rc: %s failed\n", runq->argv->words->word);
		setstatus("rfork failed");
	}
	else
		setstatus("");
	poplist();
}

void
Vinit(void)
{
	char *s;
	struct word *plan9;
	char **env = _getenvp();
	environp = env;

	for(;*env;env++){
		for(s=*env;*s && *s!='(' && *s!='=';s++);
		switch(*s){
		case '\0':
			pfmt(err, "environment %q?\n", *env);
			break;
		case '=':
			*s='\0';
			setvar(*env, enval(s+1));
			*s='=';
			break;
		case '(':	/* ignore functions for now */
			break;
		}
	}

	if((plan9 = vlook("plan9")->val))
		Rcmain = smprint("%s/rc/lib/rcmain", plan9->word);
}

static char **envp;

void
Xrdfn(void)
{
	int len;
	char *s, *e, *x;

	while(envp && *envp){
		e = *(envp++);
		if(strncmp(e, "fn#", 3) != 0)		/* ignore variables */
			continue;
		if((s = strchr(e, '=')) == nil)
			continue;

		x = smprint("fn %s\n", s+1);
		len = strlen(x);
		execcmds(opencore(x, len+1));
		free(x);
		return;
	}
	Xreturn();
}



union code rdfns[4];

void
execfinit(void)
{
	static int first = 1;
	if(first){
		rdfns[0].i = 1;
		rdfns[1].f = Xrdfn;
		rdfns[2].f = Xjump;
		rdfns[3].i = 1;
		first = 0;
	}
	Xpopm();
	envp = environp;
	start(rdfns, 1, runq->local);
}

/* NB: return -1 only on interrupt, may cause a retry */
int
Waitfor(int pid, int persist)
{
	thread *p;
	Waitmsg *w;
	char errbuf[ERRMAX];

	if(pid >= 0 && !havewaitpid(pid))
		return 0;

	while((w = wait()) != nil){
		delwaitpid(w->pid);
		if(w->pid==pid){
			setstatus(w->msg);
			free(w);
			return 0;
		}
		for(p = runq->ret;p;p = p->ret)
			if(p->pid==w->pid){
				p->pid=-1;
				strcpy(p->status, w->msg);
			}
		free(w);
	}

	errstr(errbuf, sizeof errbuf);
	if(strcmp(errbuf, "interrupted") == 0)
		return -1;
	return 0;
}

void
Updenv(void)
{
}

void
Execute(word *args, word *path)
{
	pfmt(err, "rc: exec not supported under 9win\n");
}

int
Abspath(char *w)
{
	if(strncmp(w, "./", 2)==0)
		return 1;
	if(strncmp(w, ".\\", 2)==0)
		return 1;
	if(strncmp(w, "../", 3)==0)
		return 1;
	if(strncmp(w, "..\\", 3)==0)
		return 1;
	if(strncmp(w, "/", 1)==0)
		return 1;
	if(strncmp(w, "\\", 1)==0)
		return 1;
	if(isalpha(w[0]) && w[1] == ':') 
		return 1;
	return 0;
}

#define	NDIR	256		/* shoud be a better way */


int
Globsize(char *p)
{
	int isglob = 0, globlen = NDIR+1;
	for(;*p;p++){
		if(*p==GLOB){
			p++;
			if(*p!=GLOB)
				isglob++;
			globlen+=*p=='*'?NDIR:1;
		}
		else
			globlen++;
	}
	return isglob?globlen:0;
}

#define	NFD	50

struct{
	Dir	*dbuf;
	int	i;
	int	n;
}dir[NFD];

int
Opendir(char *name)
{
	Dir *db;
	int f;
	f = open(name, 0);
	if(f==-1)
		return f;
	db = dirfstat(f);
	if(db!=nil && (db->mode&DMDIR)){
		if(f<NFD){
			dir[f].i = 0;
			dir[f].n = 0;
		}
		free(db);
		return f;
	}
	free(db);
	close(f);
	return -1;
}

static int
trimdirs(Dir *d, int nd)
{
	int r, w;

	for(r=w=0; r<nd; r++)
		if(d[r].mode&DMDIR)
			d[w++] = d[r];
	return w;
}

/*
 * onlydirs is advisory -- it means you only
 * need to return the directories.  it's okay to
 * return files too (e.g., on unix where you can't
 * tell during the readdir), but that just makes 
 * the globber work harder.
 */
int
Readdir(int f, void *p, int onlydirs)
{
	int n;

	if(f<0 || f>=NFD)
		return 0;
Again:
	if(dir[f].i==dir[f].n){	/* read */
		free(dir[f].dbuf);
		dir[f].dbuf = 0;
		n = dirread(f, &dir[f].dbuf);
		if(n>0){
			if(onlydirs){
				n = trimdirs(dir[f].dbuf, n);
				if(n == 0)
					goto Again;
			}	
			dir[f].n = n;
		}else
			dir[f].n = 0;
		dir[f].i = 0;
	}
	if(dir[f].i == dir[f].n)
		return 0;
	strcpy(p, dir[f].dbuf[dir[f].i].name);
	dir[f].i++;
	return 1;
}

void
Closedir(int f)
{
	if(f>=0 && f<NFD){
		free(dir[f].dbuf);
		dir[f].i = 0;
		dir[f].n = 0;
		dir[f].dbuf = 0;
	}
	close(f);
}

volatile int interrupted = 0;

void
notifyf(void *unused, char *s)
{
	int i;
	for(i = 0;syssigname[i];i++)
		if(strncmp(s, syssigname[i], strlen(syssigname[i]))==0){
			if(strncmp(s, "sys: ", 5)!=0)
				interrupted = 1;
			goto Out;
		}
	pfmt(err, "rc: note: %s\n", s);
	noted(NDFLT);
	return;
Out:
	if(strcmp(s, "interrupt")!=0 || trap[i]==0){
		trap[i]++;
		ntrap++;
	}
	if(ntrap>=32){	/* rc is probably in a trap loop */
		pfmt(err, "rc: Too many traps (trap %s), aborting\n", s);
		abort();
	}
	noted(NCONT);
}

void
Trapinit(void)
{
	notify(notifyf);
}

void
Unlink(char *name)
{
	remove(name);
}

long
Write(int fd, void *buf, long cnt)
{
	return write(fd, buf, cnt);
}

long
Read(int fd, void *buf, long cnt)
{
	int n;

	n = read(fd, buf, cnt);
	if(interrupted)
		return -1;
	return n;
}

long
Seek(int fd, long cnt, long whence)
{
	return seek(fd, cnt, whence);
}


/* used by whatis only */
int
Executable(char *file)
{
	int fd, n;
	char buf[4];

	if(access(file, AEXEC) == 0)
		return 1;

/* FIXME: maybe this logic should go into libc/win32/fileinfo.c */
	if((fd = open(file, OREAD)) == -1)
		return 0;
	n = read(fd, buf, sizeof(buf));
	close(fd);
	if(n <= 0)
		return 0;
	buf[n] = 0;

	if(strncmp(buf, "#!", 2) == 0)
		return 1;
	return 0;
}

int
Creat(char *file)
{
	return create(file, ORDWR, 0644);
}

int
Dup(int a, int b)
{
	return dup(a, b);
}

int
Dup1(int a)
{
	return dup(a, -1);
}

void
Exit(char *stat)
{
	exits(stat);
}

int
Xeintr(void)
{
	return interrupted;
}

void
Noerror(void)
{
	interrupted = 0;
}
 
int
Isatty(int fd)
{
	char buf[64];

	if(fd2path(fd, buf, sizeof buf) != 0)
		return 0;

	/* might be #c/cons during boot - fixed 22 april 2005, remove this later */
	if(strcmp(buf, "#c/cons") == 0)
		return 1;
	if(strcmp(buf, "/dev/stdin") == 0)		// fd2path should return somthing better
		return 1;

	/* might be /mnt/term/dev/cons */
	return strlen(buf) >= 9 && strcmp(buf+strlen(buf)-9, "/dev/cons") == 0;
}

char **
mkargv(struct word *a)
{
	int n;
	char **argv, **argp;

	n = count(a)+2;
	if(n < 8)			/* plenty of room for shebangs */
		n = 8;
	argv = (char **)emalloc(n*sizeof(char *));
	memset(argv, 0, n*sizeof(char *));


	argp = argv+1;	/* leave one at front for runcoms */
	for(;a;a = a->next)
		*argp++=a->word;
	*argp = 0;
	return argv;
}

int
ForkExecute(char *name, char **argv, int sin, int sout, int serr)
{
	int pid;

int i;
pfmt(err, "%s\n", name);
for(i = 0; argv[i] != nil; i++)
	pfmt(err, "	%d %s\n", i, argv[i]);
pfmt(err, "			%p %p %p\n", sin, sout, serr);

	if(rfork(RFFDG) == -1){
		pfmt(err, "rfork() failed %r\n");
		return -1;
	}
	if(dup(0, sin) < 0){
		pfmt(err, "dup stdin failed %r\n");
		return -1;
	}
	if(dup(1, sout) < 0){
		pfmt(err, "dup stdout failed %r\n");
		return -1;
	}
	if(dup(2, serr) < 0){
		pfmt(err, "dup stderr failed %r\n");
		return -1;
	}
	if((pid = _spawn(name, argv)) < 0){
		pfmt(err, "%s spawn failed %r\n", name);
		return -1;
	}
	return pid;
}

void
Abort(void)
{
	pfmt(err, "aborting\n");
	flush(err);
	Exit("aborting");
}

void
Memcpy(void *a, void *b, long n)
{
	memmove(a, b, n);
}

void *
Malloc(ulong n)
{
	return malloc(n);
}

int *waitpids;
int nwaitpids;

void
addwaitpid(int pid)
{
	waitpids = realloc(waitpids, (nwaitpids+1)*sizeof waitpids[0]);
	if(waitpids == 0)
		panic("Can't realloc %d waitpids", nwaitpids+1);
	waitpids[nwaitpids++] = pid;
}

void
delwaitpid(int pid)
{
	int r, w;
	
	for(r=w=0; r<nwaitpids; r++)
		if(waitpids[r] != pid)
			waitpids[w++] = waitpids[r];
	nwaitpids = w;
}

void
clearwaitpids(void)
{
	nwaitpids = 0;
}

int
havewaitpid(int pid)
{
	int i;

	for(i=0; i<nwaitpids; i++)
		if(waitpids[i] == pid)
			return 1;
	return 0;
}

/* avoid loading any floating-point library code */
int
_efgfmt(Fmt *unused)
{
	return -1;
}
