#include <u.h>
#include <libc.h>
#include <mswin.h>

#include <userenv.h>
#include <lm.h>

int
remote(wchar_t *old, wchar_t *new, int perm)
{
	int e;
	int flags;

	flags = 0;
	if(perm)
		flags = CONNECT_UPDATE_PROFILE;

	if((e = WNetCancelConnection2(new, flags, TRUE)) != NO_ERROR){
		werrstr("WNetCancelConnection2 - %s", _wneterr(e));
		return -1;
	}
	return 0;
}

int
local(wchar_t *old, wchar_t *new, int perm)
{
	wchar_t dev[MAX_PATH];
	int flags = DDD_RAW_TARGET_PATH | DDD_REMOVE_DEFINITION | DDD_EXACT_MATCH_ON_REMOVE;

	if(perm){
		if(DeleteVolumeMountPoint(new) != 0){
			werrstr("DeleteVolumeMountpoint - %s", _winerr());
			return -1;
		}
	}
	else{
		if(old == nil){
			if(QueryDosDevice(new, dev, MAX_PATH) == 0){
				werrstr("DeleteVolumeMountpoint - %s", _winerr());
				return -1;
			}
			old = dev;
		}

		if(DefineDosDevice(flags, new, old) == 0){
			werrstr("DefineDosDevice - %s", _winerr());
			return -1;
		}
	}
	return 0;
}

void
usage(void)
{
	fprint(2, "usage: %s [-r] old new\n", argv0);
	fprint(2, "	-r	make change permanant in the registry\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int rc, type, perm;
	wchar_t *old, *new;

	perm = 0;
	ARGBEGIN{
	case 'r':
		perm = 1;
		break;
	default:
		usage();
		break;
	}ARGEND;

	new = nil;
	old = nil;
	switch(argc){
	case 1:
		old = _winpath(argv[0]);
		break;
	case 2:
		old = _winpath(argv[0]);
		new = _winpath(argv[1]);
		break;
	default:
		usage();
	}

	rc = 0;
	type = GetDriveType(old);
	switch(type){
	case DRIVE_NO_ROOT_DIR:
		sysfatal("%s 'no root directory' unknown drive type", old);
		break;
	case DRIVE_REMOVABLE:
	case DRIVE_FIXED:
	case DRIVE_CDROM:
	case DRIVE_RAMDISK:
		rc = local(old, new, perm);
		break;
	case DRIVE_REMOTE:
		rc = remote(old, new, perm);
		break;
	case DRIVE_UNKNOWN:
	default:
		sysfatal("%s %d unknown drive type", old, type);
	}

	if(rc == -1)
		sysfatal("%q %q, %r", old, new);

	free(old);
	free(new);
	exits(nil);
}
