#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>

#include <mswin.h>

#include <winldap.h>
#include <winber.h>
#include <lmaccess.h>
#include <security.h>
#include <secext.h>

Biobuf out;
int Maxage;
int Debug;

enum {
	Tstr,
	Tdate,
	Tbytes,
	Tcn,
	Tuac,
	Tforename,
	Tsurname,
	Texpire
};

typedef struct {
	int type;
	char *name;
	wchar_t *attr;
	char **values;
	int nvals;
} Val;

void dump(void);
char *uac(char *);
void prcn(Val *);
char *prdate(char *, int);
void cpvals(Val *, int, wchar_t **);

Val fetch[] = {
	{ Tstr,			"uid:",				L"mailNickname" },
	{ Tforename,	"Forename:",		L"givenName" },
	{ Tsurname,		"Surname:",			L"sn" },
	{ Tstr,			"Title:",			L"title" },
	{ Tstr,			"Dep:",				L"department" },
	{ Tcn,			"Manager:",			L"manager" },
	{ Tstr,			"Company:",			L"company" },
	{ Tstr,			"Tel:",				L"telephoneNumber" },
	{ Tstr,			"Fax:",				L"facsimileTelephoneNumber" },
	{ Tstr,			"Misc:",			L"description" },
	{ Tstr,			"Mob:",				L"mobile" },
	{ Tstr,			"Employee:",		L"employeeID" },
	{ Tstr,			"Site:",			L"physicalDeliveryOfficeName" },
	{ Tstr,			"Email:",			L"mail" },
	{ Tstr,			"Home:",			L"homeDirectory" },
	{ Tstr,			"Terminal:",		L"extensionAttribute10" },
	{ Tstr,			"Logins:",			L"logonCount" },
	{ Tdate,		"Last login:",		L"lastLogon" },
	{ Tstr,			"Bad passwd:",		L"badPwdCount" },
	{ Texpire,		"Passwd expiry:",	L"pwdLastSet" },
	{ Tuac,			"Controls:",		L"userAccountControl" },
	{ Tcn,			"Group:",			L"memberOf" },

	{ Tstr,			"Unix name:",		L"msSFU30Name" },
	{ Tstr,			"Unix pass:",		L"msSFU30Password" },
	{ Tstr,			"Unix uid:",		L"msSFU30UidNumber" },
	{ Tstr,			"Unix gid:",		L"msSFU30GidNumber" },
	{ Tstr,			"Unix shell:",		L"msSFU30LoginShell" },
	{ Tstr,			"Unix Home:",		L"msSFU30HomeDirectory" },
	{ Tcn,			"Unix group:",		L"msSFU30PosixMemberOf" },


	{ 0, nil, nil },
};

void
cpvals(Val * v, int n, wchar_t **values)
{
	int i;

	if((v->values = malloc(n * sizeof(char *))) == nil)
		sysfatal("No memory, %r");
	for(i = 0; i < n; i++)
		v->values[i] = _ninestr(values[i]);
	v->nvals = n;
}

char *
strlwr(char *buf)
{
	char *p;

	for(p = buf; *p; p++)
		if(isupper(*p))
			*p = tolower(*p);
	return buf;
}

char *
uac(char *str)
{
	int i;
	Fmt f;
	long val;

	struct {
		int n;
		char *s;
	} tab[] = {
		{ UF_NORMAL_ACCOUNT, 				"normal account" },
		{ UF_TEMP_DUPLICATE_ACCOUNT, 		"temporary duplicate account" },
		{ UF_WORKSTATION_TRUST_ACCOUNT,		"workstation trust account" },
		{ UF_SERVER_TRUST_ACCOUNT, 			"server trust account" },
		{ UF_INTERDOMAIN_TRUST_ACCOUNT,		"interdomain trust account" },
		{ UF_SCRIPT, 						"logon script executed" },
		{ UF_ACCOUNTDISABLE,				"account disabled" },
		{ UF_HOMEDIR_REQUIRED, 				"home dir required" },
		{ UF_PASSWD_NOTREQD, 				"passwd not required" },
		{ UF_PASSWD_CANT_CHANGE, 			"cannot change password" },
		{ UF_LOCKOUT, 						"locked out" },
		{ UF_DONT_EXPIRE_PASSWD, 			"passwd does not expire" }
	};

	fmtstrinit(&f);
	val = atol(str);
	for(i = 0; i < sizeof(tab) / sizeof(tab[0]); i++)
		if((val & tab[i].n) == tab[i].n)
			fmtprint(&f, "%q  ", tab[i].s);
	return fmtstrflush(&f);
}

char *
prdate(char *str, int off)
{
	Fmt f;
	long t, d;
	char *p, *s;
	long long vl;

	vl = strtoull(str, nil, 0);
	if(vl == 0)
		return strdup("never");

	vl = vl / 10000000LL;
	vl = vl - 11644473600LL;
	vl -= off;
	d = (vl - time(nil)) / (60 * 60 * 24);

	fmtstrinit(&f);
	if(d == 0)
		fmtprint(&f, "today");
	else if(d > 0 && d < 30)
		fmtprint(&f, "in %ld days", d);
	else if(d < 0 && d > -30)
		fmtprint(&f, "%ld days ago", -d);
	else {
		t = vl;
		s = ctime(t);
		if((p = strchr(s, '\n')) != nil)
			*p = 0;
		fmtprint(&f, "%s", s);
	}
	return fmtstrflush(&f);
}

void
prcn(Val * v)
{
	int i, found;
	char *p, *q;

	Bprint(&out, "%-16s ", v->name);

	found = 0;
	for(i = 0; i < v->nvals; i++){
		if(v->values[i] && (p = strstr(v->values[i], "CN=")) != nil){
			p += 3;
			if((q = strchr(p, ',')) != nil)
				*q = 0;
			if(i > 0)
				Bprint(&out, "%-16s ", v->name);
			Bprint(&out, "%q\n", p);
			found++;
		}
	}
	if(!found)
		Bprint(&out, "\n");

}

void
dump(void)
{
	Val *v;
	char *s;
	int i, found;

	for(v = fetch; v->attr; v++){
		if(v->nvals == 0)
			continue;

		switch(v->type){
		case Tcn:
			prcn(v);
			break;
		case Tdate:
		case Texpire:
			found = 0;
			Bprint(&out, "%-16s ", v->name);
			for(i = 0; i < v->nvals; i++){
				if(v->values[i]){
					if(v->type == Tdate)
						s = prdate(v->values[i], 0);
					else
						s = prdate(v->values[i], Maxage);
					if(s == nil)
						sysfatal("No memory, %r");
					Bprint(&out, "%s\n", s);
					free(s);
					found++;
				}
				if(i < v->nvals - 1)
					Bprint(&out, "%-16s ", "");
			}
			if(!found)
				Bprint(&out, "\n");
			break;
		case Tuac:
			found = 0;
			Bprint(&out, "%-16s ", v->name);
			for(i = 0; i < v->nvals; i++){
				if(v->values[i]){
					if((s = uac(v->values[i])) == nil)
						sysfatal("No memory, %r");
					Bprint(&out, "%s\n", s);
					free(s);
					found++;
				}
				if(i < v->nvals - 1)
					Bprint(&out, "%-16s ", "");
			}
			if(!found)
				Bprint(&out, "<none>\n");
			break;
		case Tstr:
			/* FALLTHRU */
		default:
			found = 0;
			Bprint(&out, "%-16s ", v->name);
			for(i = 0; i < v->nvals; i++){
				if(v->values[i]){
					Bprint(&out, "%s\n", v->values[i]);
					found++;
				}
				if(i < v->nvals - 1)
					Bprint(&out, "%-16s ", "");
			}
			if(!found)
				Bprint(&out, "\n");
			break;
		}
	}
	for(v = fetch; v->attr; v++)
		for(i = 0; i < v->nvals; i++){
			free(v->values[i]);
			v->values[i] = nil;
		}
	Bprint(&out, "\n");

}

int
extract(LDAP *con, LDAPMessage *res, int num)
{
	Val *v;
	int i, j, n;
	wchar_t *attr, **val;
	BerElement *ber = nil;
	LDAPMessage *ent = nil;

	ent = ldap_first_entry(con, res);
	for(i = 0; i < num; i++){

		attr = ldap_first_attribute(con, ent, &ber);
		while(attr){
			if((val = ldap_get_values(con, ent, attr)) == nil){
				Bprint(&out, "<no value>");
				continue;
			}
			if((n = ldap_count_values(val)) == 0){
				Bprint(&out, "<bad value>");
				ldap_value_free(val);
				continue;
			}

			if(Debug)
				for(j = 0; j < n; j++)
					fprint(2, "%-24w %w\n", attr, val[j]);

			else
				for(v = fetch; v->attr; v++)
					if(_wincmp(v->attr, attr) == 0)
						cpvals(v, n, val);

			ldap_value_free(val);
			ldap_memfree(attr);
			attr = ldap_next_attribute(con, ent, ber);
		}

		ber_free(ber, 0);
		ent = ldap_next_entry(con, ent);
		dump();
		Bflush(&out);
	}
	return 0;

}

int
search(LDAP *con, wchar_t *dn, char *name)
{
	Val *v;
	wchar_t *tmp;
	char filt[255];
	int rc, e, n, i;
	wchar_t **fields;
	LDAPMessage *res;

	rc = 0;
	if((fields = malloc(sizeof(wchar_t *) * nelem(fetch))) == nil)
		sysfatal("No memory, %r");
	if(Debug){
		fields[0] = L"*";
		fields[1] = nil;
	}
	else{
		for(i = 0, v = fetch; v->attr; v++, i++)
			fields[i] = v->attr;
		fields[i] = nil;
	}

	snprint(filt, sizeof(filt), "(&(objectCategory=person)(objectClass=user)(cn=%s))", name);
	tmp = _winstr(filt);

	if((e = ldap_search_s(con, dn, LDAP_SCOPE_SUBTREE, tmp, fields, 0, &res)) != LDAP_SUCCESS){
		Bprint(&out, "%s: %q - search failed, %w\n", argv0, name, ldap_err2string(e));
		rc = -1;
		goto done;
	}

	if((n = ldap_count_entries(con, res)) < 1){
		Bprint(&out, "%s: %q - not found\n", argv0, name);
		rc = -1;
		goto done;
	}
	rc = extract(con, res, n);

done:
	free(tmp);
	ldap_msgfree(res);
	return rc;
}


int
maxpasswdage(LDAP * con, wchar_t *dn)
{
	char *tmp;
	double maxage;
	int n, e, i, j;
	BerElement *ber;
	LDAPMessage *ent, *res;
	wchar_t *fields[2], *attr, **val, *filt;

	ber = nil;
	filt = L"objectclass=domain";
	fields[0] = L"maxPwdage";
	fields[1] = nil;

	if((e = ldap_search_s(con, dn, LDAP_SCOPE_SUBTREE, filt, fields, 0, &res)) != LDAP_SUCCESS){
		fprint(2, "%s: max passwd age: ldap_search_s() failed, %w\n", argv0, ldap_err2string(e));
		return 0;
	}

	if((n = ldap_count_entries(con, res)) < 1){
		fprint(2, "%s: max passwd age: ldap_count_entries() failed - %s\n", argv0, _winerr());
		ldap_msgfree(res);
		return 0;
	}

	maxage = 0.;
	for(i = 0; i < n; i++){
		ent = ldap_first_entry(con, res);
		if((attr = ldap_first_attribute(con, ent, &ber)) == nil){
			fprint(2, "%s: max passwd age: ldap_first_attribute() failed - %s\n", argv0, _winerr());
			ldap_msgfree(res);
			return 0;
		}
		if((val = ldap_get_values(con, ent, attr)) == nil){
			fprint(2, "%s: max passwd age: ldap_get_values() failed - %s\n", argv0, _winerr());
			ldap_memfree(attr);
			ldap_msgfree(res);
			return 0;
		}

		for(j = 0; j < n; j++)
			if(val[j]){
				tmp = _ninestr(val[j]);
				maxage = atof(tmp);
				free(tmp);
				if(maxage != 0.0)
					break;
			}

		ldap_value_free(val);
		ldap_memfree(attr);
	}
	if(maxage == 0.0)
		return 0;
	maxage = maxage / 10e6;
	return maxage;
}

void
usage(void)
{
	fprint(2, "%s: <forename> <surname>\n", argv0);
	fprint(2, "  partial names will print every match found\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int e, rc;
	ULONG sz;
	LDAP *con;
	unsigned long val;
	char *host, name[128];
	wchar_t *whost, *dc, fqdn[1024];

	host = nil;
	ARGBEGIN {
	case 'h':
		host = EARGF(usage());
		break;
	case 'd':
		Debug++;
		break;
	default:
		usage();
		break;
	}ARGEND;

	switch(argc){
	case 1:
		snprint(name, sizeof(name), "*%s*", argv[0]);
		break;
	case 2:
		snprint(name, sizeof(name), "*%s%s*", argv[0], argv[1]);
		break;
	default:
		usage();
		break;
	}

	Binit(&out, 1, OWRITE);
	quotefmtinstall();
	_widefmtinstall();

	if(host)
		whost = _winstr(host);
	else
		whost = nil;

	if((con = ldap_init(whost, LDAP_PORT)) == nil)
		sysfatal("ldap init failed, %s", _winerr());

	val = LDAP_VERSION3;
	if((e = ldap_set_option(con, LDAP_OPT_PROTOCOL_VERSION, (void *)&val)) != LDAP_SUCCESS){
		ldap_unbind(con);
		sysfatal("ldap set protocol version option failed, %w", ldap_err2string(e));
	}

	if((e = ldap_connect(con, nil)) != LDAP_SUCCESS){
		ldap_unbind(con);
		sysfatal("%s ldap connect failed, %w", ((host)? host: "<devault host>"), ldap_err2string(e));
	}

	if((e = ldap_bind_s(con, nil, nil, LDAP_AUTH_NEGOTIATE)) != LDAP_SUCCESS){
		ldap_unbind(con);
		sysfatal("ldap bind failed, %w", ldap_err2string(e));
	}

	sz = nelem(fqdn);
	if(GetComputerObjectName(NameFullyQualifiedDN, fqdn, &sz) == 0){
		ldap_unbind(con);
		sysfatal("cannot get my DC= name, %s", _winerr());
	}
	
	for(dc = fqdn; dc[0] && dc[1] && dc[2]; dc++)
		if(dc[0] == L'D' && dc[1] == L'C' && dc[2] == L'=')
			break;
	Maxage = maxpasswdage(con, dc);
	rc = search(con, dc, name);
	ldap_unbind(con);

	if(rc != 0)
		exits("failed");
	exits(nil);
}
