#include <u.h>
#include <libc.h>
#include <mswin.h>

#include <imagehlp.h>

static int
addr2line(void *addr)
{
	static Rune name[MAX_PATH];
	MEMORY_BASIC_INFORMATION mbi;

	if(VirtualQuery(addr, &mbi, sizeof(mbi)) == 0)
		return -1;
	if(GetModuleFileName((HINSTANCE)mbi.AllocationBase, name, nelem(name)) == 0)
		return -1;

	print("%S 0x%p 0x%p\n", name, mbi.AllocationBase, addr);
	return 0;
}

static void
backtrace(CONTEXT *c)
{
	void *ret, **frm;

	if(c){
 		ret = (void *)c->Eip;
		frm = (void **)c->Ebp;
	}
	else{
		ret = __builtin_return_address(0);
		asm("movl %%ebp, %0;" : "=r" (frm));	/* the syntax is painful... */
	}

	print("\n");

	/* register dump */
	if(c->ContextFlags & CONTEXT_CONTROL)
		print("eip=0x%-8lx esp=0x%-8lx ebp=0x%-8lx efl=0x%-8lx\n",
			c->Eip, c->Esp, c->Ebp, c->EFlags);
	
	if(c->ContextFlags & CONTEXT_INTEGER)
		print("eax=0x%-8lx ebx=0x%-8lx ecx=0x%-8lx "
			"edx=0x%-8lx esi=0x%-8lx edi=0x%-8lx\n",
			c->Eax, c->Ebx, c->Ecx, c->Edx, c->Esi, c->Edi);

	if(c->ContextFlags & CONTEXT_SEGMENTS)
		print("cs=0x%-8lx  ss=0x%-8lx  ds=0x%-8lx  "
			"es=0x%-8lx  fs=0x%-8lx  gs=0x%-8lx\n",
			c->SegCs, c->SegSs, c->SegDs, c->SegEs, c->SegFs, c->SegGs);

	print("\n");

	print("mswin/srcline\n");

	/* stack backtrace */
 	ret = (void *)c->Eip;
	frm = (void **)c->Ebp;
	addr2line(ret);
	for(; frm && frm[0] && frm[1]; frm = frm[0])
		addr2line(frm[1]);

	print(".\n");
}

static char *
oper(EXCEPTION_RECORD *er)
{
	if(er->NumberParameters >= 1)
		switch(er->ExceptionInformation[0]){
		case 0:
			return "read";
		case 1:
			return "write";
		case 8:
			return "execute";
		default:
			break;
		}
	return "unknown";
}

void
_fault(EXCEPTION_POINTERS *ep, char *note)
{
	int pid, why;
	long what;
	void *where;
	EXCEPTION_RECORD *er;

	er = ep->ExceptionRecord;

	print("\n");

	pid = GetCurrentProcessId();
	what = er->ExceptionInformation[1];
	where = er->ExceptionAddress;
	why = (int)er->ExceptionInformation[2];

	switch(er->ExceptionCode){
	case EXCEPTION_ACCESS_VIOLATION:
		print("%s %d: suicide sys: trap: %s fault addr=0x%lx pc=0x%08p\n",
			argv0, pid, oper(er), what, where);
		break;
	case EXCEPTION_IN_PAGE_ERROR:
		print("%s %d: suicide sys: trap: demand page %s failure addr=0x%08lx pc=0x%08p - %s\n",
			argv0, pid, oper(er), what, where, _ntdllerr(why));
		break;
	default:
		print("%s %d: suicide %s pc=0x%08p\n", argv0, pid, note, where);
	}

	backtrace(ep->ContextRecord);

	/*
	 * exit the process with a simple failure code, but not the exception code.
	 * Thus the parent process (if its rc(1)) does not print a summary
	 * suicide message for this process (as we have already printed
	 * a detailed one)
	 */
	ExitProcess(1);
}
