
enum {		/* file types */
	Tinvalid = -1,
	Tunknown = 0,	/* valid for allocfile() only, means try and work it out */
	Tfile,
	Tdev,
	Tdir,
	Tpipe,
	Tcons,
	Ttcp,
	Tudp,
};

enum{	/* fake exceptions that I generate */
	ConsExceptBase =			0xE0000100,		/* console control events */
	ConsExceptLimit =			0xE0000199,

	FAKE_ALARM_CALL = 			0xE0001000,
	FAKE_BROKEN_PIPE_WRITE =	0xE0001001,
	FAKE_KILL_PROCESS =			0xE0001001,

	FAKE_LOAD_FAIL = 			0xc0000135,		/* real Windows error code returned by CreateProcess() */

	Nexcept =					0,				/* note is an exception */
	Ncons = 					1,				/* note is a console control event */
	Nterm = 					2,				/* note is process terminate */
};

enum {	/* vfork flags */
	VFnone = 0,			/* no vfork in progress */
	VFforked,			/* have forked */
	VFexeced,			/* have execed */
	VFdetach			/* new note group, i.e. detach process */
};

enum
{
	NHASH = 32		/* number of rendez hash entries */
};

	
typedef struct File File;
typedef struct Proc Proc;
typedef struct Child Child;
typedef struct Egrp Egrp;
typedef struct Fgrp Fgrp;
typedef struct Rgrp Rgrp;
typedef struct Ngrp Ngrp;
typedef struct Wgrp Wgrp;
typedef struct Evalue Evalue;
typedef struct Ref Ref;
typedef struct Vfork Vfork;
typedef struct Note Note;
typedef struct DS DS;
typedef struct Rendtag Rendtag;

struct Ref {		/* refcount */
	QLock;
	long ref;
};

struct Evalue {		/* environment variable */
	QLock;
	char	*name;
	char	*value;
};

struct Child {		/* running child process */
	Child *next;
	HANDLE hand;	/* child's process handle that we can wait on */
	int pid;		/* child's process ID */
};

struct Egrp {		/* environment group */
	Ref;
	Egrp *next;
	int nent;		/* number of active entries */
	int ment;		/* number of entries allocated in ent */
	Evalue **ent;	/* vars */
};

struct Fgrp {		/* open file group */
	Ref;
	Fgrp *next;
	int nent;		/* number of active entries */
	int ment;		/* number of entries allocated in ent */
	File **ent;
};

struct Ngrp {		/* note group */
	Ref;
	Ngrp *next;
};

struct Rgrp {		/* rendezvous group */
	Ref;			/* both refcount and lock */
	Lock;
	Rgrp *next;
	HANDLE *wait;
	Rendtag *hashlst[NHASH];
	Rendtag *freelst;
};

struct Wgrp {		/* child wait group */
	Ref;			/* both refcount and lock */
	Wgrp *next;
	int nowait;		/* if true, don't add entries */
};

struct File {		/* open file */
	QLock;
	int fd;			/* file desceiptor */
	HANDLE hand;	/* windows handle */
	vlong off;		/* current file offset */
	int type;		/* file type (above) */
	int omode;		/* open mode */
	char *path;		/* path (for fd2path) */
	int slot;		/* directory slot */
};

struct Vfork {		/* active forks */
	Vfork *next;	
	int pid;		/* pid in forked environment */
	int ppid;		/* parent's pid in forked environment */
	int flags;		/* rfork flags */
};

struct Proc {
	Lock;
	int pid;							/* process ID */
	int ppid;							/* parent process ID */
	long expiry;						/* timer expiry deadline */
	HANDLE timer;						/* timer proc */
	HANDLE notethread;					/* thread which called notify */
	int notified;						/* flag, wait signaled due to interrupt */
	jmp_buf noted;						/* we jump here on noted() */
	void (*notify)(void *, char *);		/* note handler */
	char errstr[ERRMAX];				/* latest error string */
	char **argv;						/* command line */
	int argc;							/* number of args in command line */
	char *cwd;							/* current dir */
	int nprivates;						/* # private memory pointers allocated */
	void *privates[32];					/* private memory pointers */
	char *pathext;						/* contents of $PATHEXT from environment */
	Vfork *vfork;						/* current vforks in operation */
	Child *children;					/* children we can wait for */
	Fgrp *fgrp;							/* open files */
	Egrp *egrp;							/* environment variable group */
	Rgrp *rgrp;							/* rendezvous group */
	Ngrp *ngrp;							/* note group */
	Wgrp *wgrp;							/* child wait group */
};

struct Note {
	int code;
	int type;
	char *name;
};

struct DS {	/* dial string */
	char	buf[1024];
	char	*netdir;
	char	*proto;
	char	*host;
	char	*port;
};

void _cleanerr(char *);

char *_reslash(char *str);
int _allocfile(Fgrp *fg, HANDLE hand, int type, int omode, char *path, int wantfd);
File *_lookfile(Fgrp *fg, int fd);
File *_lookname(Fgrp *fg, char *path);
int _freefile(Fgrp *fg, int fd);
Fgrp *_dupfgrp(Fgrp *old);
void _pushfgrp(Proc *p, Fgrp *fg);
void _popfgrp(Proc *p);

char *_getcwd();

void _exits(char *s);

void _initproc(void);
Proc *_allocproc(void);
Proc *_getproc(void);
Proc *_getmainproc(void);
void _addchild(int pid, HANDLE hand);
void _freechild(int pid);
void _pushvfork(Proc *p, int flags);
void _revert(Proc *p);

int _tas(int *x);

int _wsainit(void);
void *_namequery(char *name);

int _dostat(char *path,  Dir *d);
int _dowstat(char *path, Dir *d);
int _readdir(File *f, void *edir, int nedir);
int _info2dir(char *path, wchar_t *tmp, BY_HANDLE_FILE_INFORMATION *in, Dir *d);
int _find2dir(char *path, wchar_t *tmp, WIN32_FIND_DATA *fd, Dir *d);
int _attr2dir(char *path, wchar_t *tmp, int attr, Dir *d);

void _importenv(wchar_t *env);
wchar_t *_exportenv();
char **_getenvp(void);
wchar_t *_winenviron(void);
Egrp *_dupegrp(Egrp *old);
void _pushegrp(Proc *p, Egrp *fg);
void _popegrp(Proc *p);

void _pushwgrp(Proc *p, Wgrp *fg);
void _popwgrp(Proc *p);

void _pushngrp(Proc *p, Ngrp *fg);
void _popngrp(Proc *p);

void _pushrgrp(Proc *p, Rgrp *fg);
void _poprgrp(Proc *p);

int _initsec(void);
char *_sid2name(PSID sid, int agression);

int GetParentProcessId(void);

void _panic(char *fmt, ...);
void _debug(char *fmt, ...);

void 		_heapdebug(char *file, int line);
#define heapchk		_heapdebug(__FILE__, __LINE__)

void incref(Ref *r);	/* lifted fron libthread */
long decref(Ref *r);

void _fault(EXCEPTION_POINTERS *ep, char *note);
void _fprst(CONTEXT *c);

BOOL WINAPI _catchcons(DWORD code);
LONG CALLBACK _catchexcept(EXCEPTION_POINTERS *ep);

Note *_code2note(int code);
Note *_name2note(char *note);

int _checkcase(wchar_t *tmp);

int _windevice(char *name);
