#include <u.h>
#include <libc.h>

#include <mswin.h>
#include "9win.h"

#include <winsock2.h>
#include <ws2tcpip.h>

static struct {
	char *proto;
	char *serv;
	char *port;
} porttbl[] = {
	"tcp",	"9fs",		"564",
	"tcp",	"whoami",	"565",
	"tcp",	"guard",	"566",
	"tcp",	"ticket",	"567",
	"tcp",	"exportfs",	"17007",
	"tcp",	"rexexec",	"17009",
	"tcp",	"ncpu",		"17010",
	"tcp",	"cpu",		"17013",
	"tcp",	"venti",	"17034",
	"tcp",	"wiki",		"17035",
	"tcp",	"secstore",	"5356",
	"udp",	"dns",		"53",
	"tcp",	"dns",		"53",
};

static char *
nineport(char *proto, char *serv)
{
	int i;

	for(i = 0; i < nelem(porttbl); i++)
		if(strcmp(proto, porttbl[i].proto) == 0 && strcmp(serv, porttbl[i].serv) == 0)
			return porttbl[i].port;

	if(strcmp(proto, "net") != 0)
		return serv;

	for(i = 0; i < nelem(porttbl); i++)
		if(strcmp(serv, porttbl[i].serv) == 0)
			return porttbl[i].port;

	return serv;
}

/*
 *  parse a dial string
 */
static void
_dial_string_parse(char *str, DS *ds)
{
	char *p, *p2;

	strncpy(ds->buf, str, sizeof(ds->buf));
	ds->buf[sizeof(ds->buf)-1] = 0;

	p = strchr(ds->buf, '!');
	if(p == 0) {
		ds->netdir = 0;
		ds->proto = "net";
		ds->host = ds->buf;
		
	} else {
		if(*ds->buf != '/' && *ds->buf != '#'){
			ds->netdir = 0;
			ds->proto = ds->buf;
		} else {
			/* expecting /net.alt/tcp!foo or #I1/tcp!foo */
			for(p2 = p; p2 > ds->buf && *p2 != '/'; p2--)
				;
			*p2++ = 0;
			ds->netdir = ds->buf;
			ds->proto = p2;
		}
		*p = 0;
		ds->host = p + 1;
		p = strchr(ds->host, '!');
		if(p){
			*p = 0;
			ds->port = p+1;
		}

	}
}

/* return a void* to so winsock doesn't have to be included everywhere */
void *
_namequery(char *str)
{
	DS ds;
	struct addrinfo *res, hints;

	if(str == nil)
		return nil;

	_dial_string_parse(str, &ds);
	if(ds.netdir != nil &&
	   strcmp(ds.netdir, "/net") != 0 &&
	   strcmp(ds.netdir, "#I") != 0){
		werrstr("'%s' bad net dir, only '/net' and '#I' supported", ds.netdir);
		return nil;
	}
	if(ds.host == nil || *ds.host == 0){
		werrstr("dial() %s malformed address", str);
		return nil;
	}

	memset(&hints, 0, sizeof(hints));
	if(strcmp(ds.proto, "udp") == 0){
		hints.ai_socktype = SOCK_DGRAM;
		hints.ai_protocol = IPPROTO_UDP;
	}
	else
	if(strcmp(ds.proto, "tcp") == 0){
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_protocol = IPPROTO_TCP;
	}
	else
	if(strcmp(ds.proto, "net") != 0){
		werrstr("dial: %s protocol not supported", ds.proto);
		return nil;
	}
	if(strcmp(ds.host, "*") == 0)
		ds.host = nil;

	ds.port = nineport(ds.proto, ds.port);

	res = nil;
	if(getaddrinfo(ds.host, ds.port, &hints, &res) != 0){
		werrstr("dial() getaddrinfo failed, %s", _skt_errstr());
		return nil;
	}
	return res;
}

