#include <u.h>
#include <libc.h>
#include <mswin.h>

#include <winsock2.h>
#include <ws2tcpip.h>

#include "9win.h"

static char *Protocols[] = {
	[Tudp] "udp",
	[Ttcp] "tcp",
};

static File *
getf(char *path)
{
	File *f;
	Fgrp *fg;

	fg = _getproc()->fgrp;
	qlock(fg);
	if ((f = _lookname(fg, path)) == nil){
		qunlock(fg);
		werrstr("%s not an open file\n", path);
		return NULL;
	}
	qlock(f);
	qunlock(fg);
	return f;
}

static int
putf(HANDLE h, int type, char *fmt, ...)
{
	int fd;
	Fgrp *fg;
	va_list ap;
	char path[NETPATHLEN];
	
	va_start(ap, fmt);
	vsnprint(path, sizeof(path), fmt, ap);
	va_end(ap);

	fg = _getproc()->fgrp;
	qlock(fg);
	fd = _allocfile(fg, h, type, ORDWR, path, -1);
	qunlock(fg);

	return fd;
}

int
announce(char *addr, char *dir)
{
	int n, type, ctl;
	struct addrinfo *loc, *l;

	if(_skt_startup() == -1)
		return -1;

	if((loc = _namequery(addr)) == nil)
		return -1;

	ctl = -1;
	loc = nil;
	for(l = loc; l; l = l->ai_next){
		if(l->ai_family == AF_INET || l->ai_family == AF_INET6){
			if((ctl = socket(l->ai_family, l->ai_socktype, l->ai_protocol)) == -1){
				werrstr("announce() socket failed, %r\n");
				continue;
			}

			if(l->ai_socktype == SOCK_STREAM){
				n = 1;
				if(setsockopt(ctl, SOL_SOCKET, SO_REUSEADDR, (char*)&n, sizeof n) != 0)
					fprint(2, "announce() warning, failed to enable address reuse, %s\n");
			}
			if(_skt_bind(ctl, l->ai_addr, l->ai_addrlen) < 0){
				werrstr("announce() bind failed, %r\n");
				closesocket(ctl);
				continue;
			}
			if(l->ai_socktype == SOCK_STREAM){
				if(_skt_listen(ctl, 8) < 0)
					werrstr("announce() listen failed, %r\n");
					closesocket(ctl);
					continue;
			}

			break;
		}
	}
	freeaddrinfo(loc);

	if(l == nil)
		return -1;

	if(l->ai_socktype == SOCK_STREAM)
		type = Ttcp;
	else
		type = Tudp;
	snprint(dir, NETPATHLEN, "#I/%s/%d", Protocols[type], ctl);
	return putf((HANDLE)ctl, type, "#I/%s/%d/ctl", Protocols[type], ctl);
}

int
listen(char *dir, char *newdir)
{
	File *f;
	int fd, skt;

	if((f = getf(dir)) == NULL)
		return -1;

	if((skt = _skt_accept((int)f->hand, nil, nil)) < 0)
		return -1;

	fd = putf((HANDLE)skt, f->type, "#I/%s/%d/data", Protocols[f->type], skt);
	if(fd == -1)
		_skt_close(skt);
	qunlock(f);
	return fd;
}

int
accept(int ctl, char *dir)
{
	File *f;
	int fd, one;

	if((f = getf(dir)) == NULL)
		return -1;

	one = 1;
	if(_skt_setsockopt((int)f->hand, IPPROTO_TCP, TCP_NODELAY, (char*)&one, sizeof(one)) != 0)
		fprint(2, "accept() Warning, failed to disable Naggle\n");

	fd = f->fd;
	qunlock(f);
	return fd;
}

int
reject(int ctl, char *dir, char *cause)
{
	File *f;

	USED(cause);

	if((f = getf(dir)) == NULL)
		return -1;
	_skt_shutdown((int)f->hand, SD_BOTH);
	_skt_close((int)f->hand);
	qunlock(f);
	return 0;
}
