#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"
#include "error.h"
#include <tlhelp32.h>

enum {
	WaitAny = 0,	/* return if any event is signaled */
	WaitAll = 1,	/* return only when all events have been signaled */

};

vlong
_ft2elap(FILETIME *ft)		/* returns absolute milliseconds */
{
	return (((vlong)ft->dwHighDateTime << 32) | ft->dwLowDateTime) / 10000LL;
}

int
await(char *buf, int len)
{
	Proc *p;
	Note *n;
	ulong status;
	HANDLE *hands;
	Child *c;
	int *pids;
	FILETIME us, sy, cr, ex;
	vlong user, syst, elap;
	int e, num, rc, idx, tot;

	p = _getproc();

	num = 0;
	for(c = p->children; c; c = c->next)
		num++;

	if(num == 0){
		werrstr(Enochild);
		return -1;
	}
	pids = malloc(sizeof(int) * (num+1));		/* +1 to leave space for p->rgrp->wait */
	hands = malloc(sizeof(HANDLE) * (num+1));

	tot = 0;
	hands[tot++] = p->rgrp->wait;
	for(c = p->children; c; c = c->next){
		pids[tot] = c->pid;
		hands[tot] = c->hand;
		tot++;
	}

	e = WaitForMultipleObjects(tot, hands, WaitAny, INFINITE);

	if(e > WAIT_OBJECT_0 && e < WAIT_OBJECT_0+tot){
		idx = e-WAIT_OBJECT_0;
		user = syst = elap = 0;

		if(GetProcessTimes(hands[idx], &cr, &ex, &sy, &us) != 0){
			user = _ft2elap(&us);
			syst = _ft2elap(&sy);
			elap = _ft2elap(&ex) - _ft2elap(&cr);
		}

		status = 0;
		if(GetExitCodeProcess(hands[idx], &status) < 0)
			_debug("GetExitCodeProcess() failed %s\n", _winerr());

		if((n = _code2note(status)) != nil)
			snprint(buf, len, "%d %llud %lld %llud 'suicide: %s'", pids[idx], user, syst, elap, n->name);
		else
		if(status == 0)
			snprint(buf, len, "%d %llud %lld %llud ''", pids[idx], user, syst, elap);
		else
			snprint(buf, len, "%d %llud %llud %llud %d", pids[idx], user, syst, elap, status);
		_freechild(pids[idx]);
		rc = strlen(buf);
	}
	else
	if(e == WAIT_OBJECT_0){
		rc = -1;
		werrstr("interrupted");
	}
	else{
		rc = -1;
		werrstr("await: %s", _winerr());
	}

	free(hands);
	free(pids);
	return rc;
}
