#include <u.h>
#include <libc.h>
#include <ctype.h>

/*
 * In place, rewrite name to compress multiple /, eliminate ., and process ..
 * - for win32 we add translating '\' to '/'
 * - added basic support for drive letters but
 *   this probably needs more work as d: and d:/ are not 
 *   equally valid to all win32 apis so we may need to do
 *   some more rewriting which could cause the string to grow
 *   (h:fred -> h:/fred) and this will break cleanname's API (sigh).
 * - don't compress two leading slashes as this is a valid windows path syntax.
 */
#define SEP(x)	((x)=='/' || (x) == 0)
char*
cleanname(char *name)
{
	char *p, *q, *dotdot;
	int rooted, erasedprefix;

	for(p = name; *p; p++)
		if(*p == '\\')
			*p = '/';

	rooted = 0;
	if(name[0] == '/'){
		rooted = 1;
		if(name[1] == '/')
			rooted = 2;
	}
	if(name[1] == ':' && isalpha(name[0])){
		rooted = 2;
		if(name[2] == '/')
			rooted++;
	}

	erasedprefix = 0;
	/*
	 * invariants:
	 *	p points at beginning of path element we're considering.
	 *	q points just past the last path element we wrote (no slash).
	 *	dotdot points just past the point where .. cannot backtrack
	 *		any further (no slash).
	 */
	p = q = dotdot = name+rooted;
	while(*p) {
		if(p[0] == '/')	/* null element */
			p++;
		else if(p[0] == '.' && SEP(p[1])) {
			if(p == name)
				erasedprefix = 1;
			p += 1;	/* don't count the separator in case it is nul */
		} else if(p[0] == '.' && p[1] == '.' && SEP(p[2])) {
			p += 2;
			if(q > dotdot) {	/* can backtrack */
				while(--q > dotdot && *q != '/')
					;
			} else if(!rooted) {	/* /.. is / but ./../ is .. */
				if(q != name)
					*q++ = '/';
				*q++ = '.';
				*q++ = '.';
				dotdot = q;
			}
			if(q == name)
				erasedprefix = 1;	/* erased entire path via dotdot */
		} else {	/* real path element */
			if(q != name+rooted)
				*q++ = '/';
			while((*q = *p) != '/' && *q != 0)
				p++, q++;
		}
	}
	if(q == name)	/* empty string is really ``.'' */
		*q++ = '.';
	*q = '\0';
	if(erasedprefix && name[0] == '#'){	
		/* this was not a #x device path originally - make it not one now */
		memmove(name+2, name, strlen(name)+1);
		name[0] = '.';
		name[1] = '/';
	}
	return name;
}
