#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"

int
create(char *path, int mode, ulong perm)
{
	int fd;
	Fgrp *fg;
	HANDLE h;
	wchar_t *tmp;
	int access, share, disp, attr, type;

	fg = _getproc()->fgrp;
	h = BadHand;
	tmp = _winpath(path);
	if(perm & DMDIR){
		if(mode != OREAD){
			werrstr("invalid mode");
			free(tmp);
			return -1;
		}

		if(CreateDirectory(tmp, nil) == 0){
			werrstr("'%s' %s", path, _winerr());
			free(tmp);
			return -1;
		}
		if((perm & 0222) == 0){
			if(SetFileAttributes(tmp, FILE_ATTRIBUTE_READONLY) == 0){
				RemoveDirectory(tmp);
				werrstr("'%s' %s", path, _winerr());
				free(tmp);
				return -1;
			}
		}
		qlock(fg);
		fd = _allocfile(fg, BadHand, Tdir, mode, path, -1);
		qunlock(fg);
	}
	else{
		switch(mode&3){
		default:
			access = 0;
			break;
		case OREAD:
			access = GENERIC_READ;
			break;
		case OWRITE:
			access = GENERIC_WRITE;
			break;
		case ORDWR:
			access = GENERIC_READ|GENERIC_WRITE;
			break;
		case OEXEC:
			access = GENERIC_EXECUTE;
			break;
		}

		disp = CREATE_ALWAYS;
		if(mode&OTRUNC)
			disp = TRUNCATE_EXISTING;

		attr = FILE_ATTRIBUTE_NORMAL;
		if((perm & 0222) == 0)
			attr |= FILE_ATTRIBUTE_READONLY;
		if(mode&ORCLOSE)
			attr |= FILE_FLAG_DELETE_ON_CLOSE;

		share = FILE_SHARE_WRITE|FILE_SHARE_READ|FILE_SHARE_DELETE;
		if(mode&OEXCL)
			share = 0;
		if((h = CreateFile(tmp, access, share, nil, disp, attr, nil)) == BadHand){
			werrstr("'%s' %s", path, _winerr());
			free(tmp);
			return -1;
		}
		type = Tfile;
	}
	free(tmp);
	qlock(fg);
	fd = _allocfile(fg, h, type, mode, path, -1);
	if(fd == -1)
		CloseHandle(h);
	qunlock(fg);
	return fd;

}
