#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"

int _nprivates;
void **_privates;

int
dequote(int n, char *argv[], char *p)
{
	char *p2;
	int i, j, quote, nbs;

	for(i=0; *p && i<n-1; i++) {
		while(*p == ' ' || *p == '\t')
			p++;
		quote = 0;
		argv[i] = p2 = p;
		for(;*p; p++) {
			if(!quote && (*p == ' ' || *p == '\t'))
				break;
			for(nbs=0; *p == '\\'; p++,nbs++)
				;
			if(*p == '"') {
				for(j=0; j<(nbs>>1); j++)
					*p2++ = '\\';
				if(nbs&1)
					*p2++ = *p;
				else
					quote = !quote;
			} else {
				for(j=0; j<nbs; j++)
					*p2++ = '\\';
				*p2++ = *p;
			}
		}
		/* move p up one to avoid pointing to null at end of p2 */
		if(*p)
			p++;
		*p2 = 0;
	
	}
	argv[i] = 0;
	return i;
}

/*
 * Windows adds a guard page after the end (lower address) of
 * the allocated stack so it can automagicially grow the stack.
 * after this is a protected page. The problem comes if your app
 * has some big automatics - it can step over the guard page and
 * hit the protected one and crash. This happens for real in sed.exe
 * which uses regexp which in turn is rather stack hungry. To avoid
 * this issue I just force windows to grow the stack size for the
 * app at startup to ensure we have plenty of space.
 *
 * NB some weird code here in an effort to defeat gcc's optimiser.
 */
static void
growstack(uchar *prev, int size)	
{
	static uchar buf[1000];		/* not a powwer of 2 */

	buf[0] = size;
	buf[1] = 17;
	if(prev)
		*prev = *buf;
	size -= sizeof(buf);
	if(size > 0)
		growstack(buf, size);
}

void
__main(void)
{
	/*
	 * GNU ld inserts a call to this to call static C++ constructors.
	 */
}

void
mainCRTStartup(void)
{
	int ac;
	Proc *p;
	char **av;
	wchar_t *cli, *env;
	extern void main(int, char **);
	
	/*
	 * Align the stack to 16 bytes for the sake of SSE 
	 * ops in main or in functions inlined into main.
	 */
	asm __volatile__("andl $-16, %%esp":::"%esp");

	_qlockinit(rendezvous);

	growstack(nil, 1024*1024*2);

	quotefmtinstall();		/* used for quoting $path and $pathext - see env.c */
	_initproc();
	_initsec();
	p = _allocproc();

	SetUnhandledExceptionFilter(_catchexcept);
	SetErrorMode(SEM_FAILCRITICALERRORS|SEM_NOGPFAULTERRORBOX|SEM_NOOPENFILEERRORBOX);

	if(SetConsoleCtrlHandler(_catchcons, 1) == 0)
		_panic("cannot catch ctrl-c etc - %s\n", _winerr());
	
	if((env = GetEnvironmentStrings()) == nil)
		_panic("cannot get environment - %s\n", _winerr());
	_importenv(env);

	if((cli = GetCommandLine()) == nil)
		_panic("cannot get command line - %s\n", _winerr());
	p->argc = dequote(ac, av, _nineargv(cli, &ac, &av));
	p->argv = av;

	/*
	 * in case the user doesn't use ARGBEGIN, we set argv0
	 * so fault() can print somthing useful if we hit an
	 * exception
	 */
	argv0 = p->argv[0];


	p->pathext = getenv("pathext");
	if(p->pathext == nil)
		p->pathext = ".exe .bat";

	qlock(p->fgrp);
	if(_allocfile(p->fgrp, GetStdHandle(STD_INPUT_HANDLE), Tunknown, OREAD, "/dev/stdin", 0) == -1)
		_panic("cannot attach stdin - %r\n");
	if(_allocfile(p->fgrp, GetStdHandle(STD_OUTPUT_HANDLE), Tunknown, OWRITE, "/dev/stdout", 1) == -1)
		_panic("cannot attach stdout - %r\n");
	if(_allocfile(p->fgrp, GetStdHandle(STD_ERROR_HANDLE), Tunknown, OWRITE, "/dev/stderr", 2) == -1)
		_panic("cannot attach stderr - %r\n");
	qunlock(p->fgrp);

	_nprivates = p->nprivates;
	_privates = p->privates;

	main(p->argc, p->argv);
	exits(nil);					/* be tolerant of misbehaviour */
}
