/*
 * dial - connect to a service
 */
#include <u.h>
#include <libc.h>
#include <mswin.h>

#include <winsock2.h>
#include <ws2tcpip.h>

#include "9win.h"

static char *Protocols[] = {
	[Tudp] "udp",
	[Ttcp] "tcp",
};

int
dial(char *remote, char *local, char *dir, int *cfdp)
{
	Fgrp *fg;
	int fd, type, skt, n;
	struct addrinfo *loc, *rem, *l, *r;

	if(_skt_startup() == -1){
		werrstr("dial() skt_startup failed, %r\n");
		return -1;
	}

	if((rem = _namequery(remote)) == nil)
		return -1;

	if(cfdp){
		*cfdp = -1;
		fprint(2, "dial: cfdp expected\n");
	}

	skt = -1;
	loc = nil;
	for(r = rem; r; r = r->ai_next){
		if(r->ai_family == AF_INET || r->ai_family == AF_INET6){
			if((skt = _skt_socket(r->ai_family, r->ai_socktype, r->ai_protocol)) == -1){
				werrstr("dial() socket failed, %r\n");
				continue;
			}

			if(local){
				if((loc = _namequery(local)) == nil)
					continue;
		
				for(l = loc; l; l = l->ai_next)
					if(l->ai_family == AF_INET || l->ai_family == AF_INET6){
						if(_skt_bind(skt, l->ai_addr, l->ai_addrlen) != -1)
							break;
				}
				if(l == nil){
					werrstr("dial() bind failed, %r\n");
					freeaddrinfo(loc);
					continue;
				}
			}
		
			if(r->ai_socktype == SOCK_STREAM){
				if(_skt_connect(skt, r->ai_addr, r->ai_addrlen) == -1){
					werrstr("dial() connect failed, %r\n");
					freeaddrinfo(loc);
					continue;
				}
				n = 1;
				if(_skt_setsockopt(skt, IPPROTO_TCP, TCP_NODELAY, (char*)&n, sizeof n) != 0)
					fprint(2, "dial() Warning, cannot disable Naggle, %r\n");
			}
			break;
		}
	}
	freeaddrinfo(loc);
	freeaddrinfo(rem);

	if(r == nil)
		return -1;

	if(r->ai_socktype == SOCK_STREAM)
		type = Ttcp;
	else
		type = Tudp;

	if(dir)
		snprint(dir, NETPATHLEN, "#I/%s/%d", Protocols[type], skt);

	fg = _getproc()->fgrp;
	qlock(fg);
	fd = _allocfile(fg, (HANDLE *)skt, type, ORDWR, "network", -1);
	if(fd == -1)
		_skt_close(skt);
	qunlock(fg);
	return fd;
}
