#define _GETENVP_EXTENSION
#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"

/*
 * Windows preserves case but ignores it, even in environment variables.
 * We need to be able to find these vars so we force their
 * case to that below as they are parsed from the windows.
 *
 * Maybe these should be held in a file and read by a 9win dll when
 * the first app attaches to the DLL?
 */
static struct {
	char *name;
	int split;
} Envfix[] = {
	{ "path",		1 },
	{ "pathext",	1 }
};

/*
 * quote whitespace so people can use 
 * tokenize() to split variables like $path
 */
static int
expand(char *name, char *value)
{
	int rc;
	Fmt f;
	char end, *e, *s, *tmp;

	rc = 0;

	fmtstrinit(&f);
	end = ';';
	s = value;
	for(s = value; end != 0; s = e+1){
		if((e = strchr(s, ';')) == nil)
			e = strchr(s, 0);
		end = *e;
		*e = 0;
		fmtprint(&f, "%q ", s);
	}
	tmp = fmtstrflush(&f);
	if(putenv(name, tmp) < 0)
		rc = -1;
	free(tmp);
	return rc;
}

void
_importenv(wchar_t *env)		/* used my crt0.c */
{
	int i;
	char *name, *value;

	while(*env){
		name = _ninestr(env);

		if((value = strchr(name, '=')) != nil){
			*value = 0;
			for(i = 0; i < nelem(Envfix); i++)
				if(cistrcmp(Envfix[i].name, name) == 0)
					break;

			if(i < nelem(Envfix)){
				if(Envfix[i].split)
					if(expand(Envfix[i].name, value+1) < 0)
						_panic("importenv: %s expand failed %r\n", Envfix[i].name);
			}
			else{
				if(putenv(name, value+1) < 0)
					_panic("importenv: %s putenv failed %r\n", name);
			}
		}
		free(name);
		while(*env++)
			continue;
	}
}

static void
contract(char *s, char *value)
{
	int n, i;
	char *a[1024];

	n = tokenize(value, a, nelem(a));
	if(n > 0)
		strcat(s, a[0]);
	for(i = 1; i < n; i++){
		strcat(s, ";");
		strcat(s, a[i]);
	}
}

/* used by spawn() */
wchar_t *
_exportenv(void)
{
	int i, n;
	Egrp *eg;
	Evalue **e;
	wchar_t *win;
	char *s, *str;

	eg = _getproc()->egrp;
	qlock(eg);
	n = 0;
	for(e = eg->ent; e < &eg->ent[eg->nent]; e++)
		n += strlen((*e)->name)+1 + strlen((*e)->value)+1;
	n++;

	if((str = malloc(n)) == nil){
		qunlock(eg);
		return nil;
	}

	s = str;
	for(e = eg->ent; e < &eg->ent[eg->nent]; e++){
		*s = 0;
		strcat(s, (*e)->name);
		strcat(s, "=");

		for(i = 0; i < nelem(Envfix); i++)
			if(strcmp((*e)->name, Envfix[i].name) == 0)
				break;
		if(i < nelem(Envfix) && Envfix[i].split)
			contract(s,  (*e)->value);
		else
			strcat(s, (*e)->value);
		s += strlen(s)+1;
	}
	*s = 0;
	qunlock(eg);


	win = _winenv(str);
	free(str);
	return win;
}

static Evalue *
lookenv(Egrp *eg, char *name)
{
	int i;
	Evalue *e;

	for(i=0; i<eg->nent; i++){
		e = eg->ent[i];
		if(name && strcmp(e->name, name) == 0)
			return e;
	}
	return nil;
}

/* used by rc(1) */
char **
_getenvp(void)
{
	int i, n;
	Egrp *eg;
	Evalue *e;
	char **envp, *strs;

	eg = _getproc()->egrp;

	n = 0;
	qlock(eg);
	for(i=0; i<eg->nent; i++){
		e = eg->ent[i];
		n += strlen(e->name)+1 + strlen(e->value)+1;
	}

	if((envp = malloc(n + (eg->nent+1)*sizeof(char *))) == nil){
		qunlock(eg);
		return nil;
	}
	strs = (char *)&envp[eg->nent+1];

	for(i=0; i<eg->nent; i++){
		e = eg->ent[i];
		envp[i] = strs;
		sprint(strs, "%s=%s", e->name, e->value);
		strs = strchr(strs, 0) +1;
	}
	envp[i] = 0;

	qunlock(eg);
	return envp;
}

char *
getenv(char *name)
{
	char *v;
	Egrp *eg;
	Evalue *e;

	eg = _getproc()->egrp;
	qlock(eg);
	e = lookenv(eg, name);
	if(e == nil){
		qunlock(eg);
		return nil;
	}
	v = strdup(e->value);
	qunlock(eg);
	return v;
}

int
putenv(char *name, char *value)
{
	Egrp *eg;
	Evalue *e, **ent;

	eg = _getproc()->egrp;

	qlock(eg);
	e = lookenv(eg, name);
	if(e){
		free(e->value);
		e->value = strdup(value);
		qunlock(eg);
		if(e->value == nil)
			return -1;
		return 0;
	}

	if(eg->nent >= eg->ment){
		eg->ment += 32;
		ent = malloc(sizeof(eg->ent[0])*eg->ment);
		if(ent == nil){
			qunlock(eg);
			return -1;
		}
		if(eg->nent)
			memmove(ent, eg->ent, sizeof(eg->ent[0])*eg->nent);
		free(eg->ent);
		eg->ent = ent;
	}

	e = malloc(sizeof(Evalue));
	if(e == nil){
		qunlock(eg);
		return -1;
	}
	e->name = strdup(name);
	if(e->name == nil){
		free(e);
		qunlock(eg);
		return -1;
	}
	e->value = strdup(value);
	if(e->value == nil){
		free(e->value);
		free(e);
		qunlock(eg);
		return -1;
	}

	eg->ent[eg->nent++] = e;
	qunlock(eg);
	return 0;
}

static int
envcmp(void *a, void *b)
{
	Evalue **x = a;
	Evalue **y = b;

	return cistrcmp((*x)->name, (*y)->name);
}

wchar_t *
_winenviron(void)		// called by spawn()
{
	char *s;
	Rune r;
	Evalue *e;
	int n, i, l;
	wchar_t *w, *win;
	Egrp *eg;

	eg = _getproc()->egrp;
	qlock(eg);

	/* Windows _needs_ its environment sorted - Ah bless... */
	qsort(eg->ent, eg->nent, sizeof(Evalue *), envcmp);

	l = 0;
	n = 0;
	for(i = 0; i < eg->nent; i++){
		e = eg->ent[i];
		l += utflen(e->name);
		if(e->value)
			l += utflen(e->value);
		l += 2;			// +2 for '\0' and '='
		n++;
	}
	l += 1;				// +1 for final '\0'

	if((win = malloc(l*sizeof(wchar_t))) == nil)
		_panic("newegrp: No memory - %r\n");

	w = win;
	for(i = 0; i < eg->nent; i++){
		e = eg->ent[i];
		s = e->name;
		while(*s){
			s += chartorune(&r, s);
			*w++ = r;
		}

		*w++ = L'=';

		s = e->value;
		while(s && *s){
			s += chartorune(&r, s);
			*w++ = r;
		}
		*w++ = 0;
	}
	*w++ = 0;
	qunlock(eg);
	return win;		
}

Egrp *
_dupegrp(Egrp *old)
{
	int i;
	Egrp *new;

	new = mallocz(sizeof(Egrp), 1);

	qlock(old);
	new->nent = old->nent;
	new->ment = old->ment;
	new->ent = malloc(sizeof(Evalue *)*new->ment);
	if(new->ent == nil)
		_panic("newegrp: No memory - %r\n");

	for(i = 0; i < new->nent; i++){
		new->ent[i] = malloc(sizeof(Evalue));
		if(new->ent[i] == nil)
			_panic("newegrp: No memory - %r\n");
		new->ent[i]->name = strdup(old->ent[i]->name);
		if(new->ent[i]->name == nil)
			_panic("newegrp: No memory - %r\n");
		new->ent[i]->value = strdup(old->ent[i]->value);
		if(new->ent[i]->value == nil)
			_panic("newegrp: No memory - %r\n");
	}
	qunlock(old);
	return new;
}

void
_pushegrp(Proc *p, Egrp *eg)
{
	if(eg == nil)
		eg = mallocz(sizeof(Egrp), 1);
	incref(eg);
	eg->next = p->egrp;
	p->egrp = eg;
}

void
_popegrp(Proc *p)
{
	int i;
	Egrp *eg;

	eg = p->egrp;
	if(decref(eg) != 0)
		return;

	p->egrp = eg->next;
	for(i = 0; i < eg->nent; i++){
		free(eg->ent[i]->name);
		free(eg->ent[i]->value);
		free(eg->ent[i]);
	}
	free(eg->ent);
	free(eg);
}


