#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <mswin.h>
#include "9win.h"
#include <sddl.h>
#include <aclapi.h>

int _basic_dirstat;

/*
 * - files created by Adminstrators are always in the
 * Adminstartors group, this is a feature of Windows.
 *
 * - files created by mortal users appear to be in the
 * group "Domain Users", if the computer is part of a domain.
 * If it is a standalone machine then the same SID resolves
 * to the user "None".
 *
 * This can be confusing.
 */
int
filesec(Dir *d, wchar_t *file, int attr, int agression)
{
	ACE_HEADER *hdr;
	ACL *dacl, *sacl;
	PSECURITY_DESCRIPTOR sd;
	SECURITY_INFORMATION si;
	ACL_SIZE_INFORMATION sz;
	ACCESS_DENIED_ACE *deny;
	ACCESS_ALLOWED_ACE *allow;
	PSID owner, group, sid;
	int rc, i, *p, m, err;
	int allowmask, denymask, mask;

	if(_basic_dirstat)
		return 0;

	rc = 0;
	sd = nil;
	dacl = nil;
	sacl = nil;
	owner = nil;
	group = nil;

	si = OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|DACL_SECURITY_INFORMATION;
	if((err = GetNamedSecurityInfo(file, SE_FILE_OBJECT, si, &owner, &group, &dacl, &sacl, &sd)) != ERROR_SUCCESS){
		SetLastError(err);
		werrstr("%s", _winerr());
		rc = -1;
		goto done;
	}

	free(d->uid);
	d->uid = _sid2name(owner, agression);
	free(d->gid);
	d->gid = _sid2name(group, agression);
	free(d->muid);
	d->muid = strdup(d->uid);

return 0;		// FIXME: disabled for now. need better algorithm for mapping ACLS to perms

	if(dacl == NULL){
		d->mode = 777;
		goto done;
	}
	if(GetAclInformation(dacl, (void *)&sz, sizeof(sz), AclSizeInformation) == 0){
		d->mode = 000;
		goto done;
	}
	if(sz.AceCount == 0){
		d->mode = 000;
		goto done;
	}

	allowmask = denymask = 0;
	for(i = 0; i < sz.AceCount; i++){
		if(GetAce(dacl, i, (void **)&hdr) == 0)
			continue;
		if(hdr->AceFlags & INHERIT_ONLY_ACE)
			continue;

		switch(hdr->AceType){
		case ACCESS_ALLOWED_ACE_TYPE:
			allow = (ACCESS_ALLOWED_ACE *)hdr;
			sid = &allow->SidStart;
			mask = allow->Mask;
			p = &allowmask;
			break;
		case ACCESS_DENIED_ACE_TYPE:
			deny = (ACCESS_DENIED_ACE *)hdr;
			sid = &deny->SidStart;
			mask = deny->Mask;
			p = &denymask;
			break;
		default:
			continue;
		}

		m = 0;
		if(mask & FILE_EXECUTE)
			m |= DMEXEC;
		if((mask & FILE_WRITE_DATA) || (mask & FILE_APPEND_DATA))
			m |= DMWRITE;
		if(mask & FILE_READ_DATA)
			m |= DMREAD;

		if(EqualSid(owner, sid))
			*p |= (m << 6) & ~(allowmask|denymask) & 0700;
		else
		if(EqualSid(group, sid))
			*p |= (m << 3) & ~(allowmask|denymask) & 0070;
		else
// FIXME		if(EqualSid(others, sid))
// FIXME			*p |= m & ~(allowmask|denymask) & 0007;
// FIXME		else
			continue;
	}

	d->mode = (d->mode & ~0777) | ((allowmask&~denymask) & 0777);

	if(attr & FILE_ATTRIBUTE_READONLY)
		d->mode &= ~0222;
done:
	free(dacl);
	free(sacl);
	free(owner);
	free(group);
	return rc;
}

static int
isexecutable(char *name)
{
	Proc *p;
	char *s, *e, *ext;

	if((ext = strrchr(name, '.')) == nil)
		return 0;

	p = _getproc();
	for(s = p->pathext; s; s = e+1){
		if((e = strchr(s, ' ')) == nil)
			e = strchr(s, 0);
		if(cistrncmp(ext, s, e-s) == 0)
			return 1;
		if(*e == 0)
			break;
	}
	return 0;
}


static void
baseinfo(Dir *d, wchar_t *file, int attr)
{
	wchar_t *w;
	char drive[UTFmax];

	d->type = 'M';
	d->dev = 0;							// 0 = unknown
	if(file[1] == L':'){
		runetochar(drive, file);
		if(islower(*drive))
			*drive = toupper(*drive);
		if(isalpha(*drive))
			d->dev = *drive - '@';			// 1 = a:, etc
	}

	d->qid.path = d->dev;
	for(w = file; *w; w++)
		d->qid.path = d->qid.path*1000003 + *w;

	d->uid = strdup("unknown");
	d->gid = strdup("unknown");
	d->muid = strdup("unknown");
	if(d->uid == nil || d->gid == nil || d->muid == nil)
		sysfatal("baseinfo: No memory, %r");

	d->qid.type = 0;
	if((attr & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY){
		d->mode = 0555;
		d->mode |= DMDIR;
		d->qid.type |= QTDIR;
	}
	else{
		d->mode = 0444;
		if(isexecutable(d->name))
			d->mode |= 0111;
		d->qid.type |= QTFILE;
	}

	if((attr & FILE_ATTRIBUTE_READONLY) == 0)
		d->mode |= 0222;
	if((attr & FILE_ATTRIBUTE_TEMPORARY) == FILE_ATTRIBUTE_TEMPORARY){
		d->mode |= DMTMP;
		d->qid.type |= QTTMP;
	}
	if((attr & FILE_ATTRIBUTE_DEVICE) == FILE_ATTRIBUTE_DEVICE){
		d->mode |= DMMOUNT;
		d->qid.type |= QTMOUNT;		/* never seen in the wild */
	}
	if((attr & FILE_ATTRIBUTE_REPARSE_POINT) == FILE_ATTRIBUTE_REPARSE_POINT){
		d->mode |= DMLINK;
		d->qid.type |= QTLINK;		/* win32 symlink, hidden in a spare mode bit */
	}
}

static char *
basename(char *path)
{
	char *p, *name;

	name = path;
	if((p = strrchr(path, '/')) != nil ||
	   (p = strrchr(path, '\\')) != nil)
		if(! isalpha(path[0]) || path[1] != ':' || p != path+2)
			name = p+1;

	if((name = strdup(name)) == nil)
		sysfatal("basename: No memory, %r");
	return name;
}

int
_info2dir(char *path, wchar_t *tmp, BY_HANDLE_FILE_INFORMATION *in, Dir *d)
{
	wchar_t *freeme;

	freeme = nil;
	if(tmp == nil)
		freeme = tmp = _winpath(path);

	d->name = basename(path);
	baseinfo(d, tmp, in->dwFileAttributes);
	filesec(d, tmp, in->dwFileAttributes, 99);

	free(freeme);

	d->length = ((vlong)in->nFileSizeHigh << 32) | in->nFileSizeLow;
	d->atime = _ft2epoch(&in->ftLastAccessTime)/1000;
	d->mtime = _ft2epoch(&in->ftLastWriteTime)/1000;
	d->qid.vers = d->mtime;
	return 0;
}

int
_find2dir(char *path, wchar_t *tmp, WIN32_FIND_DATA *fd, Dir *d)
{
	wchar_t *freeme;

	freeme = nil;
	if(tmp == nil){
		freeme = tmp = _winpath2(path, fd->cFileName);
		d->name = _ninepath(fd->cFileName);
	}
	else{
		d->name = basename(path);
	}

	baseinfo(d, tmp, fd->dwFileAttributes);
	filesec(d, tmp, fd->dwFileAttributes, 99);
	
	free(freeme);

	d->length = ((long long)fd->nFileSizeHigh<<32)|fd->nFileSizeLow;
	d->atime = _ft2epoch(&fd->ftLastAccessTime)/1000;
	d->mtime = _ft2epoch(&fd->ftLastWriteTime)/1000;
	d->qid.vers = d->mtime;
	return 0;
}

int
_attr2dir(char *path, wchar_t *tmp, int attr, Dir *d)
{
	d->name = basename(path);
	baseinfo(d, tmp, attr);
	filesec(d, tmp, attr, 99);

	d->length = 0;
	d->atime = time(nil);
	d->mtime = time(nil);
	d->qid.vers = d->mtime;
	return 0;
}

