#include <u.h>
#include <libc.h>
#include <fcall.h>
#include "error.h"
#include <mswin.h>
#include "9win.h"

int
fstat(int fd, uchar *edir, int nedir)
{
	Dir d;
	int rc;
	File *f;
	Fgrp *fg;
	ulong avail;
	BY_HANDLE_FILE_INFORMATION hfi;

	fg = _getproc()->fgrp;
	qlock(fg);
	if ((f = _lookfile(fg, fd)) == nil){
		qunlock(fg);
		return -1;
	}
	qlock(f);
	qunlock(fg);

	switch(f->type){
	case Tfile:
	case Tdir:
		if(f->hand == BadHand){
			if(_dostat(f->path, &d) == -1){
				qunlock(f);
				return -1;
			}
		}
		else{
			if(GetFileInformationByHandle(f->hand, &hfi) == 0){
				/*
				 * GetFileInformationByHandle() seems to fail 
				 * on devices so we call back to slower techniques
				 */
				if(_dostat(f->path, &d) == -1){
					qunlock(f);
					return -1;
				}
			}
			_info2dir(f->path, nil, &hfi, &d);
		}
		break;
	case Tpipe:
		if(PeekNamedPipe(f->hand, NULL, 0, NULL, &avail, NULL) == 0){
			werrstr("%s", _winerr());
			qunlock(f);
			return -1;
		}
		d.type = '|';
		d.length = avail;
		break;
	case Tinvalid:
		werrstr("0x%p, bad handle", f->hand);
		qunlock(f);
		return -1;
	default:
		werrstr(Ebadusefd);
		qunlock(f);
		return -1;
	}

	rc = convD2M(&d, edir, nedir);
	free(d.muid);
	free(d.name);
	free(d.uid);
	free(d.gid);
	qunlock(f);
	return rc;
}

