#include <u.h>
#include <libc.h>
#include <mswin.h>

#include "9win.h"
#include "error.h"

enum {
	NoteOff = 16,
	CodeKill = 16,		/* pretend code for kill note */
};

int
notify(void (*f)(void *uregs, char *note))
{
	Proc *p;

	p = _getproc();
	if(p->notethread)
		CloseHandle(p->notethread);

	p->notify = f;
	p->notethread = OpenThread(THREAD_ALL_ACCESS, FALSE, GetCurrentThreadId());
	return 0;
}

int
noted(int ret)
{
	Proc *p;

	p = _getproc();
	longjmp(p->noted, ret+NoteOff);
}

/*
 * Exceptions are Async Proceedure Calls made from the kernel,
 * wether we are in alertable state or not, thus up-> should
 * relate to our proc struct as expected.
 */
LONG CALLBACK
_catchexcept(EXCEPTION_POINTERS *ep)
{
	Proc *p;
	char *note;
	int rc, code;
	char buf[ERRMAX];
	EXCEPTION_RECORD *er;

_debug("%s:%d catching except\n", __FILE__, __LINE__);
	if((p = _getproc()) == nil)
		_panic("no proc\n");

	if(ep == nil)
		_panic("catchexcept: no exception pointers\n");
	if((er = ep->ExceptionRecord) == nil)
		_panic("catchexcept: no exception record\n");
	code = ep->ExceptionRecord->ExceptionCode;

_debug("%s:%d catching except %x\n", __FILE__, __LINE__, code);
	note = nil;
	if((note = _except2note(code)) == nil){
		snprint(buf, sizeof(buf), "unknown exception code=0x%x", code);
		note = buf;
	}

	if(p->notify){
		rc = setjmp(p->noted);
		switch(rc){
		case 0:
			p->notify((void *)ep, note);
			break;
		case NDFLT+NoteOff:
			break;
		case NCONT+NoteOff:
			if(er->ExceptionFlags & EXCEPTION_NONCONTINUABLE)
				sysfatal("continuing a non-continuable exception"); 
			SetEvent(p->rgrp->wait);		/* interrupt system call */
			p->notified = 1;
			if(strncmp(note, "sys: fp ", 8) == 0)
				_fprst(ep->ContextRecord);
			return 0;
		default:
			_panic("%d - bad noted() value\n", rc);
		}
	}

	_fault(ep, note);
	return 0;			/* shut up gcc */
}

static VOID CALLBACK
interrupt(ULONG_PTR code)
{
_debug("%s:%d raise %x\n", __FILE__, __LINE__, code);
	RaiseException((int)code, 0, 0, nil);
}

/*
 * Console events are a seperate thread created in our proc created
 * by the kernel, thus we must not use Thread Local Storage here.
 */
BOOL WINAPI
_catchcons(DWORD code)
{
	Proc *p;
_debug("%s:%d caught cons %d\n", __FILE__, __LINE__, code);

	p = _getproc();
//	if(p->notify == nil)
//		return 0; 	/* default handler */

_debug("%s:%d queue %d\n", __FILE__, __LINE__, code);
	QueueUserAPC(interrupt, p->notethread, (ULONG_PTR)code+ConsExceptBase);
	return 1;		/* ignore */
}


static int
postone(int code, int pid)
{
	HANDLE h;

	if(code == FAKE_KILL_PROCESS){
		if((h = OpenProcess(PROCESS_TERMINATE, FALSE, pid)) == NULL){
			werrstr("%s", _winerr());
			return -1;
		}
		if(TerminateProcess(h, 1) == 0){
			werrstr("%s", _winerr());
			CloseHandle(h);
			return -1;
		}
		CloseHandle(h);
	}

	if(code < ConsExceptBase || code >= ConsExceptLimit){
		werrstr("cannot post this type of note");
		return -1;
	}

	if(GenerateConsoleCtrlEvent(code - ConsExceptBase, pid) != 0){
		werrstr("%s", _winerr());
		return -1;
	}
	return 0;
}


int
postnote(int group, int pid, char *note)
{
	int code;

	if((code = _note2except(note)) == 0){
		werrstr("unknown note");
		return -1;
	}

	switch(group){
	case PNPROC:
		if(postone(code, pid) < 0)
			return -1;
		break;
	case PNGROUP:
		/* FIXME: walk over proc table looking for processes
		 * in this processes group an send them a signal
		 */
		werrstr("postnote(PNGROUP,...) not supported");
		return -1;
		break;
	default:
		werrstr(Ebadarg);
		return -1;
	}
	return 0;
}

void
notejmp(void *uregs, jmp_buf env, int val)
{
	longjmp(env, val);
}

