#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "error.h"
#include "9win.h"

long
pread(int fd, void *s, long n, vlong off)
{
	File *f;
	Proc *p;
	Fgrp *fg;
	DWORD got;
	OVERLAPPED ol;

	p = _getproc();
	fg = p->fgrp;
	qlock(fg);
	if ((f = _lookfile(fg, fd)) == nil){
		qunlock(fg);
		return -1;
	}
	qlock(f);
	qunlock(fg);

	/* FIXME: race on file offset here.
	 *
	 * We should use an offset of -1 if we know
	 * that we are writing at end of file. This will
	 * close the windows of the race except for a write
	 * which crosses the end-of-file boundry.
	 */
	if(off == -1LL)
		off = f->off;

	switch(f->type){
	case Tcons:
	case Tfile:
	case Tdev:
		if(f->omode == OWRITE){
			werrstr("file opened for write only");
			qunlock(f);
			return -1;
		}
		/* FALLTHRU */
	case Tpipe:

		memset(&ol, 0, sizeof(ol));
		ol.hEvent = p->rgrp->wait; 
		ol.Offset = off; 
		ol.OffsetHigh = off>>32;

		ResetEvent(p->rgrp->wait);
		if(ReadFile(f->hand, s, n, &got, &ol) == 0){
			switch(GetLastError()){
			case ERROR_SUCCESS:
				break;
			case ERROR_BROKEN_PIPE: 		/* p9 errrors on reader only */
			case ERROR_HANDLE_EOF:
			case ERROR_OPERATION_ABORTED:	/* console read was interrupted */
				qunlock(f);
				return 0;
			default:
				werrstr("%s", _winerr());
				CancelIo(f->hand);
				qunlock(f);
				return -1;
				break;
			}
		}
		break;
	case Tdir:
		got = _readdir(f, s, n);
		break;
	case Ttcp:
	case Tudp:
		if((got = _skt_read((int)f->hand, s, n)) == -1){
			qunlock(f);
			return -1;
		}
		break;
	case Tinvalid:
		werrstr("0x%p, bad handle", f->hand);
		qunlock(f);
		return -1;
	default:
		werrstr(Ebadusefd);
		qunlock(f);
		return -1;
	}
	f->off = off + got;
	qunlock(f);
	return got;
}


