#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"

/*
 * curproc will move to TLS when we port libthread
 */
static Proc *curproc;
static Proc *mainproc;

void
_panic(char *fmt, ...)
{
	ulong n;
	va_list ap;
	char buf[2048];

	n = snprint(buf, sizeof(buf), "%s: panic ", argv0);

	va_start(ap, fmt);
	n += vsnprint(buf+n, sizeof(buf)-n, fmt, ap);
	va_end(ap);

	WriteFile(GetStdHandle(STD_ERROR_HANDLE), buf, n, &n, nil);
	abort();
}

void
_debug(char *fmt, ...)
{
	ulong n;
	va_list ap;
	char buf[2048];

	va_start(ap, fmt);
	n = vsnprint(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	WriteFile(GetStdHandle(STD_ERROR_HANDLE), buf, n, &n, nil);
	Sleep(10);
}

void
_initproc(void)
{
}

Proc *
_getproc(void)
{
	return curproc;
}

Proc *
_getmainproc(void)
{
	return mainproc;
}

Proc *
_allocproc()
{
	Proc *p;

	if((p = mallocz(sizeof(Proc), 1)) == nil)
		_panic("newproc: No memory for proc\n");

	curproc = p;
	if(mainproc == nil)
		mainproc = curproc;

	_pushfgrp(p, nil);
	_pushegrp(p, nil);
	_pushrgrp(p, nil);
	_pushngrp(p, nil);
	_pushwgrp(p, nil);
	p->pid = GetCurrentProcessId();
	p->ppid = GetParentProcessId();
	p->cwd = _getcwd();
	p->nprivates = nelem(p->privates);
		
	return p;
}

void
_pushwgrp(Proc *p, Wgrp *wg)
{
	if(wg == nil)
		wg = mallocz(sizeof(Wgrp), 1);
	incref(wg);
	wg->next = p->wgrp;
	p->wgrp = wg;
}

void
_popwgrp(Proc *p)
{
	Wgrp *wg;

	wg = p->wgrp;

	p->wgrp = wg->next;
	if(decref(wg) != 0)
		return;
	free(wg);
}

void
_addchild(int pid, HANDLE hand)
{
	Proc *p;
	Child *c;

	p = _getproc();
	c = mallocz(sizeof(Child), 1);
	if(c == nil)
		_panic("_addchild: No memory - %r\n");
	c->pid = pid;
	c->hand = hand;
	c->next = p->children;
	p->children = c;
}

void
_freechild(int pid)
{
	Proc *p;
	Child **pc, *c;

	p = _getproc();
	pc = &p->children;
	for(c = p->children; c; pc = &c->next, c = c->next)
		if(c->pid == pid){
			*pc = c;
			return;
		}

	_panic("_freechild: %d child pid unknown\n", pid);
}

static jmp_buf Forktramp;

// void
// _trampoline(int argc, char *argv[])
// {
// 	Proc *p;
// 	Vfork *vf;
// 
// 	p = _getproc();
// 	if(setjmp(Forktramp)){
// 		vf = p->vfork;
// 		p->vfork = vf->next;
// 		free(vf);
// 		returntofork();				/* FIXME: Not implemented yet */
// 	}
// 	main(argc, argv);
// }


/* always called with p locked */
void
_pushvfork(Proc *p, int flags)
{
	Vfork *vf;

	vf = mallocz(sizeof(Vfork), 1);
	if(vf == nil)
		_panic("_pushvfork: No memory - %r\n");

	vf->pid = -1;
	vf->ppid = p->pid;
	vf->flags = flags;

	vf->next = p->vfork;
	p->vfork = vf;
}

void
_revert(Proc *p)
{
	int flags;
	Vfork *vf;

	if(p->vfork == nil)
		return;

	_popegrp(p);
	_popfgrp(p);
	_poprgrp(p);
	_popngrp(p);
	_popwgrp(p);
	
	lock(p);
	vf = p->vfork;
	if(vf == nil){
		unlock(p);
		return;
	}
	p->vfork = vf->next;
	unlock(p);

	flags = vf->flags;
	free(vf);

	if((flags & RFPROC) == RFPROC && (flags & RFMEM) != RFMEM)
		longjmp(Forktramp, 1);
}
