#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "error.h"
#include "9win.h"

enum {
	synchWrites = 0		/* true = synchronous writes, false = perform write behind */
};

long
pwrite(int fd, void *s, long n, vlong off)
{
	File *f;
	Fgrp *fg;
	Proc *p;
	DWORD got;
	OVERLAPPED ol;

	p = _getproc();
	fg = p->fgrp;
	qlock(fg);
	if ((f = _lookfile(fg, fd)) == nil){
		qunlock(fg);
		return -1;
	}
	qlock(f);
	qunlock(fg);

	if(off == -1)
		off = f->off;

	switch(f->type){
	case Tfile:
	case Tdev:
	case Tcons:
		if(f->omode == OREAD){
			werrstr("file opened for read only", f->omode);
			qunlock(f);
			return -1;
		}
		/* FALLTHRU */
	case Tpipe:

		memset(&ol, 0, sizeof(ol));
		ol.hEvent = p->rgrp->wait; 
		ol.Offset = off; 
		ol.OffsetHigh = off>>32; 
	
		ResetEvent(ol.hEvent);
		if(WriteFile(f->hand, s, n, &got, &ol) == 0){
			switch(GetLastError()){
			case ERROR_SUCCESS:
				break;
			case ERROR_BROKEN_PIPE:
				RaiseException(FAKE_BROKEN_PIPE_WRITE, 0, 0, nil);
				CancelIo(f->hand);
				qunlock(f);
				return -1;
			default:
				werrstr("%s", _winerr());
				CancelIo(f->hand);
				qunlock(f);
				return -1;
			}
		}
		break;
	case Tdir:
		werrstr("cannot write directory");
		qunlock(f);
		return -1;
	case Ttcp:
	case Tudp:
		if((got = _skt_write((int)f->hand, s, n)) == -1){
			qunlock(f);
			return -1;
		}
		break;
	case Tinvalid:
		werrstr("0x%p, bad handle", f->hand);
		qunlock(f);
		return -1;
	default:
		werrstr(Ebadusefd);
		qunlock(f);
		return -1;
	}
	f->off = off + got;
	qunlock(f);
	return got;
}
