#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"

int
rfork(int flags)
{
	Proc *p;
	int fail;

	fail = 0;
	if((flags & RFPROC) == RFPROC && 
	   (flags & RFMEM) == RFMEM){
		fprint(2, "rfork(RFPROC|RFMEM) - fork thread, not implemented\n");
		fail = 1;
	}
	if((flags & (RFCNAMEG)) == RFCNAMEG){
		fprint(2, "rfork(RFCNAMEG) - clear name group, not implemented\n");
		fail = 1;
	}
	if((flags & (RFNAMEG)) == RFNAMEG){
		fprint(2, "rfork(RFNAMEG) - copy name group, not implemented\n");
		fail = 1;
	}
	if((flags & RFNOMNT) == RFNOMNT){
		fprint(2, "rfork(RFNOMNT) - disallow new mounts, not implemented\n");
		fail = 1;
	}

	if(fail){
		werrstr("rfork(), feature not implemented");
		return -1;
	}

	p = _getproc();
	lock(p);

	/********** environment variables */
	if((flags & RFENVG) == RFENVG)
		_pushegrp(p, _dupegrp(p->egrp));
	else
	if((flags & RFCENVG) == RFCENVG)
		_pushegrp(p, nil);
	else
		incref(p->egrp);

	/********** open file descriptors */
	if((flags & RFFDG) == RFFDG)
		_pushfgrp(p, _dupfgrp(p->fgrp));
	else
	if((flags & RFCFDG) == RFCFDG)
		_pushfgrp(p, nil);
	else
		incref(p->fgrp);

	/********** rendezvouz */
	if((flags & RFREND) == RFREND)
		_pushrgrp(p, nil);
	else
		incref(p->rgrp);

	/********** notes */
	if((flags & RFNOTEG) == RFNOTEG)
		_pushngrp(p, nil);
	else
		incref(p->ngrp);

	/********** don't wait for children */
	if((flags & RFNOWAIT) == RFNOWAIT)
		_pushwgrp(p, nil);
	else
		incref(p->wgrp);

	/********** new process (vfork) */
	if((flags & RFPROC) == RFPROC && 
	  (flags & RFMEM) != RFMEM)
		_pushvfork(p, flags);

	unlock(p);
	return 0;
}

