#define _SPAWN_EXTENSION
#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"
#include "error.h"

enum {
	Debugpath = 0,
};

static int expand(char *exe, char *file, char **bangv);

/*
 * parse shebang into an argv array
 */
static int
shargs(char *s, int n, char **bangv)
{
	int i;

	for(i=0; s[i]!='\n'; i++)
		if(i == n-1)
			return 0;
	s[i] = 0;

	if(Debugpath)
		fprint(2, "got %s\n", s);

	*bangv = 0;
	i = 0;
	for(;;) {
		while(*s==' ' || *s=='\t')
			s++;
		if(*s == 0)
			break;
		i++;
		*bangv++ = s;
		*bangv = 0;
		while(*s && *s!=' ' && *s!='\t')
			s++;
		if(*s == 0)
			break;
		else
			*s++ = 0;
	}
	return i;
}

static int
shebang(char *exe, char *file, char **bangv)
{
	int n, fd;
	char *plan9, *cputype, tmp[MAX_PATH];

	/* is it an rc script, or an extension-less windows executable? */
	if((fd = open(file, OREAD)) == -1)
		return -1;
	n = read(fd, tmp, sizeof(tmp));
	close(fd);
	if(n <= 0)
		return -1;
	tmp[n-1] = 0;

	if(strncmp(tmp, "MZ", 2) == 0)			/* is an exe but has no extension */
		return 0;
	if(strncmp(tmp, "#!", 2) != 0)			/* no shebang */
		return -1;
	if(shargs(tmp+2, n-2, bangv) <= 0)		/* parse args */
		return -1;

	if((plan9 = getenv("plan9")) == nil)
		sysfatal("plan9 - environment variable not set\n", 0);
	if((cputype = getenv("cputype")) == nil)
		sysfatal("$cputype environment variable not set\n", 0);
	snprint(tmp, sizeof(tmp), "%s/%s%s", plan9, cputype, bangv[0]);
	free(plan9);
	free(cputype);

	/*
	 * may gave to recurse into expand as we may heve been given
	 * /bin/rc which we convert to $plan9/$cputype/bin/rc but
	 * we need expand to append the .exe to it.
	 */
	expand(exe, tmp, nil);
	bangv[0] = exe;
	return 0;
}

/*
 * This could become a cached lookup to compensate for the
 * miserable performance of CIFS remote directory lookups
 */
static int
exists(char *exe, char *fmt, ...)
{
	int a;
	va_list ap;
	wchar_t *wexe;

	va_start(ap, fmt);
	vsnprint(exe, MAX_PATH, fmt, ap);
	va_end(ap);

	wexe = _winpath(exe);
	a = GetFileAttributes(wexe);
	free(wexe);

	if(Debugpath)
		fprint(2, "exists: %s\n", exe);

	if(a == -1)
		return -1;
	if(a == FILE_ATTRIBUTE_DIRECTORY)
		return -1;
	return 0;
}

static int
expand(char *exe, char *file, char **bangv)
{
	Proc *p;
	char *ext, *s, *e;
	int hasext;
	
	p = _getproc();

	/* has it a valid looking extension? */
	hasext = 0;
	if((ext = strrchr(file, '.')) != 0){
		for(s = p->pathext; s; s = e+1){
			if((e = strchr(s, ' ')) == nil)
				e = strchr(s, 0);
			if(cistrncmp(ext, s, e-s) == 0){
				hasext++;
				if(exists(exe, "%s", file) == 0)
					return 0;
			}
		if(*e == 0)
			break;
		}
	}

	/* try appending a known extensions (.EXE files etc)*/
	if(! hasext){
		for(s = p->pathext; s; s = e+1){
			if((e = strchr(s, ' ')) == nil)
				e = strchr(s, 0);
			if(exists(exe, "%s%.*s", file, e-s, s) == 0)
				return 0;
			if(*e == 0)
				break;
		}
	}

	/* try the unadorned path */
	if(exists(exe, "%s", file) == 0){

		/* expand #! in a shell script */
		if(bangv)
			if(shebang(exe, file, bangv) == 0)
				return 0;
	}
	return -1;
}

/*
 * windows quoting rules - I think
 * Words are seperated by space or tab
 * Words containing a space or tab can be quoted using "
 * 2N backslashes + " ==> N backslashes and end quote
 * 2N+1 backslashes + " ==> N backslashes + literal "
 * N backslashes not followed by " ==> N backslashes
 */
static char *
dblquote(char *cmd, char *s)
{
	int nb;
	char *p;

	for(p=s; *p; p++)
		if(*p == ' ' || *p == '\t' || *p == '\n' || *p == '\r' || *p == '"')
			break;

	if(p == s){					/* empty arg */
		strcpy(cmd, "\"\"");
		return cmd+2;
	}

	if(*p == 0){				/* easy case */
		strcpy(cmd, s);
		return cmd+(p-s);
	}

	*cmd++ = '"';
	for(;;) {
		for(nb=0; *s=='\\'; nb++)
			*cmd++ = *s++;

		if(*s == 0) {				/* trailing backslashes -> 2N */
			while(nb-- > 0)
				*cmd++ = '\\';
			break;
		}

		if(*s == '"') {			/* literal quote -> 2N+1 backslashes */
			while(nb-- > 0)
				*cmd++ = '\\';
			*cmd++ = '\\';			/* escape the quote */
		}
		*cmd++ = *s++;
	}

	*cmd++ = '"';
	*cmd = 0;

	return cmd;
}

static char *
proccmd(char **bangv, char **argv)
{
	int i, n;
	char *cmd, *p;

	/* conservatively calculate length of command;
	 * backslash expansion can cause growth in dblquote().
	 */
	n = 0;
	for(i=0; bangv[i]; i++)
		n += (bangv[i])? 2*strlen(bangv[i]): 2;
	for(i=0; argv[i]; i++)
		n += (argv[i])? 2*strlen(argv[i]): 2;
	n++;
	
	cmd = malloc(n * sizeof(char));
	if(cmd == nil)
		_panic("no memory for exec\n");

	p = cmd;
	for(i=0; bangv[i]; i++){
		p = dblquote(p, bangv[i]);
		*p++ = ' ';
	}
	for(i=0; argv[i]; i++){
		p = dblquote(p, argv[i]);
		*p++ = ' ';
	}
	if(p != cmd)
		p--;
	*p = 0;

	return cmd;
}

static HANDLE
fdexport(Fgrp *fg, int fd)
{
	File *f;
	HANDLE r;
	enum { Inherit = 1 };

	qlock(fg);
	if((f = _lookfile(fg, fd)) == nil){
		qunlock(fg);
		_debug("spawn/fdexport: fd=%d cannot find file descriptor\n", fd);
		return INVALID_HANDLE_VALUE;
	}
	qlock(f);
	qunlock(fg);

	if(!DuplicateHandle(GetCurrentProcess(), f->hand,
			GetCurrentProcess(), &r, DUPLICATE_SAME_ACCESS,
			Inherit, DUPLICATE_SAME_ACCESS)){
		qunlock(f);
		_debug("spawn/fdexport: fd=%d cannot duplicate handle, %s\n", _winerr());
		return INVALID_HANDLE_VALUE;
	}
	SetHandleInformation(r, HANDLE_FLAG_INHERIT, 1);
	qunlock(f);
	return r;
}

int
_spawn(void *tok, char *name, char *argv[])
{
	int r;
	Proc *p;
	STARTUPINFO si;
	wchar_t *wexe, *wcmd, *wenv;
	PROCESS_INFORMATION pi;
	char *plan9, *cputype, rcpath[MAX_PATH];
	char exe[MAX_PATH], *bangv[1024], *cmd;

	p = _getproc();
	bangv[0] = 0;

	/*
	 * we treat /bin/rc as special.
	 * plan9 rarely uses absolute paths but this one is common
	 */
	if(strcmp(name, "/bin/rc") == 0){
		if((plan9 = getenv("plan9")) != nil && (cputype = getenv("cputype")) != nil)
			snprint(rcpath, sizeof(rcpath), "%s/%s/bin/rc.exe", plan9, cputype, name);
		free(plan9);
		name = rcpath;
	}

	if(expand(exe, name, bangv) == -1)
		return -1;
	wexe = _winpath(exe);

	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	si.dwFlags = STARTF_USESHOWWINDOW|STARTF_USESTDHANDLES;
	si.wShowWindow = SW_HIDE;
	si.hStdInput = fdexport(p->fgrp, 0);
	si.hStdOutput = fdexport(p->fgrp, 1);
	si.hStdError = fdexport(p->fgrp, 2);
	si.lpDesktop = L"";		// we want a new WinStation per session

	wenv = _exportenv();
	cmd = proccmd(bangv, argv);
	wcmd = _winstr(cmd);

	memset(&pi, 0, sizeof(pi));
	if(tok == INVALID_HANDLE_VALUE || tok == nil)
		r = CreateProcess(wexe, wcmd, nil, nil, TRUE,
			CREATE_NEW_PROCESS_GROUP|CREATE_UNICODE_ENVIRONMENT, wenv, nil, &si, &pi);
	else
		r = CreateProcessAsUser(tok, wexe, wcmd, nil, nil, TRUE,
			CREATE_NEW_PROCESS_GROUP|CREATE_UNICODE_ENVIRONMENT|CREATE_DEFAULT_ERROR_MODE,
			wenv, nil, &si, &pi);

	Sleep(2);			/* allow child to run */

	free(cmd);
	free(wenv);
	free(wexe);
	free(wcmd);
	CloseHandle(si.hStdInput);
	CloseHandle(si.hStdOutput);
	CloseHandle(si.hStdError);

	if(r == 0){
		werrstr("spawn failed, %s", _winerr());
		return -1;
	}

	if(p->wgrp->nowait)
		CloseHandle(pi.hProcess);
	else
		_addchild(pi.dwProcessId, pi.hProcess);

	_revert(p);
	return pi.dwProcessId;
}
