#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <mswin.h>

#include <sddl.h>
#include <aclapi.h>

typedef struct User User;
struct User {
	User *next;
	SID *sid;
	wchar_t *dom;
	wchar_t *nam;
};

static struct {
	User *first;
	Lock lock;
} Usrlst;

static SID *others;
static SID *creatorowner;
static SID *creatorgroup;

static wchar_t Username[128];
static wchar_t Domainname[128];
static wchar_t Sysname[128];

/* Well known dull domain names */
static wchar_t *Wellknown[] = {
	L"NT AUTHORITY",
	L"NT SERVICE",
	L"BUILTIN",
};

static wchar_t *
wstrdup(wchar_t *s)
{
	int n;
	wchar_t *t;

	n = _winlen(s);
	if((t = malloc(sizeof(wchar_t)*(n+1))) == nil)
		sysfatal("wstrdup: No memory, %r");
	memmove(t, s, sizeof(wchar_t)*(n+1));
	return t;
}

static int
isdull(wchar_t *dom, int agressive)
{
	int i;

	if(agressive > 0)		/* domain is our login domain */
		if(_wincmp(dom, Domainname) == 0)
			return 1;
	if(agressive > 1)		/* domain is local host name */
		if(_wincmp(dom, Sysname) == 0)
			return 1;
	if(agressive > 2)		/* domain is "Well known" */
		for(i = 0; i < nelem(Wellknown); i++)
			if(_wincmp(dom, Wellknown[i]) == 0)
				return 1;
	return 0;
}

static void
addsid(PSID sid, wchar_t *dom, wchar_t *nam)
{
	int n;
	User *u;

	if((u = malloc(sizeof(User))) == nil)
		sysfatal("addsid: No memory, %r");
	n = GetLengthSid(sid);
	if((u->sid = malloc(n)) == nil)
		sysfatal("addsid: No memory, %r");
	if(CopySid(n, u->sid, sid) == 0)
		sysfatal("addsid: CopySid failed, %s", _winerr());
	if((u->dom = wstrdup(dom)) == nil)
		sysfatal("addsid: No memory, %r");
	if((u->nam = wstrdup(nam)) == nil)
		sysfatal("addsid: No memory, %r");

	lock(&Usrlst.lock);
	u->next = Usrlst.first;
	Usrlst.first = u;
	unlock(&Usrlst.lock);

}

char *
looksid(PSID sid, int agression)
{
	User *u;

	lock(&Usrlst.lock);
	for(u = Usrlst.first; u; u = u->next)
		if(EqualSid(u->sid, sid))
			break;
	unlock(&Usrlst.lock);

	if(u == nil)
		return nil;
	if(isdull(u->dom, agression))
		return _nineuser(nil, u->nam);
	else
		return _nineuser(u->dom, u->nam);

}

static void
allocsid(SID_IDENTIFIER_AUTHORITY *sia, int s1, int s2, int s3, SID **p)
{
	AllocateAndInitializeSid(sia, s1, s2, s3, 0, 0, 0, 0, 0, 0, (void **)p);
}

int
_initsec(void)
{
	HANDLE h;
	TOKEN_USER *pu;
	SID_NAME_USE snu;
	DWORD sz, sz1, sz2;
	unsigned char buf[1024];

	/*
	 * SID_IDENTIFIER_AUTHORITY creator = SECURITY_CREATOR_SID_AUTHORITY;
	 * SID_IDENTIFIER_AUTHORITY world = SECURITY_WORLD_SID_AUTHORITY;
	 * SID_IDENTIFIER_AUTHORITY nt = SECURITY_NT_AUTHORITY;
	 * Not quite sure of the "correct" thing to do here.
	 * We can check for permission for the Everyone" group as others
	 * but in our (domain) install this is not used.
	 * Instead the "Users" group seems to make more sense and
	 * that is why we use this here; YMMV.
	 * if(0)
	 * 	allocsid(&world, 1, SECURITY_WORLD_RID, 0, &others);
	 * else
	 * 	allocsid(&nt, 2, SECURITY_BUILTIN_DOMAIN_RID, DOMAIN_ALIAS_RID_USERS, &others);
	 * allocsid(&creator, 1, SECURITY_CREATOR_OWNER_RID, 0, &creatorowner);
	 * allocsid(&creator, 1, SECURITY_CREATOR_GROUP_RID, 0, &creatorgroup);
	 */

	sz = sizeof(Sysname);
	GetComputerName(Sysname, &sz);

	if(OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY|TOKEN_QUERY_SOURCE, &h) == 0)
		sysfatal("_initsec: cannot open current process, %s", _winerr());
	sz = sizeof(buf);
	if(GetTokenInformation(h, TokenUser, buf, sz, &sz) == 0)
		sysfatal("_initsec: cannot open current process token, %s", _winerr());
	CloseHandle(h);
	sz1 = sizeof(Username);
	sz2 = sizeof(Domainname);
	pu = (TOKEN_USER *)buf;
	if(LookupAccountSid(nil, pu->User.Sid, Username, &sz1, Domainname, &sz2, &snu) == 0)
		sysfatal("_initsec: cannot find current account name, %s", _winerr());

	return 0;
}

char *
_sid2name(PSID sid, int agression)
{
	char *name;
	SID_NAME_USE snu;
	DWORD sz1, sz2;
	wchar_t dom[128], nam[128];

	if((name = looksid(sid, agression)) != nil)
		return name;

	sz1 = sizeof(nam);
	sz2 = sizeof(dom);
	if(LookupAccountSid(nil, sid, nam, &sz1, dom, &sz2, &snu) == 0){
		werrstr("sid2name: LookupAccountSid: %s", _winerr());
		return strdup("failed");
	}

	switch(snu) {
	case SidTypeDeletedAccount:
		name = strdup("deleted");
		break;
	case SidTypeInvalid:
		name = strdup("invalid");
		break;
	case SidTypeUnknown:
		name = strdup("unknown");
		break;
	default:
		addsid(sid, dom, nam);
		name = looksid(sid, agression);
		break;
	}
	if(name == nil)
		sysfatal("_sid2name: No memory, %r");
	return name;
}

